/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Locale;

import org.vono.narau.db.CommonDB;
import org.vono.narau.models.kanji.ConvKanaRomanji;
import org.vono.narau.models.kanji.ConvRomanjiKana;
import org.vono.narau.models.kanji.Kanji;
import org.vono.narau.preferences.Preferences;

import android.content.Context;
import android.content.pm.PackageManager.NameNotFoundException;
import android.os.Environment;
import android.util.Log;

public class Common {
	private static final String TAG = Common.class.getSimpleName();

	public static final String EMPTY_STRING = "";
	public static final char SLASH = '/';
	public static final String SCHEME = "narau";

	public enum KanaType {
		ROMANJI, HIRAGANA, KATAKANA, KANJI
	}

	/**
	 * R = Romanji<br/>
	 * H = Hiragana<br/>
	 * K = Katakana
	 */
	private enum ConvType {
		RH, RK, HR, HK, KR, KH
	}

	private static class ConvData {
		private HashMap<String, String> map;
		private int maxStrLen;
	}

	private static boolean initDone = false;
	private static Context context = null;
	private static HashMap<String, Locale> mapLocales = null;
	public static String packageName = null;
	public static File dataPath = null;
	private static HashMap<ConvType, ConvData> mapConv = null;

	private static ConvData buildConvData(ConvType convType) {

		HashMap<String, String> map = null;
		int maxStrLen = 0;

		String[][] sourceArray = null;
		int sourcePos = 0;
		int destPos = 0;

		// Get source and dest informations
		switch (convType) {
		case RH:
			sourceArray = ConvRomanjiKana.ROMANJI_KANA;
			sourcePos = 0;
			destPos = 1;
			break;
		case RK:
			sourceArray = ConvRomanjiKana.ROMANJI_KANA;
			sourcePos = 0;
			destPos = 2;
			break;
		case HK:
			sourceArray = ConvRomanjiKana.ROMANJI_KANA;
			sourcePos = 1;
			destPos = 2;
			break;
		case HR:
			sourceArray = ConvKanaRomanji.HIRAGANA_ROMANJI;
			sourcePos = 0;
			destPos = 1;
			break;
		case KH:
			sourceArray = ConvRomanjiKana.ROMANJI_KANA;
			sourcePos = 2;
			destPos = 1;
			break;
		case KR:
		default:
			sourceArray = ConvKanaRomanji.KATAKANA_ROMANJI;
			sourcePos = 0;
			destPos = 1;
			break;
		}

		// Build the map
		int arraySize = sourceArray.length;
		map = new HashMap<String, String>(arraySize);
		for (int i = 0; i < arraySize; i++) {
			String source = sourceArray[i][sourcePos];
			map.put(source, sourceArray[i][destPos]);
			if (maxStrLen < source.length()) {
				maxStrLen = source.length();
			}
		}

		ConvData convData = new ConvData();
		convData.map = map;
		convData.maxStrLen = maxStrLen;

		mapConv.put(convType, convData);

		return convData;
	}

	private static ConvType KanaTypeToConvType(KanaType from, KanaType to) {
		String convString = null;
		switch (from) {
		case HIRAGANA:
			convString = "H";
			break;
		case KATAKANA:
			convString = "K";
			break;
		case ROMANJI:
			convString = "R";
			break;
		case KANJI:
			Log.e(TAG, "Can't convert from kanji");
			throw new RuntimeException("Can't convert from kanji");
		}
		switch (to) {
		case HIRAGANA:
			convString = convString + "H";
			break;
		case KATAKANA:
			convString = convString + "K";
			break;
		case ROMANJI:
			convString = convString + "R";
			break;
		case KANJI:
			Log.e(TAG, "Can't convert to kanji");
			throw new RuntimeException("Can't convert to kanji");
		}

		return ConvType.valueOf(convString);
	}

	/**
	 * Based on: <a
	 * href="http://www.mauvecloud.net/charsets/kanagenerator.html">
	 * kanagenerator</a> and <a
	 * href="http://www.mauvecloud.net/charsets/romaji.html">romaji</a>
	 * 
	 * @param originalText
	 *            The text to convert
	 * @param from
	 *            The text kana type
	 * @param to
	 *            The Kana type to convert the text
	 * @return the converted string
	 */
	public static String convertKana(String originalText, KanaType from, KanaType to) {
		if (originalText == null || originalText.length() == 0) {
			return EMPTY_STRING;
		}
		String text = originalText;
		if (KanaType.ROMANJI == from) {
			text = text.toLowerCase(Locale.US);
		}

		ConvType convType = Common.KanaTypeToConvType(from, to);
		ConvData data = mapConv.get(convType);
		if (null == data) {
			data = buildConvData(convType);
		}

		int pos = 0;
		int textLen = text.length();
		int maxLen = data.maxStrLen;
		HashMap<String, String> map = data.map;
		StringBuilder sb = new StringBuilder();

		while (pos < textLen) {
			int len = maxLen;
			if (textLen - pos < len) {
				len = textLen - pos;
			}

			boolean found = false;
			while (len > 0 && !found) {
				String subStr = text.substring(pos, pos + len);
				String convStr = map.get(subStr);
				if (null != convStr) {
					sb.append(convStr);
					pos += len;
					found = true;
				}
				len--;
			}

			if (!found) {
				sb.append(text.charAt(pos));
				pos++;
			}
		}

		return sb.toString();
	}

	public static void init(Context ctx) throws AppException {
		if (!initDone && null != ctx) {
			Common.context = ctx;
			String state = Environment.getExternalStorageState();

			if (!(Environment.MEDIA_MOUNTED.equals(state) || Environment.MEDIA_MOUNTED_READ_ONLY.equals(state))) {
				throw new AppException(ctx.getString(R.string.initSdcardNotMounted));
			}

			Common.packageName = ctx.getPackageName();
			// Android create the path if it does not exists
			Common.dataPath = ctx.getExternalFilesDir(null);

			if (null == Common.dataPath || !Common.dataPath.exists()) {
				throw new AppException(ctx.getString(R.string.initNotAvailable));
			}

			Common.mapConv = new HashMap<Common.ConvType, Common.ConvData>(ConvType.class.getEnumConstants().length);

			Common.initDone = true;
			CommonDB.initDatabases();
			Preferences.init(ctx);
		}
	}

	/**
	 * Return a localized string from the application's package's default string
	 * table.
	 * 
	 * @param id
	 *            The resource id for the string
	 */
	public static String getString(int id) {
		if (initDone) {
			return Common.context.getString(id);
		}
		return Common.EMPTY_STRING;
	}

	/**
	 * Return a localized formatted string from the application's package's
	 * default string table, substituting the format arguments as defined in
	 * java.util.Formatter and java.lang.String.format.
	 * 
	 * @param id
	 *            Resource id for the format string formatArgs The format
	 * @param formatArgs
	 *            that will be used for substitution.
	 */
	public static String getString(int id, Object... formatArgs) {
		if (initDone) {
			return Common.context.getString(id, formatArgs);
		}
		return Common.EMPTY_STRING;
	}

	/**
	 * 
	 * @param id
	 *            The resource id, as found in res/values/values.xml
	 * @return The string or null if the init is not done.
	 */
	public static String getResourcesString(int id) {
		if (initDone) {
			return Common.context.getResources().getString(id);
		}
		return null;
	}

	/**
	 * 
	 * @param id
	 *            The resource id, as found in res/values/values.xml
	 * @return the value of the resource
	 */
	public static int getResourcesInteger(int id) {
		if (initDone) {
			return Common.context.getResources().getInteger(id);
		}
		return 0;
	}

	/**
	 * @param id
	 *            The resource id, as found in res/values/values.xml
	 * @return the value of the resource, false if init is not done
	 */
	public static boolean getResourcesBoolean(int id) {
		if (initDone) {
			return Common.context.getResources().getBoolean(id);
		}
		return false;
	}

	/**
	 * @param id
	 *            The resource id, as found in res/values/values.xml
	 * @return the value of the resource, Float.NaN if init is not done
	 */
	public static float getResourcesFloat(int id) {
		if (initDone) {
			final String strVal = Common.context.getResources().getString(id);
			try {
				return Float.parseFloat(strVal);
			} catch (NumberFormatException nfe) {
				// Do nothing
			}
		}
		return Float.NaN;
	}

	/**
	 * 
	 * @return The application version code read in the manifest file.
	 */
	public static int getVersionCode() {
		int versionCode = 0;
		if (null != Common.context) {
			try {
				versionCode = Common.context.getPackageManager().getPackageInfo(Common.context.getPackageName(), 0).versionCode;
			} catch (NameNotFoundException nnfe) {
				Log.e(TAG, nnfe.getMessage());
			}
		}
		return versionCode;
	}

	public static Locale getLocale(String lang) {

		if (null == Common.mapLocales) {
			Locale[] localesAvail = Locale.getAvailableLocales();
			Common.mapLocales = new HashMap<String, Locale>(localesAvail.length);
			for (Locale l : localesAvail) {
				mapLocales.put(l.getISO3Language(), l);
			}
		}
		Locale locale = Common.mapLocales.get(lang);
		if (null == locale) {
			// Build a factice one if not available
			locale = new Locale(lang);
			Common.mapLocales.put(lang, locale);
			Log.w(TAG, "No phone locale for lang " + lang);
		}

		return locale;
	}

	/**
	 * nullify the hashmap used by convert.
	 */
	public static void clean() {

		if (null != Common.mapLocales) {
			Common.mapLocales.clear();
			Common.mapLocales = null;
		}

		CommonDB.closeDatabases();
		Kanji.clearCache();
		Preferences.clean();

		if (null != Common.mapConv) {
			Common.mapConv.clear();
			Common.mapConv = null;
		}
		Common.packageName = null;
		Common.dataPath = null;
		Common.initDone = false;
		
		System.runFinalization();
		System.gc();
	}

	public static boolean isInitDone() {
		return Common.initDone;
	}

	/**
	 * My split string function (String.toSplit() don't work like I want) May
	 * return an array of 0 length.
	 * 
	 * @param toSplit
	 *            The String to split
	 * @param splitChar
	 *            the splitter char
	 * @return a String array
	 */
	public static ArrayList<String> splitString(String toSplit, char splitChar) {
		if (null == toSplit) {
			return null;
		}

		ArrayList<String> list = new ArrayList<String>();
		int strLen = toSplit.length();
		int first = 0;
		for (int i = 0; i < strLen; i++) {
			char c = toSplit.charAt(i);
			if (c == splitChar) {
				if (i > first) {
					list.add(toSplit.substring(first, i));
					first = i + 1;
				} else {
					first = i + 1;
				}
			}
		}
		if (strLen > first) {
			list.add(toSplit.substring(first, strLen));
		}

		return list;
	}

	public static String listToString(ArrayList<String> list, char sep) {
		String str = null;
		if (null != list) {
			int size = list.size();
			StringBuilder sb = new StringBuilder(20);
			for (int i = 0; i < size; i++) {
				String elem = list.get(i);
				if (sb.length() > 0)
					sb.append(sep);
				sb.append(elem);

			}
			str = sb.toString();
		}
		return str;
	}

	public static String DEVICE_LANG = Locale.getDefault().getISO3Language();
	public static String ENGLISH_LANG = Locale.ENGLISH.getISO3Language();

	/**
	 * Add "lang" to "list". If "lang" == phone language, then put it at the
	 * first position in the list.<br/>
	 * If "lang" = english put if first or second depending on the current phone
	 * language.
	 * 
	 * @param lang
	 *            The lang to add
	 * @param list
	 *            The list to add the lang
	 */
	public static void addLangToList(String lang, ArrayList<String> list) {
		if (lang.equals(DEVICE_LANG)) {
			// put the phone language first, no matter what
			list.add(0, lang);
		} else if (lang.equals(ENGLISH_LANG)) {
			// for English language put it second, of the final array contains
			// the phone language or at the first position
			if (list.contains(DEVICE_LANG)) {
				list.add(1, lang);
			} else {
				list.add(0, lang);
			}
		} else {
			// other language, simply add it to the the list
			list.add(lang);
		}
	}

}
