/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.db;

import java.io.File;
import java.util.ArrayList;
import java.util.Locale;
import java.util.regex.Pattern;

import org.vono.narau.Common;

import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteException;
import android.util.Log;

/** SQLite db common */
public class CommonDB {
	private static final String TAG = CommonDB.class.getSimpleName();

	public static final String MAIN_LANG_NAME = Locale.JAPANESE.getDisplayLanguage();
	static final String MAIN_DB_LANG = Locale.JAPANESE.getISO3Language().intern();

	public enum Database {
		dictonary(1), kanji(2);

		private int value;

		private Database(int value) {
			this.value = value;
		}

		public int getValue() {
			return this.value;
		}

	};

	public enum DatabseState {
		bad, good
	}

	public static final String TABLE_DBINFO = "db_info";
	public static final String KEY_DB_VERSION = "version";
	public static final String KEY_DB_CREATION_TIME = "creation_time";
	public static final String KEY_DB_LANG = "lang";
	public static final String[] COLS_DBINFO = { "key", "value_str", "value_int" };

	public static final String DB_DICT_NAME_OLD = "JMdict.db";
	public static final String DB_KANJI_NAME_OLD = "kanjis.db";

	public static final String DICT_DB_FILENAME_TEMPLATE = "dict_%1$s-%2$d.db";
	public static final String KANJI_DB_FILENAME_TEMPLATE = "kanjis_%1$s-%2$d.db";
	private static final String DICT_DB_FILENAME_REGEX = "^dict_([a-z]{3})-([0-9]+)\\.db$";
	private static final String KANJI_DB_FILENAME_REGEX = "^kanjis_([a-z]{3})-([0-9]+)\\.db$";

	public static final Pattern PATTERN_DICT = Pattern.compile(DICT_DB_FILENAME_REGEX);
	public static final Pattern PATTERN_KANJI = Pattern.compile(KANJI_DB_FILENAME_REGEX);

	public static final ArrayList<File> oldDBFiles = new ArrayList<File>();
	public static final ArrayList<InfosDB> listDBInfos = new ArrayList<InfosDB>(12);

	public static boolean haveDB(int dbMask) {
		boolean ret = true;

		if ((dbMask & Database.dictonary.getValue()) != 0) {
			ret = ret && DictionaryDB.isReady();
		}
		if ((dbMask & Database.kanji.getValue()) != 0) {
			ret = ret && KanjiDB.isReady();
		}

		return ret;
	}

	public static boolean haveDB(Database database) {
		if (Database.dictonary == database) {
			return DictionaryDB.isReady();
		}
		return KanjiDB.isReady();
	}

	/**
	 * Get the list of existing database, check if everything is OK.
	 */
	public static void initDatabases() {
		final File dataDirectory = Common.dataPath;
		final File[] files = dataDirectory.listFiles();

		for (int i = 0; i < files.length; i++) {
			final File file = files[i];
			final String filename = file.getName();
			if (PATTERN_DICT.matcher(filename).matches()) {
				CommonDB.buildDBInfo(Database.dictonary, file);
			} else if (PATTERN_KANJI.matcher(filename).matches()) {
				CommonDB.buildDBInfo(Database.kanji, file);
			} else if (DB_DICT_NAME_OLD.equals(filename)) {
				oldDBFiles.add(file);
			} else if (DB_KANJI_NAME_OLD.equals(filename)) {
				oldDBFiles.add(file);
			}
			// else ignore file
		}

		DictionaryDB.initDB();
		KanjiDB.initDB();
	}

	public static void closeDatabases() {
		DictionaryDB.clean();
		KanjiDB.clean();

		for (InfosDB infosDB : listDBInfos) {
			if (null != infosDB.db) {
				infosDB.db.close();
			}
		}

		oldDBFiles.clear();
		listDBInfos.clear();
	}

	/**
	 * Build InfosDB, open database
	 * 
	 * @param type
	 *            Database type
	 * @param file
	 *            Database file
	 */
	private static void buildDBInfo(Database type, File file) {
		// Get db version and lang
		int dbVersion = -1;
		String lang = null;
		long creationTime = 0;

		SQLiteDatabase db = null;
		if (file.canRead()) {

			Cursor cursor = null;
			try {
				db = SQLiteDatabase.openDatabase(file.getAbsolutePath(), null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);
				cursor = db.query(CommonDB.TABLE_DBINFO, CommonDB.COLS_DBINFO, null, null, null, null, null);
				if (cursor.moveToFirst()) {
					int idxColKey = cursor.getColumnIndex(CommonDB.COLS_DBINFO[0]);
					int idxColValueStr = cursor.getColumnIndex(CommonDB.COLS_DBINFO[1]);
					int idxColValueInt = cursor.getColumnIndex(CommonDB.COLS_DBINFO[2]);
					do {
						String key = cursor.getString(idxColKey);
						if (CommonDB.KEY_DB_VERSION.equals(key)) {
							dbVersion = cursor.getInt(idxColValueInt);
						} else if (CommonDB.KEY_DB_LANG.equals(key)) {
							lang = cursor.getString(idxColValueStr);
						} else if (CommonDB.KEY_DB_CREATION_TIME.equals(key)) {
							creationTime = (long) cursor.getInt(idxColValueInt);
						}
					} while (cursor.moveToNext());
				}

			} catch (SQLiteException se) {
				Log.e(TAG, se.getMessage(), se);
			} finally {
				if (null != cursor) {
					cursor.close();
				}

			}
		}
		if (-1 == dbVersion || null == lang) {
			// Wrong database type, or something unknown
			// will ask the user to delete this file
			oldDBFiles.add(file);
			if (null != db) {
				db.close();
			}
			return;
		}

		int expectedDBVersion = -1;
		if (type == Database.dictonary) {
			expectedDBVersion = DictionaryDB.VALUE_DB_VERSION;
		} else {
			expectedDBVersion = KanjiDB.VALUE_DB_VERSION;
		}
		DatabseState state = DatabseState.bad;
		if (expectedDBVersion == dbVersion) {
			state = DatabseState.good;
		}

		if (state != DatabseState.good) {
			oldDBFiles.add(file);
			if (null != db) {
				db.close();
			}
			return;
		}

		InfosDB infosDB = getDBInfos(type, lang);
		if (null == infosDB) {
			infosDB = new InfosDB(type, lang, db);
			CommonDB.listDBInfos.add(infosDB);
		}
		infosDB.setCreationTime(creationTime);
	}

	public static String getDBFilename(Database database, String lang) {
		String template = DICT_DB_FILENAME_TEMPLATE;
		int version = 0;
		if (database == Database.kanji) {
			template = KANJI_DB_FILENAME_TEMPLATE;
			version = KanjiDB.VALUE_DB_VERSION;
		} else {
			version = DictionaryDB.VALUE_DB_VERSION;
		}

		return String.format(template, lang, version);

	}

	static InfosDB getDBInfos(Database type, String lang) {
		final ArrayList<InfosDB> listDB = CommonDB.listDBInfos;
		final int numDBInfos = listDB.size();
		InfosDB infosDB = null;
		for (int i = 0; i < numDBInfos; i++) {
			infosDB = listDB.get(i);
			if (infosDB.type == type && infosDB.lang.equals(lang)) {
				// OK we have it
				i = numDBInfos;
			} else {
				infosDB = null;
			}
		}

		return infosDB;
	}
}
