/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.db;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeSet;

import org.vono.narau.Common;
import org.vono.narau.db.CommonDB.Database;
import org.vono.narau.db.CommonDB.DatabseState;
import org.vono.narau.models.kanji.Kanji;
import org.vono.narau.preferences.Preferences;

import android.database.Cursor;
import android.database.SQLException;
import android.database.sqlite.SQLiteDatabase;
import android.util.Log;

/**
 * 
 * Read the kanji svg database.
 * 
 */
public class KanjiDB {
	private static final String TAG = KanjiDB.class.getSimpleName();

	public static final int VALUE_DB_VERSION = 4;

	// KanjiDict
	private static final String TABLE_DICT_KANJI = "kdict_kanji";
	private static final String TABLE_DICT_MEANINGS = "kdict_meanings";
	// private static final String TABLE_DICT_INDEX = "kdict_index";
	// private static final String TABLE_DICT_REF = "kdict_ref";
	// private static final String TABLE_DICT_LANGS = "kdict_langs";

	// KanjiVG
	private static final String TABLE_KVG_GROUP = "kvg_group";
	private static final String TABLE_KVG_PATH = "kvg_path";
	private static final String TABLE_KVG_RAD = "kvg_radicals";
	private static final String TABLE_KVG_POS = "kvg_positions";
	private static final String VIEW_KVG_SCHOOL = "kvg_school";

	public static final String WHERE_CODEPOINT = "codepoint=?";
	public static final String WHERE_CP_IS_AND_RAD_NOT_NULL = WHERE_CODEPOINT + " AND radical IS NOT NULL";

	private static final String COL_CODEPOINT = "codepoint";
	private static final String COL_KANJI_GRADE = "grade";
	private static final String COL_KANJI_FREQUENCY = "frequency";
	private static final String COL_KANJI_JLPT = "jlpt";
	private static final String COL_KANJI_STROKE_COUNT = "stroke_count";
	private static final String COL_KANJI_READING_ON = "reading_on";
	private static final String COL_KANJI_READING_KUN = "reading_kun";
	private static final String COL_KVG_GROUPS_ELEMENT = "element";
	private static final String COL_KVG_GROUPS_ORIGINAL = "original";

	private static final String[] COLS_KANJI_WITH_READING = { COL_CODEPOINT, COL_KANJI_GRADE, COL_KANJI_FREQUENCY,
			COL_KANJI_JLPT, COL_KANJI_STROKE_COUNT, COL_KANJI_READING_ON, COL_KANJI_READING_KUN };
	private static final String[] COLS_KANJI_NO_READING = { COL_CODEPOINT, COL_KANJI_GRADE, COL_KANJI_FREQUENCY,
			COL_KANJI_JLPT, COL_KANJI_STROKE_COUNT };
	private static final String[] COLS_KANJIVG_CODEPOINT = { COL_CODEPOINT };
	private static final String[] COLS_KVG_RADPOS = { "value" };
	private static final String[] COLS_DICT_LANG_MEANINGS = { "meanings" };
	private static final String[] COLS_KVG_GROUPS = { "id", "parent", "element", "position", "radical", "original" };
	private static final String[] COLS_KVG_GROUPS_RADS = { "element", "original" };
	private static final String[] COLS_KVG_PATH = { "parent", "type", "draw" };

	// private static final String[] COLS_DICT_LANGS = { "lang" };

	private static final String ORDER_BY_CODEPOINT = "codepoint asc";
	private static final String ORDER_BY_ID = "id asc";
	private static final String ORDER_BY_PARENT = "parent ASC";

	// Database management datas
	private static InfosDB mainDB = null;
	private static final ArrayList<InfosDB> langsDB = new ArrayList<InfosDB>();
	private static CommonDB.DatabseState state;
	private static String[] radicals = null;
	private static String[] positions = null;

	/**
	 * Init the KanjiDB.state value
	 */
	static void initDB() {
		KanjiDB.state = DatabseState.bad;

		boolean haveMain = false;
		boolean haveLang = false;
		final ArrayList<InfosDB> listDBs = CommonDB.listDBInfos;
		final int size = listDBs.size();
		for (int i = 0; i < size; i++) {
			final InfosDB dbInfos = listDBs.get(i);
			if (dbInfos.type == Database.kanji) {
				if (CommonDB.MAIN_DB_LANG.equals(dbInfos.lang)) {
					haveMain = true;
					KanjiDB.mainDB = dbInfos;
				} else {
					haveLang = true;
					KanjiDB.langsDB.add(dbInfos);
				}
			}
		}
		if (haveMain && haveLang) {
			KanjiDB.state = DatabseState.good;
			KanjiDB.initRadicalPositions();
		}
	}

	private static void initRadicalPositions() {
		// Fetch some datas

		Cursor cursor = null;
		try {
			final SQLiteDatabase db = KanjiDB.mainDB.db;

			cursor = db.query(TABLE_KVG_POS, COLS_KVG_RADPOS, null, null, null, null, ORDER_BY_ID);
			if (cursor.moveToFirst()) {
				final int colIdxVal = cursor.getColumnIndex(COLS_KVG_RADPOS[0]);
				final int count = cursor.getCount();
				KanjiDB.positions = new String[count];

				for (int i = 0; i < count; i++) {
					KanjiDB.positions[i] = cursor.getString(colIdxVal);
					cursor.moveToNext();
				}
			}
			cursor.close();

			cursor = db.query(TABLE_KVG_RAD, COLS_KVG_RADPOS, null, null, null, null, ORDER_BY_ID);
			if (cursor.moveToFirst()) {
				final int colIdxVal = cursor.getColumnIndex(COLS_KVG_RADPOS[0]);
				final int count = cursor.getCount();
				KanjiDB.radicals = new String[count];

				for (int i = 0; i < count; i++) {
					KanjiDB.radicals[i] = cursor.getString(colIdxVal);
					cursor.moveToNext();
				}
			}

		} catch (SQLException se) {
			Log.e(TAG, "initDB()", se);
			KanjiDB.state = DatabseState.bad;
		} finally {
			if (null != cursor) {
				cursor.close();
			}
		}
	}

	static void clean() {
		KanjiDB.state = DatabseState.bad;
		KanjiDB.radicals = null;
		KanjiDB.positions = null;
		KanjiDB.mainDB = null;
		KanjiDB.langsDB.clear();
	}

	public static boolean isReady() {
		return DatabseState.good == KanjiDB.state;
	}

	public static DatabseState getState() {
		return KanjiDB.state;
	}

	/**
	 * Can be called only by Preferences.getLangsPriorities()<br/>
	 * Use Preferences.getLangsPriorities() instead
	 * 
	 * @return a list of langs where there is meanings, or null if the db is not
	 *         Initialised.
	 */
	public static ArrayList<String> getLangs() {
		ArrayList<String> langs = null;
		if (DatabseState.good == KanjiDB.state) {
			langs = new ArrayList<String>(langsDB.size());
			for (InfosDB infosDB : langsDB) {
				Common.addLangToList(infosDB.lang, langs);
			}
		}

		return langs;
	}

	public static boolean getKanji(Kanji kanji) {
		boolean ret = true;
		final boolean needToLoad = !(kanji.kanjiDBLoadDictRef && kanji.kanjiDBLoadKanji && kanji.kanjiDBLoadMeaning && kanji.kanjiDBLoadSVG);

		if (needToLoad && (DatabseState.good == KanjiDB.state)) {
			Cursor cursor = null;
			try {
				final SQLiteDatabase db = mainDB.db;

				String[] selectionArgs = new String[1];
				selectionArgs[0] = Integer.toString(kanji.codepoint);
				cursor = db.query(TABLE_DICT_KANJI, COLS_KANJI_WITH_READING, WHERE_CODEPOINT, selectionArgs, null,
						null, null);
				if (cursor.moveToFirst()) {
					ArrayList<InfosDB> listLangDB = KanjiDB.getSortedListLangDB();
					ret = KanjiDB.cursorToKanji(cursor, kanji, listLangDB, true, true, true) != null;
				} else {
					ret = false;
				}
			} finally {
				if (null != cursor)
					cursor.close();
			}
		}

		return ret;
	}

	/**
	 * Load only the radicals from the SVG informations.
	 * 
	 * @param kanji
	 */
	public static TreeSet<Integer> getKanjiRadicals(int[] codepoints) {

		final TreeSet<Integer> setRadicals = new TreeSet<Integer>();

		if (DatabseState.good == KanjiDB.state) {
			Cursor cursor = null;

			try {
				final SQLiteDatabase db = mainDB.db;

				final String[] selectionArgs = new String[1];

				final int numKanjis = codepoints.length;
				for (int i = 0; i < numKanjis; i++) {
					final Kanji kanji = Kanji.getKanji(codepoints[i]);
					if (kanji.kanjiDBLoadRadicals) {
						// Radical already loaded
						setRadicals.addAll(kanji.radicals);
						continue;
					}

					selectionArgs[0] = Integer.toString(kanji.codepoint);
					cursor = db.query(TABLE_KVG_GROUP, COLS_KVG_GROUPS_RADS, WHERE_CP_IS_AND_RAD_NOT_NULL,
							selectionArgs, null, null, null);

					final TreeSet<Integer> radCodePoints = new TreeSet<Integer>();

					if (cursor.moveToFirst()) {
						final Integer colElemIdx = cursor.getColumnIndex(COL_KVG_GROUPS_ELEMENT);
						final Integer colOrigIdx = cursor.getColumnIndex(COL_KVG_GROUPS_ORIGINAL);

						do {

							Integer cpElem = cursor.getInt(colElemIdx);
							Integer cpOrig = cursor.getInt(colOrigIdx);
							if (null != cpElem && !radCodePoints.contains(cpElem)) {
								radCodePoints.add(cpElem);
							}
							if (null != cpOrig && !radCodePoints.contains(cpOrig)) {
								radCodePoints.add(cpOrig);
							}
						} while (cursor.moveToNext());

					}

					kanji.radicals = new ArrayList<Integer>(radCodePoints);
					kanji.kanjiDBLoadRadicals = true;
					setRadicals.addAll(radCodePoints);
					cursor.close();
					cursor = null;
				}

			} finally {
				if (null != cursor) {
					cursor.close();
				}
			}
		}

		return setRadicals;
	}

	/**
	 * 
	 * @param where
	 *            The query where clause.
	 * @return An iterator for the kanjis
	 */
	public static KanjiDBIterator getKanjiIterator(String selection, String[] selectionArgs, boolean loadMeaning,
			boolean loadSvg, boolean loadReading) {
		if (DatabseState.good != KanjiDB.state) {
			return null;
		}

		final SQLiteDatabase db = mainDB.db;

		Cursor cursor = null;

		if (loadReading) {
			cursor = db.query(TABLE_DICT_KANJI, COLS_KANJI_WITH_READING, selection, selectionArgs, null, null,
					ORDER_BY_CODEPOINT);
		} else {
			cursor = db.query(TABLE_DICT_KANJI, COLS_KANJI_NO_READING, selection, selectionArgs, null, null,
					ORDER_BY_CODEPOINT);
		}

		KanjiDB.KanjiDBIterator ite = new KanjiDB.KanjiDBIterator(cursor, loadMeaning, loadSvg, loadReading);
		return ite;
	}

	/**
	 * Return an iterator on kanjis codepoint where an SVG drawing is available.
	 * 
	 * @param selection
	 *            where clause
	 * @param selectionArgs
	 *            where values
	 * @return an KanjiSVGIterator or null if the database do not have the good
	 *         version.
	 */
	public static KanjiSVGIterator getKanjiSVGIterator(String selection, String[] selectionArgs) {
		if (DatabseState.good != KanjiDB.state) {
			return null;
		}

		final SQLiteDatabase db = mainDB.db;

		Cursor cursor = db.query(VIEW_KVG_SCHOOL, COLS_KANJIVG_CODEPOINT, selection, selectionArgs, null, null,
				ORDER_BY_CODEPOINT);

		KanjiDB.KanjiSVGIterator ite = new KanjiSVGIterator(cursor);

		return ite;
	}

	private static ArrayList<InfosDB> getSortedListLangDB() {
		final ArrayList<InfosDB> listLangDB = new ArrayList<InfosDB>();
		final ArrayList<String> lLangs = Preferences.getLangsPriorities(Database.kanji);

		// Quick & Dirty
		for (String lang : lLangs) {
			for (InfosDB dbInfos : KanjiDB.langsDB) {
				if (lang.equals(dbInfos.lang)) {
					listLangDB.add(dbInfos);
				}
			}
		}

		return listLangDB;
	}

	private static void getMeanings(Kanji kanji, ArrayList<InfosDB> listLangDB) {
		if (kanji.kanjiDBLoadMeaning) {
			return;
		}

		final int numLangs = listLangDB.size();
		final String[] selectionArgs = new String[1];
		selectionArgs[0] = Integer.toString(kanji.codepoint);

		if (null == kanji.meanings) {
			kanji.meanings = new HashMap<String, String>(1);
		}
		final HashMap<String, String> map = kanji.meanings;

		for (int i = 0; i < numLangs; i++) {
			final InfosDB dbInfos = listLangDB.get(i);

			final Cursor cursor = dbInfos.db.query(TABLE_DICT_MEANINGS, COLS_DICT_LANG_MEANINGS, WHERE_CODEPOINT,
					selectionArgs, null, null, null);

			if (cursor.moveToFirst()) {
				final int idxColMeaning = cursor.getColumnIndex(COLS_DICT_LANG_MEANINGS[0]);

				do {
					String meanings = cursor.getString(idxColMeaning);
					map.put(dbInfos.lang, meanings);

				} while (cursor.moveToNext());
				i = numLangs;
			}

			cursor.close();
		}
		kanji.kanjiDBLoadMeaning = true;
	}

	private static void getKanjiVG(Kanji kanji) {
		if (kanji.kanjiDBLoadSVG) {
			return;
		}
		final SQLiteDatabase db = mainDB.db;

		Kanji.Group[] groups = null;
		final String[] selectionArgs = new String[1];
		selectionArgs[0] = Integer.toString(kanji.codepoint);
		Cursor cursor = db.query(TABLE_KVG_GROUP, COLS_KVG_GROUPS, WHERE_CODEPOINT, selectionArgs, null, null,
				ORDER_BY_ID);

		if (cursor.moveToFirst()) {
			final int idxColId = cursor.getColumnIndex(COLS_KVG_GROUPS[0]);
			final int idxColParent = cursor.getColumnIndex(COLS_KVG_GROUPS[1]);
			final int idxColElement = cursor.getColumnIndex(COL_KVG_GROUPS_ELEMENT);
			final int idxColPosition = cursor.getColumnIndex(COLS_KVG_GROUPS[3]);
			final int idxColRadical = cursor.getColumnIndex(COLS_KVG_GROUPS[4]);
			final int idxColOriginal = cursor.getColumnIndex(COL_KVG_GROUPS_ORIGINAL);

			groups = new Kanji.Group[cursor.getCount()];

			final TreeSet<Integer> radCodePoints = new TreeSet<Integer>();
			do {
				final int id = cursor.getInt(idxColId);
				final int parent = cursor.getInt(idxColParent);

				Kanji.Group group = new Kanji.Group(id);
				group.element = cursor.getInt(idxColElement);
				group.position = cursor.isNull(idxColPosition) ? null
						: KanjiDB.positions[cursor.getInt(idxColPosition)];
				group.radical = cursor.isNull(idxColRadical) ? null : KanjiDB.radicals[cursor.getInt(idxColRadical)];
				group.original = cursor.getInt(idxColOriginal);

				if (null != group.element)
					radCodePoints.add(group.element);
				if (null != group.original)
					radCodePoints.add(group.original);

				if (-1 != parent) {
					// parent==-1 => Root element
					if (null == groups[parent].elements) {
						groups[parent].elements = new ArrayList<Kanji.Element>(1);
					}
					groups[parent].elements.add(group);
				}
				groups[id] = group;

			} while (cursor.moveToNext());

			kanji.group = groups[0];
			kanji.radicals = new ArrayList<Integer>(radCodePoints.size());
			kanji.radicals.addAll(radCodePoints);
		}

		cursor.close();

		// Path
		// codepoint INTEGER NOT NULL, parent INTEGER NOT NULL, type TEXT, draw
		// TEXT NOT NULL
		cursor = db.query(TABLE_KVG_PATH, COLS_KVG_PATH, WHERE_CODEPOINT, selectionArgs, null, null, ORDER_BY_PARENT);
		if (cursor.moveToFirst()) {
			int idxColParent = cursor.getColumnIndex(COLS_KVG_PATH[0]);
			int idxColType = cursor.getColumnIndex(COLS_KVG_PATH[1]);
			int idxColDraw = cursor.getColumnIndex(COLS_KVG_PATH[2]);

			if (null == kanji.group) {
				kanji.group = new Kanji.Group(0);
			}
			if (null == groups) {
				groups = new Kanji.Group[1];
				groups[0] = kanji.group;
			}
			final int numGroups = groups.length;
			int id = 1;

			do {
				int parent = cursor.getInt(idxColParent);
				String type = cursor.getString(idxColType);
				String draw = cursor.getString(idxColDraw);

				if (parent < numGroups) {
					Kanji.Path path = new Kanji.Path(id, type, draw);
					if (null == groups[parent].elements) {
						groups[parent].elements = new ArrayList<Kanji.Element>(1);
					}
					groups[parent].elements.add(path);
				} else {
					Log.e(TAG, "Error in parenting path for kanji " + kanji.codepoint + ", no parent group id "
							+ parent);
				}

			} while (cursor.moveToNext());
		}
		cursor.close();

		kanji.kanjiDBLoadSVG = true;
		kanji.kanjiDBLoadRadicals = true;
	}

	private static Kanji cursorToKanji(Cursor cursor, Kanji kanji, ArrayList<InfosDB> listLangDB, boolean loadMeaning,
			boolean loadSVG, boolean loadReading) { // boolean loadDictRef
		if (cursor.isClosed() || cursor.isAfterLast() || cursor.isBeforeFirst()) {
			return null;
		}

		Kanji lKanji = kanji;

		if (null == lKanji || !lKanji.kanjiDBLoadKanji) {
			int idxColCodepoint = cursor.getColumnIndex(COL_CODEPOINT);
			int idxColGrade = cursor.getColumnIndex(COL_KANJI_GRADE);
			int idxColFrequency = cursor.getColumnIndex(COL_KANJI_FREQUENCY);
			int idxColJLPT = cursor.getColumnIndex(COL_KANJI_JLPT);
			int idxColStrokeCount = cursor.getColumnIndex(COL_KANJI_STROKE_COUNT);
			int idxColReadingOn = cursor.getColumnIndex(COL_KANJI_READING_ON);
			int idxColReadingKun = cursor.getColumnIndex(COL_KANJI_READING_KUN);

			final int codepoint = cursor.getInt(idxColCodepoint);
			if (null == lKanji) {
				lKanji = Kanji.getKanji(codepoint);
			} else if (codepoint != lKanji.codepoint) {
				Log.w(TAG, "Codepoint mismatch cursor: " + codepoint + ", kanji: " + lKanji.codepoint);
				return null;
			}

			if (!cursor.isNull(idxColGrade))
				lKanji.grade = cursor.getInt(idxColGrade);

			if (!cursor.isNull(idxColFrequency))
				lKanji.frequency = cursor.getInt(idxColFrequency);

			if (!cursor.isNull(idxColJLPT))
				lKanji.jlpt = cursor.getInt(idxColJLPT);

			if (!cursor.isNull(idxColStrokeCount))
				lKanji.strokeCount = cursor.getInt(idxColStrokeCount);

			if (loadReading && idxColReadingKun != -1 && idxColReadingOn != -1) {
				lKanji.readingsOn = cursor.getString(idxColReadingOn);
				lKanji.readingsKun = cursor.getString(idxColReadingKun);
			}

			lKanji.kanjiDBLoadKanji = true;
		}

		if (loadMeaning && !lKanji.kanjiDBLoadMeaning) {
			KanjiDB.getMeanings(lKanji, listLangDB);
		}
		if (loadSVG && !lKanji.kanjiDBLoadSVG) {
			KanjiDB.getKanjiVG(lKanji);
		}

		// TODO LoadDictRef
		lKanji.kanjiDBLoadDictRef = true;

		return lKanji;
	}

	/**
	 * Iterator on a kanji db
	 * 
	 */
	public static class KanjiDBIterator implements Iterator<Kanji> {
		private final Cursor cursor;
		private int current;
		private int count;
		private boolean loadMeaning;
		private boolean loadSVG;
		private boolean loadReading;
		private final ArrayList<InfosDB> listLangDB;

		private KanjiDBIterator(Cursor cursor, boolean loadMeaning, boolean loadSVG, boolean loadReading) {
			this.cursor = cursor;
			this.current = 0;
			this.count = cursor.getCount();
			this.loadMeaning = loadMeaning;
			this.loadSVG = loadSVG;
			this.loadReading = loadReading;
			this.listLangDB = KanjiDB.getSortedListLangDB();
		}

		/**
		 * 
		 * @return The current kanji number or -1 if this iterator is closed or
		 *         0 the first() method was not called.
		 */
		public int getCurrent() {
			return this.current;
		}

		/**
		 * 
		 * @return The number of resuls in the query.
		 */
		public int getCount() {
			return this.count;
		}

		/**
		 * 
		 * @return the first kanji, or null if there is no kanji in the list.
		 */
		public Kanji first() {
			Kanji kanji = null;
			if (this.cursor.moveToFirst()) {
				kanji = KanjiDB.cursorToKanji(cursor, null, this.listLangDB, this.loadMeaning, this.loadSVG,
						this.loadReading);
				this.current = 0;
			}
			return kanji;
		}

		/**
		 * 
		 * @return The next kanji.
		 */
		public Kanji next() {
			Kanji kanji = null;
			if (!this.cursor.isClosed() && this.cursor.moveToNext()) {
				kanji = KanjiDB.cursorToKanji(cursor, null, this.listLangDB, this.loadMeaning, this.loadSVG,
						this.loadReading);
				this.current++;
			}
			return kanji;
		}

		/**
		 * 
		 * @return The previous kanji
		 */
		public Kanji previous() {
			Kanji kanji = null;
			if (this.cursor.moveToPrevious()) {
				kanji = KanjiDB.cursorToKanji(cursor, null, this.listLangDB, this.loadMeaning, this.loadSVG,
						this.loadReading);
				this.current--;
			}
			return kanji;
		}

		/**
		 * 
		 * @return The last kanji
		 */
		public Kanji last() {
			Kanji kanji = null;
			if (this.cursor.moveToLast()) {
				kanji = KanjiDB.cursorToKanji(cursor, null, this.listLangDB, this.loadMeaning, this.loadSVG,
						this.loadReading);
				this.current = this.count;
			}
			return kanji;
		}

		/**
		 * Move the the iterator to the given position
		 * 
		 * @param newPosition
		 *            The new position (>=0 && < count)
		 * @return The kanji, or null if the cursor is close or the new position
		 *         is out of range.
		 */
		public Kanji position(int newPosition) {
			Kanji kanji = null;
			if (!this.cursor.isClosed() && newPosition >= 0 && newPosition < this.count) {
				if (this.cursor.moveToPosition(newPosition)) {
					this.current = newPosition;
					kanji = KanjiDB.cursorToKanji(cursor, null, this.listLangDB, this.loadMeaning, this.loadSVG,
							this.loadReading);
				}
			}
			return kanji;
		}

		/**
		 * Get the specific kanji at index position
		 * 
		 * @param index
		 *            The kanji position
		 * @return the kanji at index position or null if index is invalid.
		 */
		public Kanji get(int index) {
			Kanji kanji = null;
			if (this.cursor.moveToPosition(index)) {
				kanji = KanjiDB.cursorToKanji(cursor, null, this.listLangDB, this.loadMeaning, this.loadSVG,
						this.loadReading);
				this.cursor.moveToPosition(this.current);
			}
			return kanji;
		}

		/**
		 * Must be call, to close the db cursor and the database.
		 */
		public void close() {
			this.cursor.close();
			this.current = -1;
			this.count = 0;
		}

		public boolean hasNext() {
			return !this.cursor.isClosed() && !this.cursor.isLast();
		}

		public boolean hasPrevious() {
			return !this.cursor.isClosed() && !this.cursor.isFirst();
		}

		/**
		 * throw UnsupportedOperationException.
		 */
		public void remove() {
			throw new UnsupportedOperationException("Can't remove a Kanji from the database");

		}
	}

	/**
	 * Iterator on codepoint which have an SVG path available.
	 * 
	 */
	public static class KanjiSVGIterator implements Iterator<Integer> {
		private int codepoint;

		private final Cursor cursor;
		private final int count;
		private int current;

		private KanjiSVGIterator(Cursor cursor) {
			this.cursor = cursor;
			this.count = cursor.getCount();
			this.codepoint = -1;
			this.current = 0;
		}

		/**
		 * 
		 * @return The number of resuls in the query.
		 */
		public int getCount() {
			return this.count;
		}

		/**
		 * 
		 * @return The current kanji number or -1 if this iterator is closed or
		 *         0 the first() method was not called.
		 */
		public int getCurrent() {
			return this.current;
		}

		public boolean hasNext() {
			return !this.cursor.isClosed() && !this.cursor.isLast();
		}

		/**
		 * Get the next kanji codepoint.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer next() {
			this.codepoint = -1;
			if (!this.cursor.isAfterLast() && !this.cursor.isClosed()) {
				if (cursor.moveToNext()) {
					this.codepoint = cursor.getInt(0);
					this.current++;
				}
			}
			return this.codepoint;
		}

		/**
		 * Get the previous kanji codepoint.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer previous() {
			this.codepoint = -1;
			if (!this.cursor.isBeforeFirst() && !this.cursor.isClosed()) {
				if (cursor.moveToPrevious()) {
					this.codepoint = cursor.getInt(0);
					this.current--;
				}
			}
			return this.codepoint;
		}

		/**
		 * Get the first kanji.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer first() {
			this.codepoint = -1;
			if (!this.cursor.isClosed()) {
				if (cursor.moveToFirst()) {
					this.codepoint = cursor.getInt(0);
					this.current = 0;
				}
			}
			return this.codepoint;
		}

		/**
		 * Get the last kanji.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer last() {
			this.codepoint = -1;
			if (!this.cursor.isClosed()) {
				if (cursor.moveToLast()) {
					this.codepoint = cursor.getInt(0);
					this.current = this.count;
				}
			}
			return this.codepoint;
		}

		/**
		 * Move the the iterator to the given position
		 * 
		 * @param newPosition
		 *            The new position (>=0 && < count)
		 * @return The kanji codepoint, or -1 if the cursor is close or the new
		 *         position is out of range.
		 */
		public Integer position(int newPosition) {
			this.codepoint = -1;
			if (!this.cursor.isClosed() && newPosition >= 0 && newPosition < this.count) {
				if (this.cursor.moveToPosition(newPosition)) {
					this.current = newPosition;
					this.codepoint = cursor.getInt(0);
				}
			}
			return this.codepoint;
		}

		/**
		 * Must be call, to close the db cursor and the database.
		 */
		public void close() {
			this.cursor.close();
			this.current = -1;
			this.codepoint = -1;
		}

		/**
		 * UnsupportedOperationException.
		 */
		public void remove() {
			throw new UnsupportedOperationException("Can't remove a Kanji from the database");

		}

	}
}
