/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.download;

import java.io.IOException;
import java.io.InputStream;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Locale;

import org.vono.narau.AppException;
import org.vono.narau.Common;
import org.vono.narau.Narau;
import org.vono.narau.QuitThread;
import org.vono.narau.R;
import org.vono.narau.db.CommonDB;
import org.vono.narau.db.InfosDB;
import org.vono.narau.db.InfosDB.Action;

import android.app.NotificationManager;
import android.app.ProgressDialog;
import android.content.Intent;
import android.content.res.ColorStateList;
import android.os.Bundle;
import android.os.Environment;
import android.os.StatFs;
import android.support.v4.app.FragmentActivity;
import android.text.Html;
import android.text.Spanned;
import android.text.method.LinkMovementMethod;
import android.util.Log;
import android.view.View;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.LinearLayout;
import android.widget.ProgressBar;
import android.widget.TextView;

/**
 * Manage the download database process (first run & update).
 * 
 */
public class DownloaderActivity extends FragmentActivity {
	private static final String TAG = DownloaderActivity.class.getSimpleName();

	public static final String UPDATE_PROCESS = "upd";
	public static final String KEY_LOG_MSGS = "log";
	public static final String KEY_CURRENT_DB = "db";

	private enum CurrentScreen {
		NONE, FIRST_RUN, QUESTION, DOWNLOADING, DOWNLOAD_DONE, INIT_ERROR;
	}

	public enum LauncherSource {
		NONE, FIRST_RUN, UPDATE, NOTIFICATION
	}

	/**
	 * The current download screen.
	 */
	private static CurrentScreen currentScreen = CurrentScreen.NONE;
	private LauncherSource launcherSource;

	private ProgressDialog progressDialog;
	private TextView textViewDownloadSpeed;
	private TextView textViewLog;
	private TextView textViewDatabase;
	private ProgressBar progressBar;
	private ProgressBar progressBarTotal;

	public DownloaderActivity() {
		this.progressBar = null;
		this.progressBarTotal = null;
		this.progressDialog = null;
		this.textViewDatabase = null;
		this.textViewDownloadSpeed = null;
		this.textViewLog = null;
		this.launcherSource = LauncherSource.NONE;
	}

	/** Called when the activity is created. */
	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		if (!Common.isInitDone()) {
			try {
				Common.init(super.getApplicationContext());
			} catch (AppException ae) {
				super.setContentView(R.layout.menu_init_error);
				TextView tv = (TextView) super.findViewById(R.id.initErrorTextView);
				tv.setText(ae.getMessage());
				DownloaderActivity.currentScreen = CurrentScreen.INIT_ERROR;
			}
		}

		this.launcherSource = (LauncherSource) super.getIntent().getSerializableExtra(UPDATE_PROCESS);

		if (this.launcherSource != LauncherSource.NOTIFICATION) {
			NotificationManager nm = (NotificationManager) getSystemService(NOTIFICATION_SERVICE);
			nm.cancel(R.id.download_db_notification);
		} else {
			Narau.finishOnUpdateDB();
		}

		switch (DownloaderActivity.currentScreen) {
		case NONE:
			if (Environment.MEDIA_MOUNTED_READ_ONLY.equals(Environment.getExternalStorageState())) {
				// Error we can't write a file in a read only mounted
				// storage.
				super.setContentView(R.layout.menu_init_error);
				TextView tv = (TextView) super.findViewById(R.id.initErrorTextView);
				tv.setText(R.string.initSdcardReadOnly);
			} else if (this.launcherSource != LauncherSource.FIRST_RUN) {
				this.firstRunContinue(null);
			} else {
				this.showFirstRunScreen();
			}

			if (!CommonDB.oldDBFiles.isEmpty()) {
				DialogDeleteOldDB dialog = new DialogDeleteOldDB();
				dialog.show(super.getSupportFragmentManager(), "DeleteOldDB");
			}
			break;
		case FIRST_RUN:
			this.showFirstRunScreen();
			break;
		case QUESTION:
			this.showDownloadViewQuestion();
			break;
		case DOWNLOADING:
		case DOWNLOAD_DONE:
			this.showDownloadView();
			break;
		case INIT_ERROR:
			break;
		}
	}

	@Override
	protected void onDestroy() {
		super.onDestroy();
		DownloaderGzipDB.updateUIActivity(null);
	}

	@Override
	public void onBackPressed() {
		// Don't finish this activity, instead show the device home screen.
		final Intent setIntent = new Intent(Intent.ACTION_MAIN);
		setIntent.addCategory(Intent.CATEGORY_HOME);
		setIntent.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
		super.startActivity(setIntent);
	}

	private void showFirstRunScreen() {
		super.setContentView(R.layout.download_firstrun);

		// Read the welcome HTML file
		final InputStream inputStream = getResources().openRawResource(R.raw.welcome);
		final byte[] buffer = new byte[512];
		final StringBuilder sbWelcome = new StringBuilder(512);
		try {
			boolean eof = false;
			do {
				final int count = inputStream.read(buffer);
				if (-1 == count) {
					eof = true;
				} else {
					final String str = new String(buffer, 0, count, "UTF-8");
					sbWelcome.append(str);
				}
			} while (!eof);
		} catch (IOException ioe) {
			Log.e(TAG, ioe.getMessage(), ioe);
		}

		try {
			inputStream.close();
		} catch (IOException ioe) {
			Log.e(TAG, ioe.getMessage(), ioe);
		}

		final Spanned html = Html.fromHtml(sbWelcome.toString());
		TextView tv = (TextView) super.findViewById(R.id.downloadFirstRunTextViewWelcome);
		tv.setMovementMethod(LinkMovementMethod.getInstance());
		tv.setText(html);

		DownloaderActivity.currentScreen = CurrentScreen.FIRST_RUN;
	}

	private void showDownloadViewQuestion() {
		super.setContentView(R.layout.download_question);
		DownloaderActivity.currentScreen = CurrentScreen.QUESTION;

		final LinearLayout linearLayout = (LinearLayout) super.findViewById(R.id.downloadQuestionLayoutCheckboxes);

		final ArrayList<InfosDB> listDBInfos = CommonDB.listDBInfos;
		final int numDownloads = listDBInfos.size();
		InfosDB currentMainDB = null;
		for (int i = 0; i < numDownloads; i++) {
			final InfosDB dbInfo = listDBInfos.get(i);

			View partView = null;
			if (dbInfo.isMain()) {
				currentMainDB = dbInfo;

				partView = super.getLayoutInflater().inflate(R.layout.download_question_cb_main, null);

				final CheckBox cb = (CheckBox) partView.findViewById(R.id.downloadQuestionPartCheckBox);

				final String cbText = String.format(super.getString(R.string.downloadQuestionSizeMain),
						dbInfo.userNameDBType, this.bytesToHumanSize(dbInfo.gzSize),
						this.bytesToHumanSize(dbInfo.dbSize));
				cb.setText(cbText);
				cb.setTag(dbInfo);
				dbInfo.checkBox = cb;

				final TextView tvDate = (TextView) partView.findViewById(R.id.downloadQuestionPartTextViewInfosDate);
				DateFormat df = DateFormat.getDateTimeInstance();
				tvDate.setText(super.getString(R.string.downloadQuestionUpdateDate,
						df.format(dbInfo.getDownloadCreationTime())));
				final TextView tvUpdate = (TextView) partView.findViewById(R.id.downloadQuestionPartTextViewInfosOther);

				if (dbInfo.exists()) {
					tvUpdate.setVisibility(View.VISIBLE);

					if (dbInfo.haveLatest()) {
						dbInfo.action = Action.none;
						tvUpdate.setText(R.string.downloadQuestionHaveLatest);
						tvUpdate.setTextColor(super.getResources().getColorStateList(R.drawable.text_color_orange));
						cb.setChecked(false);
					} else {
						dbInfo.action = Action.download;
						cb.setChecked(true);
						tvUpdate.setText(R.string.downloadQuestionUpdateAvail);
						tvUpdate.setTextColor(super.getResources().getColorStateList(R.drawable.text_color_green));
					}
				} else {
					dbInfo.action = Action.download;
					cb.setChecked(true);
					tvUpdate.setVisibility(View.GONE);
				}

			} else {
				partView = super.getLayoutInflater().inflate(R.layout.download_question_cb_sub, null);

				final CheckBox cb = (CheckBox) partView.findViewById(R.id.downloadQuestionPartCheckBox);
				final String cbText = String.format(super.getString(R.string.downloadQuestionSizeSub),
						dbInfo.userNameDBLang, this.bytesToHumanSize(dbInfo.gzSize),
						this.bytesToHumanSize(dbInfo.dbSize));
				cb.setText(cbText);
				dbInfo.checkBox = cb;

				cb.setTag(dbInfo);

				if (currentMainDB.haveLatest()) {
					cb.setChecked(false);
					dbInfo.action = Action.none;
				} else if (currentMainDB.exists()) {
					if (dbInfo.exists()) {
						cb.setChecked(true);
						dbInfo.action = Action.download;
					} else {
						cb.setChecked(false);
						dbInfo.action = Action.none;
					}
				} else {
					if (dbInfo.haveLatest()) {
						cb.setChecked(false);
						dbInfo.action = Action.none;
					} else {
						cb.setChecked(true);
						dbInfo.action = Action.download;
					}
				}

			}

			partView.setId(i);
			linearLayout.addView(partView);
		}

		this.onClickCheckBoxDB(null);
	}

	/**
	 * Show the download view
	 */
	private void showDownloadView() {
		super.setContentView(R.layout.download);
		this.textViewDownloadSpeed = (TextView) super.findViewById(R.id.downloadTextViewSpeed);
		this.textViewLog = (TextView) super.findViewById(R.id.downloadTextViewLogs);
		this.textViewDatabase = (TextView) super.findViewById(R.id.downloadTextViewDatabase);
		this.progressBar = (ProgressBar) super.findViewById(R.id.downloadProgressBarDatabase);
		this.progressBarTotal = (ProgressBar) super.findViewById(R.id.downloadProgressBarTotal);

		DownloaderGzipDB.updateUIActivity(this);
	}

	/**
	 * menu_download_question.xml<br/>
	 * Click on OK button in the download now view.
	 * 
	 * @param v
	 *            the clicked button
	 */
	public void onClickButtonOk(View v) {
		DownloaderActivity.currentScreen = CurrentScreen.DOWNLOADING;

		this.showDownloadView();
		DownloaderGzipDB.download(this);
	}

	/**
	 * Click on the continue button in the first run layout.
	 * 
	 * @param v
	 *            the button
	 */
	public void firstRunContinue(View v) {
		super.setContentView(R.layout.menu_init_error);
		final String waitMessage = super.getString(R.string.downloadQuestionFetchURL);
		this.progressDialog = ProgressDialog.show(this, Common.EMPTY_STRING, waitMessage, true);
		DownloaderDBInfos.startDownload(this);
	}

	/**
	 * 
	 * @param v
	 *            The continue button on the download.xml
	 */
	public void showMainMenu(View v) {
		DownloaderActivity.currentScreen = CurrentScreen.NONE;
		DownloaderGzipDB.clean();
		Intent intent = new Intent(super.getApplicationContext(), Narau.class);
		super.startActivity(intent);
		super.finish();
	}

	public void dbInfoJobDone(boolean retValue, String message) {
		this.progressDialog.dismiss();
		this.progressDialog = null;
		if (retValue) {
			Collections.sort(CommonDB.listDBInfos);
			this.showDownloadViewQuestion();
		} else {
			TextView tv = (TextView) super.findViewById(R.id.initErrorTextView);
			tv.setText(message);
		}
	}

	/**
	 * menu_download_question.xml<br/>
	 * Update the total download size, when click on a checkbox.
	 * 
	 * @param v
	 *            The checkbox clicked
	 */
	public void onClickCheckBoxDB(View v) {
		// This method is complicated, I think, there is many cases to
		// handles...

		// There is 3 states for each files:
		// - (exists() == false && haveLatest() == false) => Absent
		// - (exists() == true && haveLatest() == false) => Update available
		// - (exists() == true && haveLatest() == true) => Up to date
		//
		// Two actions are possibles:
		// - If a database is absent or an update is available, and the checkbox
		// is checked, it means that it will be downloaded.
		// - If a database is up to date, and the checkbox is checked, it means
		// that it will be deleted.
		//
		// Constraints:
		// - To be valid, a main DB must have at least one sub DB at the same
		// version (creation date).
		// - If a main DB is up to date and is selected, sub DB must be deleted
		// with the main DB
		// - If a database can be downloaded, and the user uncheck all sub DB,
		// the main DB is unselected, and all sub DB checkboxes are disabled.
		// - If the main DB is selected, select the first sub DB.
		//
		// Text message & validation button:
		// - If a database can be (downloaded or deleted) and there is enough
		// space on the SD
		// card (used size - delete size), the OK button is enabled.
		// - If there is not enough space on the SD card (used size - delete
		// size), a "downloadQuestionWarnLowSpace" message is display.
		// - If a database can be downloaded, the text message
		// "downloadQuestionTotal" is display.
		// - If a database can be deleted, the text message
		// "downloadQuestionDelSize" is display.

		final ArrayList<InfosDB> listDBInfos = CommonDB.listDBInfos;
		final int numDownloads = listDBInfos.size();

		final ColorStateList orange = super.getResources().getColorStateList(R.drawable.text_color_orange);
		final ColorStateList green = super.getResources().getColorStateList(R.drawable.text_color_green);
		final String textDelete = super.getString(R.string.downloadQuestionDelete);
		final String textHaveLatest = super.getString(R.string.downloadQuestionHaveLatest);
		final String textUpdateAvail = super.getString(R.string.downloadQuestionUpdateAvail);

		if (null != v) {
			// Current action
			final InfosDB dbInfo = (InfosDB) v.getTag();
			final CheckBox checkBox = (CheckBox) v;

			if (dbInfo.haveLatest() && checkBox.isChecked()) {
				dbInfo.action = Action.delete;
			} else if (checkBox.isChecked()) {
				dbInfo.action = Action.download;
			} else {
				dbInfo.action = Action.none;
			}
		}

		// Check coherence
		for (int i = 0; i < numDownloads; i++) {
			final InfosDB dbInfo = listDBInfos.get(i);
			if (dbInfo.isMain()) {

				if (dbInfo.haveLatest()) {
					// Up to date: delete?
					View view = (View) dbInfo.checkBox.getParent();
					TextView tvInfos = (TextView) view.findViewById(R.id.downloadQuestionPartTextViewInfosOther);
					tvInfos.setVisibility(View.VISIBLE);

					final View tvDate = view.findViewById(R.id.downloadQuestionPartTextViewInfosDate);
					final View tvTrans = view.findViewById(R.id.downloadQuestionPartTextViewTranslation);

					if (dbInfo.checkBox.isChecked()) {
						dbInfo.action = Action.delete;

						tvInfos.setText(textDelete);
						tvInfos.setTextColor(orange);
						tvInfos.setEnabled(false);
						tvDate.setEnabled(false);
						tvTrans.setEnabled(false);
					} else {
						dbInfo.action = Action.none;
						tvInfos.setText(textHaveLatest);
						tvInfos.setTextColor(green);
						tvInfos.setEnabled(true);
						tvDate.setEnabled(true);
						tvTrans.setEnabled(true);
					}

					int numSubAvail = 0;
					InfosDB firstSub = null;

					for (int j = i + 1; j < numDownloads; j++) {
						final InfosDB subDB = listDBInfos.get(j);
						if (subDB.type == dbInfo.type) {

							if (null == firstSub) {
								firstSub = subDB;
							}

							view = (View) subDB.checkBox.getParent();
							tvInfos = (TextView) view.findViewById(R.id.downloadQuestionPartTextViewInfosOther);

							if (Action.delete == dbInfo.action) {
								// Disabled all sub, mark sub as to be deleted
								subDB.checkBox.setEnabled(false);
								if (subDB.exists()) {
									subDB.action = Action.delete;
									tvInfos.setVisibility(View.VISIBLE);
									tvInfos.setText(textDelete);
									tvInfos.setTextColor(orange);
									subDB.checkBox.setChecked(true);
								} else {
									subDB.action = Action.none;
									tvInfos.setVisibility(View.GONE);
									subDB.checkBox.setChecked(false);
								}
							} else { // Action.none == dbInfo.action
								subDB.checkBox.setEnabled(true);
								// if the current action is on the main DB
								// uncheck all
								if (v == dbInfo.checkBox) {
									subDB.checkBox.setChecked(false);
									subDB.action = Action.none;
									tvInfos.setVisibility(View.GONE);
								}

								boolean haveLatest = subDB.haveLatest();
								boolean checked = subDB.checkBox.isChecked();
								if ((haveLatest && !checked) || (!haveLatest && checked)) {
									numSubAvail++;
								}
							}

						} else {
							// Another db type, end here
							j = numDownloads;
						}
					}

					if (numSubAvail == 0 && dbInfo.action == Action.none) {
						if (v != dbInfo.checkBox) {
							// Mark the main db as to be deleted
							// Rerun all
							dbInfo.checkBox.setChecked(true);
							dbInfo.action = Action.delete;
						} else if (firstSub.haveLatest()) {
							firstSub.checkBox.setChecked(false);
							firstSub.action = Action.none;
						} else {
							firstSub.checkBox.setChecked(true);
							firstSub.action = Action.download;
						}

						this.onClickCheckBoxDB(null);
						return;
					}

				} else {
					// Absent or Update available: download
					View view = (View) dbInfo.checkBox.getParent();
					TextView textView = (TextView) view.findViewById(R.id.downloadQuestionPartTextViewInfosOther);
					textView.setTextColor(green);
					textView.setText(textUpdateAvail);

					if (dbInfo.checkBox.isChecked()) {
						dbInfo.action = Action.download;

						int numSubAvail = 0;
						InfosDB firstSub = null;

						for (int j = i + 1; j < numDownloads; j++) {
							final InfosDB subDB = listDBInfos.get(j);
							if (subDB.type == dbInfo.type) {
								subDB.checkBox.setEnabled(true);

								if (null == firstSub) {
									firstSub = subDB;
								}

								if (subDB.checkBox.isChecked() && !subDB.haveLatest()) {
									numSubAvail++;
								} else if (!subDB.checkBox.isChecked() && subDB.haveLatest()) {
									numSubAvail++;
								}
							} else {
								// Another db type, end here
								j = numDownloads;
							}
						}

						if (0 == numSubAvail) {
							if (firstSub.haveLatest()) {
								firstSub.checkBox.setChecked(false);
								firstSub.action = Action.none;
							} else {
								firstSub.checkBox.setChecked(true);
								firstSub.action = Action.download;
							}
						}

					} else {
						if (dbInfo.exists()) {
							dbInfo.action = Action.delete;
							textView.setText(textDelete);
							textView.setTextColor(orange);
						} else {
							dbInfo.action = Action.none;
							textView.setText(textUpdateAvail);
							textView.setTextColor(green);
						}

						// Uncheck & disabled all sub db:
						for (int j = i + 1; j < numDownloads; j++) {
							final InfosDB subDB = listDBInfos.get(j);
							if (subDB.type == dbInfo.type) {
								subDB.checkBox.setEnabled(false);
								view = (View) subDB.checkBox.getParent();
								textView = (TextView) view.findViewById(R.id.downloadQuestionPartTextViewInfosOther);
								if (subDB.exists()) {
									// Normally this case should not append,
									// unless the user have manually change
									// something in the DB.
									subDB.action = Action.delete;
									subDB.checkBox.setChecked(true);
									textView.setVisibility(View.VISIBLE);
									textView.setText(textDelete);
									textView.setTextColor(orange);
									textView.setEnabled(false);
								} else {
									subDB.checkBox.setChecked(false);
									subDB.action = Action.none;
									textView.setVisibility(View.GONE);
								}
							} else {
								// Another db type, end here
								j = numDownloads;
							}
						}
					}
				}
			} else {
				// A sub database
				View view = (View) dbInfo.checkBox.getParent();
				TextView textView = (TextView) view.findViewById(R.id.downloadQuestionPartTextViewInfosOther);

				if (dbInfo.checkBox.isEnabled()) {
					textView.setEnabled(true);

					if (dbInfo.haveLatest()) {
						if (dbInfo.checkBox.isChecked()) {
							dbInfo.action = Action.delete;
							textView.setVisibility(View.VISIBLE);
							textView.setText(textDelete);
							textView.setTextColor(orange);
						} else {
							dbInfo.action = Action.none;
							textView.setVisibility(View.VISIBLE);
							textView.setText(textHaveLatest);
							textView.setTextColor(green);
						}
					} else if (dbInfo.exists()) {
						if (dbInfo.checkBox.isChecked()) {
							dbInfo.action = Action.download;
							textView.setVisibility(View.VISIBLE);
							textView.setText(textUpdateAvail);
							textView.setTextColor(green);
						} else {
							dbInfo.action = Action.delete;
							textView.setVisibility(View.VISIBLE);
							textView.setText(textDelete);
							textView.setTextColor(orange);
						}
					} else {
						textView.setVisibility(View.GONE);
						if (dbInfo.checkBox.isChecked()) {
							dbInfo.action = Action.download;
						} else {
							dbInfo.action = Action.none;
						}
					}
				} else {
					textView.setEnabled(false);
				}
			}
		} // for (int i = 0; i < numDownloads; i++)

		// Global message & OK Button
		long dwnlSize = 0;
		long needSize = 0;

		for (int i = 0; i < numDownloads; i++) {
			final InfosDB dbInfo = listDBInfos.get(i);

			if (Action.download == dbInfo.action) {
				dwnlSize += dbInfo.gzSize;
				needSize += dbInfo.dbSize;

				if (dbInfo.exists()) {
					// Remove the actual file size, if any
					needSize -= dbInfo.file.length();
				}

			} else if (Action.delete == dbInfo.action) {
				needSize -= dbInfo.dbSize;
			}
		}

		StringBuilder sb = new StringBuilder();
		if (dwnlSize > 0) {
			sb.append(String.format(super.getString(R.string.downloadQuestionDownloadTotalSize),
					this.bytesToHumanSize(dwnlSize)));
		}

		if (needSize > 0) {
			if (sb.length() > 0)
				sb.append('\n');
			sb.append(String.format(super.getString(R.string.downloadQuestionSDUsedSize),
					this.bytesToHumanSize(needSize)));
		} else if (needSize < 0) {
			if (sb.length() > 0)
				sb.append('\n');
			sb.append(String.format(super.getString(R.string.downloadQuestionSDFreeSize),
					this.bytesToHumanSize(needSize)));
		}

		TextView tvTotal = (TextView) super.findViewById(R.id.downloadQuestionTextViewTotal);
		if (sb.length() > 0) {
			tvTotal.setVisibility(View.VISIBLE);
			tvTotal.setText(sb);
		} else {
			tvTotal.setVisibility(View.GONE);
		}

		final StatFs stat = new StatFs(Environment.getExternalStorageDirectory().getPath());
		final long sdAvailSize = ((long) stat.getAvailableBlocks()) * ((long) stat.getBlockSize());

		TextView tvLow = (TextView) super.findViewById(R.id.downloadQuestionTextViewWarnLowSpace);
		if (sdAvailSize <= needSize) {
			String text = String.format(super.getString(R.string.downloadQuestionWarnLowSpace),
					this.bytesToHumanSize(sdAvailSize), this.bytesToHumanSize(needSize));
			tvLow.setText(text);
			tvLow.setVisibility(View.VISIBLE);
		} else {
			tvLow.setVisibility(View.GONE);
		}

		Button btOK = (Button) super.findViewById(R.id.downloadQuestionButtonOk);
		if (tvTotal.getVisibility() == View.VISIBLE && tvLow.getVisibility() == View.GONE) {
			btOK.setEnabled(true);
		} else {
			btOK.setEnabled(false);
		}
	}

	/**
	 * Handle a press on a quit/cancel button:<br/>
	 * <ul>
	 * <li>on the first run screen</li>
	 * <li>on the question screen</li>
	 * <li>on the download screen.</li>
	 * </ul>
	 * 
	 * @param v
	 *            a button
	 */
	public void quitApplication(View v) {

		if (null != this.progressDialog) {
			this.progressDialog.dismiss();
			this.progressDialog = null;
		}

		DownloaderActivity.currentScreen = CurrentScreen.NONE;

		DownloaderGzipDB.prematureThreadStop();
		super.finish();

		if (this.launcherSource == LauncherSource.UPDATE) {
			Common.clean();
			Intent intent = new Intent(super.getApplicationContext(), Narau.class);
			super.startActivity(intent);
		} else {
			QuitThread.startQuitThread(2000);
		}
	}

	/**
	 * Update the ui with the download informations.
	 * 
	 * @param progressTotal
	 *            if of the download
	 * @param progress
	 *            the progress in percent
	 * @param speed
	 *            the speed in bytes/seconds
	 */
	public void updateDownloadProgressAndSpeed(int progressTotal, int progress, long speed) {
		if (null != this.textViewDownloadSpeed && speed >= 0) {
			String text = super.getString(R.string.downloadSpeed, this.bytesToHumanSize(speed));
			this.textViewDownloadSpeed.setText(text);
		} else {
			this.textViewDownloadSpeed.setText(Common.EMPTY_STRING);
		}

		this.progressBar.setProgress(progress);
		this.progressBarTotal.setProgress(progressTotal);
	}

	public void updateDownloadLog(StringBuilder log) {
		this.textViewLog.setText(Html.fromHtml(log.toString()));
	}

	public void updateDownloadName(String value) {
		this.textViewDatabase.setText(value);
	}

	/**
	 * The downloader thread tell the ui that is job is done.
	 */
	public void downloadComplete() {
		Button button = (Button) super.findViewById(R.id.downloadButtonCancel);
		button.setVisibility(View.GONE);
		button = (Button) super.findViewById(R.id.downloadButtonContinue);
		button.setVisibility(View.VISIBLE);
		this.textViewDownloadSpeed.setText("");
		this.updateDownloadProgressAndSpeed(100, 100, -1);
		DownloaderActivity.currentScreen = CurrentScreen.DOWNLOAD_DONE;

		Common.clean();
	}

	private String bytesToHumanSize(long value) {
		final String[] sizes = super.getResources().getStringArray(R.array.byteSizeHumanDisplay);
		float humanSize = Math.abs(value);
		int count = 0;

		while (humanSize > 1024 && count < sizes.length) {
			humanSize = humanSize / 1024f;
			count++;
		}

		return String.format(Locale.US, "%1$.1f%2$s", humanSize, sizes[count]);
	}

	public static boolean isActive() {
		switch (DownloaderActivity.currentScreen) {
		case DOWNLOADING:
		case DOWNLOAD_DONE:
		case FIRST_RUN:
		case INIT_ERROR:
		case QUESTION:
			return true;
		default:
			break;
		}
		return false;
	}
}
