/**
 * Copyright (C) 2012 Yvon TANGUY
 * 
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.download;

import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;

import org.vono.narau.Common;
import org.vono.narau.R;
import org.vono.narau.db.InfosDB;

import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.util.Log;

/**
 * 
 * Download a small text file with the downloads informations.
 */
public class DownloaderDBInfos {
	private static final String TAG = DownloaderDBInfos.class.getSimpleName();

	private static final String KEY_MSG_DBINFO_MSG = "msg";
	private static final String KEY_MSG_DBINFO_DONE = "dbi";

	private static final String VER_CODE = "NarauVersionCode:";

	/**
	 * <p>
	 * Download the narau_db_infos.txt files on launchpad.net.<br/>
	 * See DATABASE_FILEINFO_URL.
	 * </p>
	 * Fill informations in DownloaderDBInfos.dbinfos
	 */
	private static class DBInfoDownloadThread extends Thread {

		@Override
		public void run() {
			boolean success = false;

			try {
				final URL url = new URL(Common.getResourcesString(R.string.db_infos_url));

				StringBuilder sb = DBInfoDownloadThread.downloadFile(url);
				int versionCode = Common.getVersionCode();
				final String searchCode = VER_CODE + versionCode;

				int startIdx = sb.indexOf(searchCode);
				if (startIdx == -1) {
					UIHandler.showMessage(Common.getString(R.string.downloadErrorNoDBInfo));
				} else {
					int endIdx = sb.indexOf(VER_CODE, startIdx + VER_CODE.length());
					if (endIdx == -1) {
						endIdx = sb.length();
					}
					success = DBInfoDownloadThread.updateDBInfos(sb.substring(startIdx + searchCode.length(), endIdx));
				}
			} catch (MalformedURLException mue) {
				String message = mue.getMessage();
				if (null == message) {
					message = "Empty URL";
				}
				UIHandler.showMessage(message);
				Log.e(TAG, message);
			}
			UIHandler.donwloadThreadEnd(success);
		}

		/**
		 * Perform the download
		 * 
		 * @param url
		 *            the url of the file to download.
		 * @return null if there is an error, or the file content.
		 */
		private static StringBuilder downloadFile(URL url) {
			HttpURLConnection connection = null;
			final StringBuilder sb = new StringBuilder();

			try {
				connection = (HttpURLConnection) url.openConnection();
				connection.connect();

				final int responseCode = connection.getResponseCode();
				if (responseCode / 100 != 2) {
					final String message = "" + responseCode + ": " + connection.getResponseMessage();
					UIHandler.showMessage(message);
					UIHandler.showMessage(Common.getString(R.string.downloadErrorGeneric,
							Common.getResourcesString(R.string.db_infos_url)));

					connection.disconnect();
					return null;
				}

				InputStream is = connection.getInputStream();
				int read = is.read();
				while (read != -1) {
					sb.append((char) read);
					read = is.read();

				}
				connection.disconnect();
			} catch (IOException ioe) {
				UIHandler.showMessage(ioe.getClass().getSimpleName() + ": " + ioe.getMessage());
				Log.e(TAG, ioe.getMessage(), ioe);
			} finally {
				if (null != connection) {
					connection.disconnect();
				}

			}
			return sb;
		}

		/**
		 * Update the dbInfos values with theses found in the downloaded file.
		 * 
		 * @param sb
		 *            The file content
		 * @return true if everything went OK.
		 */
		private static boolean updateDBInfos(String textToParse) {
			final String[] lines = textToParse.split("\n");
			final int numLines = lines.length;
			int countDB = 0;

			for (int i = 0; i < numLines; i++) {
				final String line = lines[i].trim();
				if (line.isEmpty() || line.charAt(0) == '#') {
					// ignore empty lines, and comments
					continue;
				}

				InfosDB infosDB = null;
				try {
					infosDB = InfosDB.getDBInfoFromLine(line);
				} catch (NumberFormatException nfe) {
					Log.e(TAG, nfe.getMessage(), nfe);
				}
				if (null == infosDB) {
					UIHandler.showMessage(Common.getString(R.string.downloadErrorLineFormat, line));
				} else {
					countDB++;
				}
			}

			// why 4 ?
			// at least 2 database per type (dict/kanji): one for jpn, and one
			// for a translation.
			if (countDB >= 4)
				return true;
			return false;
		}
	}

	private static class UIHandler extends Handler {
		private static UIHandler instance;
		private DownloaderActivity activity;

		private StringBuilder messages;

		public UIHandler(DownloaderActivity activity) {
			this.activity = activity;
			this.messages = new StringBuilder();

			DBInfoDownloadThread thread = new DBInfoDownloadThread();
			thread.start();
		}

		private static void donwloadThreadEnd(boolean returnValue) {
			final Message msg = Message.obtain(UIHandler.instance);
			final Bundle bundle = msg.getData();
			bundle.clear();
			bundle.putBoolean(KEY_MSG_DBINFO_DONE, returnValue);
			UIHandler.instance.sendMessage(msg);

		}

		private static void showMessage(String message) {
			final Message msg = Message.obtain(UIHandler.instance);
			final Bundle bundle = msg.getData();
			bundle.clear();
			bundle.putString(KEY_MSG_DBINFO_MSG, message);
			UIHandler.instance.sendMessage(msg);
		}

		@Override
		public void handleMessage(Message msg) {
			final Bundle bundle = msg.getData();
			if (bundle.containsKey(KEY_MSG_DBINFO_MSG)) {
				this.messages.append(bundle.getString(KEY_MSG_DBINFO_MSG));
				this.messages.append("\n");
			} else if (bundle.containsKey(KEY_MSG_DBINFO_DONE)) {
				boolean threadReturn = bundle.getBoolean(KEY_MSG_DBINFO_DONE, false);
				this.activity.dbInfoJobDone(threadReturn, this.messages.toString());
				UIHandler.instance = null;
			}
		}
	}

	public static void startDownload(DownloaderActivity activity) {
		if (null == UIHandler.instance) {
			UIHandler.instance = new UIHandler(activity);
		}
	}
}
