/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.models.dictionary;

import java.io.Serializable;
import java.util.ArrayList;

import org.vono.narau.R;
import org.vono.narau.dictionary.EntryActivity;
import org.vono.narau.preferences.Preferences;

import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.util.TypedValue;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnLongClickListener;
import android.widget.LinearLayout;
import android.widget.TextView;
import android.widget.Toast;

/**
 * Display dictionary entries for DictionaryActivity and KanjiToWords.
 * 
 */
public class ItemDisplay implements OnLongClickListener {

	private ArrayList<LinearLayout> innerLayouts;
	private Activity activity;
	private float zoomValue;
	private final int colorBlack;
	private final int colorDarkGrey;
	private ArrayList<Item> itemsList;

	public ItemDisplay(Activity parent) {
		this.activity = parent;
		this.innerLayouts = new ArrayList<LinearLayout>(100);
		this.itemsList = new ArrayList<ItemDisplay.Item>(100);
		this.zoomValue = Preferences.getFloat(R.string.prefNameTextZoom);
		this.colorBlack = activity.getResources().getColor(android.R.color.black);
		this.colorDarkGrey = activity.getResources().getColor(R.color.dark_grey);
	}

	public boolean onLongClick(View itemView) {
		Item item = (Item) itemView.getTag();
		Intent intent = new Intent(activity, EntryActivity.class);
		intent.putExtra(EntryActivity.TAG, item.sequence);
		activity.startActivity(intent);
		return true;
	}

	public void clearList() {
		this.itemsList.clear();
	}

	public void addItems(final LinearLayout layoutParent, final ArrayList<Item> items, final String toastTitle) {
		final int entriesCount = items.size();
		final ArrayList<LinearLayout> lInnerLayouts = this.innerLayouts;

		this.itemsList = items;

		if (lInnerLayouts.size() < entriesCount) {
			// Create missing inner layouts
			this.createMissingInnerLayouts(entriesCount);
		}

		for (int i = 0; i < entriesCount; i++) {
			ItemDisplay.addItemInternal(layoutParent, lInnerLayouts.get(i), items.get(i));
		}

		this.showToast(toastTitle);
	}

	public void addItem(final LinearLayout layoutParent, final Item item) {

		final int pos = this.itemsList.size();
		this.itemsList.add(item);
		final int count = this.itemsList.size();
		if (this.innerLayouts.size() < count) {
			this.createMissingInnerLayouts(count);
		}

		ItemDisplay.addItemInternal(layoutParent, this.innerLayouts.get(pos), item);
	}

	private static void addItemInternal(final LinearLayout layoutParent, final LinearLayout targetLayout,
			final Item item) {

		final TextView tvKanji = (TextView) targetLayout.findViewById(R.id.kanjiToWordsTextKanji);
		final TextView tvReading = (TextView) targetLayout.findViewById(R.id.kanjiToWordsTextReading);
		final TextView tvSense = (TextView) targetLayout.findViewById(R.id.kanjiToWordsTextSense);

		tvKanji.setText(item.kanjiElem);
		tvReading.setText(item.readingElem);
		tvSense.setText(item.sense);

		targetLayout.requestLayout();
		tvKanji.requestLayout();
		tvReading.requestLayout();
		tvSense.requestLayout();

		targetLayout.setTag(item);

		layoutParent.addView(targetLayout);
	}

	private void createMissingInnerLayouts(final int countTotal) {
		final float lZoomValue = this.zoomValue;
		final int cBlack = this.colorBlack;
		final int cGrey = this.colorDarkGrey;

		final LayoutInflater inflator = (LayoutInflater) activity.getSystemService(Context.LAYOUT_INFLATER_SERVICE);
		final ArrayList<LinearLayout> lInnerLayouts = this.innerLayouts;

		for (int i = lInnerLayouts.size(); i < countTotal; i++) {
			// Add inner linear layout
			final LinearLayout ill = (LinearLayout) inflator.inflate(R.layout.dictionary_item, null);

			final TextView tvKanji = (TextView) ill.findViewById(R.id.kanjiToWordsTextKanji);
			final TextView tvReading = (TextView) ill.findViewById(R.id.kanjiToWordsTextReading);
			final TextView tvSense = (TextView) ill.findViewById(R.id.kanjiToWordsTextSense);
			// final View vSpacer = ill.findViewById(R.id.kanjiToWordsSpacer);

			tvKanji.setTextSize(TypedValue.COMPLEX_UNIT_PX, lZoomValue);
			tvReading.setTextSize(TypedValue.COMPLEX_UNIT_PX, lZoomValue);
			tvSense.setTextSize(TypedValue.COMPLEX_UNIT_PX, lZoomValue);

			// Color alternance:
			int color = cGrey;
			if (i % 2 == 1) {
				color = cBlack;
			}

			ill.setBackgroundColor(color);
			ill.setOnLongClickListener(this);

			lInnerLayouts.add(ill);
		}
	}

	public void showToast(final String toastTitle) {
		if (null != toastTitle) {
			final String msg = String.format(activity.getString(R.string.toastNumWords), toastTitle,
					this.itemsList.size());
			Toast.makeText(activity, msg, Toast.LENGTH_LONG).show();
		}
	}

	public void setActivity(Activity activity, LinearLayout layoutParent) {
		this.activity = activity;
		final int count = this.itemsList.size();
		for (int i = 0; i < count; i++) {
			layoutParent.addView(this.innerLayouts.get(i));
		}
	}

	public void clear() {
		this.innerLayouts.clear();
		this.innerLayouts = null;
		this.itemsList.clear();
		this.itemsList = null;
	}

	public void zoomIn() {
		this.zoomValue *= 1.1f;
		this.changeZoomValue();
	}

	public void zoomOut() {
		this.zoomValue *= 0.9f;
		this.changeZoomValue();
	}

	private void changeZoomValue() {
		Preferences.setFloat(R.string.prefNameTextZoom, this.zoomValue);
		final float zoom = this.zoomValue;

		final ArrayList<LinearLayout> lInnerLayouts = this.innerLayouts;
		final int numLL = lInnerLayouts.size();

		for (int i = 0; i < numLL; i++) {
			final LinearLayout ill = lInnerLayouts.get(i);

			final TextView tvKanji = (TextView) ill.findViewById(R.id.kanjiToWordsTextKanji);
			final TextView tvReading = (TextView) ill.findViewById(R.id.kanjiToWordsTextReading);
			final TextView tvSense = (TextView) ill.findViewById(R.id.kanjiToWordsTextSense);

			tvKanji.setTextSize(TypedValue.COMPLEX_UNIT_PX, zoom);
			tvReading.setTextSize(TypedValue.COMPLEX_UNIT_PX, zoom);
			tvSense.setTextSize(TypedValue.COMPLEX_UNIT_PX, zoom);
		}
	}

	/**
	 * A single item (mini entry).
	 * 
	 */
	public static class Item implements Serializable {
		/**
		 * 
		 */
		private static final long serialVersionUID = -8748389774648139865L;

		public final String kanjiElem;
		public final String readingElem;
		public final String sense;
		public final int sequence;

		public Item(int id, String kanjiElem, String readingElem, String sense) {
			this.kanjiElem = kanjiElem;
			this.readingElem = readingElem;
			this.sense = sense;
			this.sequence = id;
		}

		// private void writeObject(java.io.ObjectOutputStream out) throws
		// IOException {
		// Log.d("Item", "writeObject");
		// }
		//
		// private void readObject(java.io.ObjectInputStream in) throws
		// IOException, ClassNotFoundException {
		// Log.d("Item", "readObject");
		// }

	}

}
