/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.models.kanji;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import android.os.Parcel;
import android.os.Parcelable;
import android.util.SparseArray;

/**
 * Informations on a kanji<br/>
 * Datas from kanjivg & kanjidict.
 */
public class Kanji implements Parcelable {
	private static final SparseArray<Kanji> kanjisArrayCache = new SparseArray<Kanji>(5000);

	/**
	 * Unicode value of this kanji.
	 */
	public final int codepoint;

	/**
	 * The main group.
	 */
	public Group group;

	/**
	 * G&lt;num&gt; -- the "grade" of the kanji. At most one per line.
	 * <ul>
	 * <li>G1 to G6 indicates the grade level as specified by the Japanese
	 * Ministry of Education for kanji that are to be taught in elementary
	 * school (1006 Kanji). These are sometimes called the "Kyouiku" (education)
	 * kanji and are part of the set of Jouyou (daily use) kanji;</li>
	 * <li>G8 indicates the remaining Jouyou kanji that are to be taught in
	 * secondary school (additional 1130 Kanji);</li>
	 * <li>G9 and G10 indicate Jinmeiyou ("for use in names") kanji which in
	 * addition to the Jouyou kanji are approved for use in family name
	 * registers and other official documents. G9 (774 kanji, of which 628 are
	 * in KANJIDIC) indicates the kanji is a "regular" name kanji, and G10 (209
	 * kanji of which 128 are in KANJIDIC) indicates the kanji is a variant of a
	 * Jouyou kanji;</li>
	 * </ul>
	 */
	public Integer grade;

	/**
	 * F&lt;num&gt; -- the frequency-of-use ranking. At most one per line. The
	 * 2,501 most-used characters have a ranking; those characters that lack
	 * this field are not ranked. The frequency is a number from 1 to 2,501 that
	 * expresses the relative frequency of occurrence of a character in modern
	 * Japanese. The data is based on an analysis of word frequencies in the
	 * Mainichi Shimbun over 4 years by Alexandre Girardi. From this the
	 * relative frequencies have been derived. Note:
	 * <ul>
	 * <li>these frequencies are biassed towards words and kanji used in
	 * newspaper articles,</li>
	 * <li>the relative frequencies for the last few hundred kanji so graded is
	 * quite imprecise.</li>
	 * </ul>
	 * 
	 * (Earlier editions of the KANJIDIC file used a frequency-of-use ranking
	 * from the National Language Research Institute (Tokyo), interpreted and
	 * adapted by Jack Halpern.)
	 */
	public Integer frequency;

	/**
	 * J&lt;num&gt; -- the level of the Japanese Language Proficiency Test
	 * (JLPT) in which the kanji occurs. (1-4)
	 */
	public Integer jlpt;

	/**
	 * S&lt;num&gt; -- the stroke count. At least one per line. If more than
	 * one, the first is considered the accepted count.
	 */
	public Integer strokeCount;

	/**
	 * ja_on - the "on" Japanese reading of the kanji, in katakana. Another
	 * attribute r_status, if present, will indicate with a value of "jy"
	 * whether the reading is approved for a "Jouyou kanji". A further attribute
	 * on_type, if present, will indicate with a value of kan, go, tou or
	 * kan'you the type of on-reading.
	 */
	public String readingsOn;

	/**
	 * ja_kun - the "kun" Japanese reading of the kanji, usually in hiragana. <br/>
	 * Where relevant the okurigana is also included separated by a ".".
	 * Readings associated with prefixes and suffixes are marked with a "-". A
	 * second attribute r_status, if present, will indicate with a value of "jy"
	 * whether the reading is approved for a "Jouyou kanji".
	 */
	public String readingsKun;

	/**
	 * Meaning map: lang for key, meanings '/' separated.
	 */
	public HashMap<String, String> meanings;

	/**
	 * This element contains the index numbers and similar unstructured
	 * information such as page numbers in a number of published dictionaries,
	 * and instructional books on kanji.
	 */
	public ArrayList<DictRef> dictRef;

	/**
	 * This kanji radicals
	 */
	public ArrayList<Integer> radicals;

	/*
	 * Flags for the KanjiDB, to know what was already loaded. 
	 */
	public boolean kanjiDBLoadKanji;
	public boolean kanjiDBLoadMeaning;
	public boolean kanjiDBLoadSVG;
	public boolean kanjiDBLoadDictRef;
	public boolean kanjiDBLoadRadicals;

	private Kanji(int codepoint) {
		this.codepoint = codepoint;
		this.group = null;
		this.frequency = null;
		this.grade = null;
		this.jlpt = null;
		this.strokeCount = null;
		this.readingsKun = null;
		this.readingsOn = null;
		this.meanings = null;
		this.dictRef = null;
		this.radicals = null;

		this.kanjiDBLoadDictRef = false;
		this.kanjiDBLoadKanji = false;
		this.kanjiDBLoadMeaning = false;
		this.kanjiDBLoadSVG = false;
		this.kanjiDBLoadRadicals = false;
	}

	/**
	 * Get the kanji for the specified codepoint.
	 * 
	 * @param codepoint
	 * @return
	 */
	public static Kanji getKanji(int codepoint) {
		Kanji kanji = Kanji.kanjisArrayCache.get(codepoint);
		if (null == kanji) {
			kanji = new Kanji(codepoint);
			Kanji.kanjisArrayCache.put(codepoint, kanji);
		}
		return kanji;
	}

	public static void clearCache() {
		Kanji.kanjisArrayCache.clear();
	}

	public void addReadingsOn(String text) {
		if (null == this.readingsOn) {
			this.readingsOn = text;
		} else {
			this.readingsOn += "/" + text;
		}
	}

	public void addReadingsKun(String text) {
		if (null == this.readingsKun) {
			this.readingsKun = text;
		} else {
			this.readingsKun += "/" + text;
		}
	}

	public int describeContents() {
		return 0;
	}

	public void writeToParcel(Parcel dest, int flags) {
		dest.writeInt(this.codepoint);
	}

	public static final Parcelable.Creator<Kanji> CREATOR = new Parcelable.Creator<Kanji>() {
		public Kanji createFromParcel(Parcel in) {
			final int codepoint = in.readInt();
			return Kanji.getKanji(codepoint);
		}

		public Kanji[] newArray(int size) {
			return new Kanji[size];
		}
	};

	/**
	 * Element of a kanji is a group or a path.
	 * 
	 */
	public static interface Element {
	}

	public static class Group implements Element {
		public final int id;
		public Integer element;
		public String position;
		public String radical;
		public Integer original;
		public String part;
		public String variant;
		public boolean partial;
		public boolean radicalForm;
		public int number;
		public boolean tradForm;
		/**
		 * TBD as of 20120505
		 */
		// public String phon;

		public List<Element> elements;

		public Group(int id) {
			this.id = id;
			this.element = null;
			this.position = null;
			this.radical = null;
			this.original = null;
			this.part = null;
			this.variant = null;
			this.partial = false;
			this.radicalForm = false;
			this.number = 0;
			this.tradForm = false;
			// this.phon = null;
		}

	}

	public static class Path implements Element {
		public final int id;
		public final String type;
		public final String draw;

		public Path(int id, String type, String draw) {
			this.type = type;
			this.draw = draw;
			this.id = id;
		}

	}

	public static class DictRef {
		public final String name;
		public final String ref;

		public DictRef(String name, String ref) {
			this.name = name;
			this.ref = ref;
		}
	}

}