/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.preferences;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.vono.narau.Common;
import org.vono.narau.R;
import org.vono.narau.db.CommonDB;
import org.vono.narau.db.CommonDB.Database;
import org.vono.narau.db.DictionaryDB;
import org.vono.narau.db.KanjiDB;

import android.content.Context;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.preference.PreferenceManager;
import android.util.Log;
import android.util.SparseArray;
import android.widget.TextView;

/**
 * Manage store preferences, and expose them to other classes.
 * 
 */
public class Preferences {
	public static final String TAG = Preferences.class.getSimpleName();
	private static final char LANG_SEP = ',';

	/**
	 * Pref database creation date
	 */
	@Deprecated
	public static final String PREF_DB_CD_PREFIX = "db.creation.";

	private static class PrefDefault {
		final int prefValueId;
		final String prefName;

		PrefDefault(int nameId, int valueId) {
			this.prefValueId = valueId;
			this.prefName = Common.getResourcesString(nameId);

			Preferences.defaultValues.put(nameId, this);
		}
	}

	private static SharedPreferences sharedPreferences = null;
	private static SparseArray<PrefDefault> defaultValues = null;
	private static final HashMap<Database, ArrayList<String>> langsPriorities;

	static {
		langsPriorities = new HashMap<Database, ArrayList<String>>();
	}

	/**
	 * DictionaryDB must have been init before.
	 * 
	 * @param context
	 */
	@SuppressWarnings("unused")
	public static void init(Context context) {
		sharedPreferences = PreferenceManager.getDefaultSharedPreferences(context);
		Preferences.migrPref();

		Preferences.defaultValues = new SparseArray<PrefDefault>(8);

		new PrefDefault(R.string.prefNameVersion, R.integer.prefValueVersion);
		new PrefDefault(R.string.prefNameTextZoom, R.string.prefValueTextZoom);
		new PrefDefault(R.string.prefNameDictionaryMaxResult, R.integer.prefValueDictionaryMaxResult);
		new PrefDefault(R.string.prefNameDictionaryWordMatch, R.bool.prefValueDictionaryWordMatch);
		new PrefDefault(R.string.prefNameKanjiShowGrid, R.bool.prefValueKanjiShowGrid);
		new PrefDefault(R.string.prefNameKanjiShowStrokesNumber, R.bool.prefValueKanjiShowStrokesNumber);
		new PrefDefault(R.string.prefNameKanjiStrokesWidth, R.string.prefValueKanjiStrokeDefaultWidth);
		new PrefDefault(R.string.prefNameKanjiDrawSpeed, R.string.prefValueKanjiStrokeDefaultDrawSpeed);
		new PrefDefault(R.string.prefNameKanjiToWordsNWTL, R.integer.prefValueKanjiToWordsNWTL);

		// Init the default text size
		// Not a very elegant method, but I don't know better
		float textZoom = Preferences.getFloat(R.string.prefNameTextZoom);
		if (0f == textZoom) {
			float textSize = (new TextView(context)).getTextSize();
			Preferences.setFloat(R.string.prefNameTextZoom, textSize);
		}
	}

	/**
	 * Used by Common, to free memory.
	 */
	public static void clean() {
		sharedPreferences = null;
		if (null != defaultValues) {
			defaultValues.clear();
			defaultValues = null;
		}
		langsPriorities.clear();
	}

	/**
	 * Clear all preferences
	 */
	public static void clear() {
		sharedPreferences.edit().clear().commit();
		langsPriorities.clear();
	}

	/**
	 * Update the preferences when there is some change (ex: key deleted, format
	 * change...).
	 */
	private static void migrPref() {
		final String prefName = Common.getResourcesString(R.string.prefNameVersion);
		final int prefValue = Common.getResourcesInteger(R.integer.prefValueVersion);
		int prefVersion = sharedPreferences.getInt(prefName, -1);

		if (prefValue == prefVersion) {
			// OK, good version
			return;
		}

		final Editor edit = sharedPreferences.edit();

		if (-1 == prefVersion) {
			// First time the application is launch
			edit.putInt(prefName, prefValue);
			edit.commit();
			return;
		}
		if (0 == prefVersion) {
			Log.i(TAG, "Preference migration from version 0 to 1");
			Map<String, ?> mapAll = sharedPreferences.getAll();
			Set<String> keys = mapAll.keySet();
			for (String key : keys) {
				if (key.startsWith("kanjiItem")) {
					edit.remove(key);
				}
			}
			prefVersion = 1;
		}
		if (1 == prefVersion) {
			Log.i(TAG, "Preference migration from version 1 to 2");
			Map<String, ?> mapAll = sharedPreferences.getAll();
			Set<String> keys = mapAll.keySet();
			for (String key : keys) {
				if (!key.contains(".")) {
					edit.remove(key);
				}
			}
			prefVersion = 2;
		}
		if (2 == prefVersion) {
			final String gzExt = ".gz";
			final long vD = sharedPreferences.getLong(PREF_DB_CD_PREFIX + CommonDB.DB_DICT_NAME_OLD + gzExt, 0);
			if (0 != vD) {
				edit.remove(PREF_DB_CD_PREFIX + CommonDB.DB_DICT_NAME_OLD + gzExt);
				edit.putLong(PREF_DB_CD_PREFIX + CommonDB.DB_DICT_NAME_OLD, vD);
			}
			final long vK = sharedPreferences.getLong(PREF_DB_CD_PREFIX + CommonDB.DB_KANJI_NAME_OLD + gzExt, 0);
			if (0 != vK) {
				edit.remove(PREF_DB_CD_PREFIX + CommonDB.DB_KANJI_NAME_OLD + gzExt);
				edit.putLong(PREF_DB_CD_PREFIX + CommonDB.DB_KANJI_NAME_OLD, vK);
			}

			if (0 != vD || 0 != vK) {
				edit.commit();
			}
			prefVersion = 3;
		}
		if (3 == prefVersion) {
			edit.remove(PREF_DB_CD_PREFIX + CommonDB.DB_DICT_NAME_OLD);
			edit.remove(PREF_DB_CD_PREFIX + CommonDB.DB_KANJI_NAME_OLD);

			prefVersion = 4;
		}

		edit.putInt(prefName, prefValue);
		edit.commit();
	}

	public static SharedPreferences getSP() {
		return Preferences.sharedPreferences;
	}

	public static String getString(int prefNameId) {
		final PrefDefault prefDefault = defaultValues.get(prefNameId);
		final String prefValue = Common.getResourcesString(prefDefault.prefValueId);

		return sharedPreferences.getString(prefDefault.prefName, prefValue);
	}

	public static int getInt(int prefNameId) {
		final PrefDefault prefDefault = defaultValues.get(prefNameId);
		final int prefValue = Common.getResourcesInteger(prefDefault.prefValueId);

		return sharedPreferences.getInt(prefDefault.prefName, prefValue);
	}

	public static boolean getBoolean(int prefNameId) {
		final PrefDefault prefDefault = defaultValues.get(prefNameId);
		final boolean prefValue = Common.getResourcesBoolean(prefDefault.prefValueId);

		return sharedPreferences.getBoolean(prefDefault.prefName, prefValue);
	}

	public static float getFloat(int prefNameId) {
		final PrefDefault prefDefault = defaultValues.get(prefNameId);
		final float prefValue = Common.getResourcesFloat(prefDefault.prefValueId);

		return sharedPreferences.getFloat(prefDefault.prefName, prefValue);
	}

	public static float getFloat(String prefName) {
		final int size = defaultValues.size();
		for (int i = 0; i < size; i++) {
			final PrefDefault prefDefault = defaultValues.valueAt(i);
			if (prefDefault.prefName.equals(prefName)) {
				final float prefValue = Common.getResourcesFloat(prefDefault.prefValueId);
				return sharedPreferences.getFloat(prefName, prefValue);
			}
		}
		return 0f;
	}

	public static void setBoolean(String name, boolean value) {
		sharedPreferences.edit().putBoolean(name, value).apply();
	}

	public static void setBoolean(int prefNameId, boolean value) {
		final String prefName = Common.getResourcesString(prefNameId);
		sharedPreferences.edit().putBoolean(prefName, value).apply();
	}

	public static void setFloat(int prefNameId, float value) {
		final String prefName = Common.getResourcesString(prefNameId);
		sharedPreferences.edit().putFloat(prefName, value).apply();
	}

	public static void setFloat(String prefName, float value) {
		sharedPreferences.edit().putFloat(prefName, value).apply();
	}

	public static void setLangsPriorities(Database database, ArrayList<String> values) {
		String strValue = Common.listToString(values, LANG_SEP);
		if (null != strValue) {
			String key = null;
			switch (database) {
			case dictonary:
				key = Common.getResourcesString(R.string.prefNameDictionaryLangPrio);
				break;
			case kanji:
				key = Common.getResourcesString(R.string.prefNameKanjiLangPrio);
				break;
			}

			langsPriorities.put(database, values);
			sharedPreferences.edit().putString(key, strValue).apply();
		}
	}

	public static ArrayList<String> getLangsPriorities(Database database) {
		ArrayList<String> langs = langsPriorities.get(database);
		if (null == langs) {

			String key = null;
			ArrayList<String> dbLangs = null;

			switch (database) {
			case dictonary:
				dbLangs = DictionaryDB.getLangs();
				key = Common.getResourcesString(R.string.prefNameDictionaryLangPrio);
				break;
			case kanji:
			default:
				dbLangs = KanjiDB.getLangs();
				key = Common.getResourcesString(R.string.prefNameKanjiLangPrio);
			}

			if (null != dbLangs) {
				// null if no DB of that type

				String prefValueLangs = sharedPreferences.getString(key, null);
				if (null != prefValueLangs) {
					ArrayList<String> prefLangs = Common.splitString(prefValueLangs, LANG_SEP);

					langs = new ArrayList<String>(dbLangs.size());
					// Add pref lang in right order
					// only if there is a DB with that lang
					for (String lang : prefLangs) {
						if (dbLangs.contains(lang)) {
							langs.add(lang);
						}
					}
					// Add others langs (new langs since last pref save)
					for (String lang : dbLangs) {
						if (!langs.contains(lang)) {
							langs.add(lang);
						}
					}

				} else {
					langs = dbLangs;
				}
			}

			if (null != langs) {
				langsPriorities.put(database, langs);
			}
		}

		return langs;
	}

}
