/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictionary;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;

import org.vono.narau.AppException;
import org.vono.narau.Common;
import org.vono.narau.Common.DatabseState;
import org.vono.narau.R;
import org.vono.narau.dictionary.jmdict.Entry;
import org.vono.narau.dictionary.jmdict.KanjiElement;
import org.vono.narau.dictionary.jmdict.ReadingElement;
import org.vono.narau.dictionary.jmdict.Sense;
import org.vono.narau.preferences.Preferences;

import android.database.Cursor;
import android.database.SQLException;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteException;
import android.util.Log;

/**
 * Function to manage the dictionaries.
 * 
 */
public class DictionaryDB {
	private static final String TAG = DictionaryDB.class.getSimpleName();
	private static final String DBL_QUOTES = "\"";
	private static final String PERCENT = "%";

	private static final String[][] TEXT_REPLACE = { { "'", "''" }, { ";", "" }, { "\"", "" }, { "/*", "" },
			{ "*/", "" }, { "--", "" } };

	public static final int VALUE_DB_VERSION = 1;

	private static final String TABLE_KANJI_ELEM = "k_ele";
	private static final String TABLE_READING_ELEM = "r_ele";
	private static final String TABLE_SENSE = "sense";
	private static final String TABLE_GLOSS = "gloss_%1$s";
	private static final String TABLE_INFO = "dict_info";

	private static final String COL_ID = "id";
	private static final String COL_KEB = "keb";
	private static final String COL_REB = "reb";
	private static final String COL_PRI = "pri";
	private static final String COL_INF = "inf";
	private static final String COL_NUM = "num";
	private static final String COL_STAGK = "stagk";
	private static final String COL_STAGR = "stagr";
	private static final String COL_POS = "pos";
	private static final String COL_XREF = "xref";
	private static final String COL_GLOSS_TEXT = "gloss";
	private static final String COL_INFO_LANG = "lang";
	private static final String COL_INFO_NBENTRIES = "nb_entries";

	private static final String[] ARRAY_COLS_ID = { COL_ID };

	/**
	 * COL_KEB, COL_PRI, COL_INF
	 */
	private static final String[] ARRAY_COLS_KELE = { COL_KEB, COL_PRI, COL_INF };

	/**
	 * COL_REB, COL_PRI, COL_INF
	 */
	private static final String[] ARRAY_COLS_RELE = { COL_REB, COL_PRI, COL_INF };

	/**
	 * COL_STAGK, COL_STAGR, COL_POS, COL_XREF
	 */
	private static final String[] ARRAY_COLS_SENSE = { COL_STAGK, COL_STAGR, COL_POS, COL_XREF, COL_NUM };

	/**
	 * COL_GLOSS_TEXT
	 */
	private static final String[] ARRAY_COLS_GLOSS = { COL_GLOSS_TEXT };

	private static final String QUERY_DEL_LANG = "DELETE FROM " + TABLE_INFO + " WHERE " + COL_INFO_LANG + "=\"%1$s\";";
	private static final String QUERY_DROP_TABLE = "DROP TABLE IF EXISTS " + TABLE_GLOSS + ";";
	private static final String QUERY_VACUUM = "VACUUM;";

	/**
	 * <ul>
	 * <li>1: Translation lang table (gloss_lang)</li>
	 * <li>2: Japanese table (k_ele or r_ele)</li>
	 * <li>3: Japanese text column</li>
	 * <li>4: Japanese text to search</li>
	 * <li>5: The max number of results (LIMIT clause)>/li>
	 * </ul>
	 */
	private static final String QUERY_SEARCH_JPN = "SELECT DISTINCT " + TABLE_GLOSS + "." + COL_ID + " FROM "
			+ TABLE_GLOSS + ",%2$s WHERE " + TABLE_GLOSS + "." + COL_ID + "=%2$s." + COL_ID
			+ " AND %2$s.%3$s LIKE '%%%4$s%%' LIMIT %5$d";

	/**
	 * A big where clause, to match only word (which can have a space or a
	 * slash, before and after.
	 */
	private static final String WHERE_LIKE_WORDMATCH = "%1$s LIKE '%2$s' " + "OR %1$s LIKE '%2$s %%' "
			+ "OR %1$s LIKE '%2$s/%%' " + "OR %1$s LIKE '%% %2$s' " + "OR %1$s LIKE '%%/%2$s' "
			+ "OR %1$s LIKE '%% %2$s %%' " + "OR %1$s LIKE '%%/%2$s/%%' " + "OR %1$s LIKE '%%/%2$s %%' "
			+ "OR %1$s LIKE '%% %2$s/%%'";
	private static final String WHERE_LIKE = "%1$s LIKE ?";
	private static final String WHERE_ID_EQUAL = COL_ID + "=%1$d";
	private static final String WHERE_ID_NUM_EQUAL = WHERE_ID_EQUAL + " AND " + COL_NUM + "=%2$d";

	private static final String LIMIT = "%1$d";
	// private static final String ORDERBY = "LENGTH (%1$s)";

	private static String dbFullName = null;
	private static Common.DatabseState state;

	public static Common.DatabseState initDB() {
		DictionaryDB.dbFullName = Common.dataPath + File.separator + Common.DB_DICT_NAME;
		DictionaryDB.state = DatabseState.missing;

		// check db version
		File dbFile = new File(DictionaryDB.dbFullName);
		if (dbFile.canRead()) {
			SQLiteDatabase db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
					| SQLiteDatabase.NO_LOCALIZED_COLLATORS);
			Cursor cursor = null;
			try {
				cursor = db.query(Common.TABLE_DBINFO, Common.COLS_DBINFO, null, null, null, null, null);
				if (cursor.moveToFirst()) {
					int idxColKey = cursor.getColumnIndex(Common.COLS_DBINFO[0]);
					int idxColValueInt = cursor.getColumnIndex(Common.COLS_DBINFO[2]);
					do {
						String key = cursor.getString(idxColKey);
						if (Common.KEY_DB_VERSION.equals(key)) {
							int version = cursor.getInt(idxColValueInt);
							if (VALUE_DB_VERSION == version) {
								DictionaryDB.state = DatabseState.ok;
							} else {
								DictionaryDB.state = DatabseState.deprecated;
							}
						}
					} while (cursor.moveToNext());
				}

			} catch (SQLiteException se) {
				Log.e(TAG, se.getMessage(), se);
			} finally {
				if (null != cursor) {
					cursor.close();
				}

				db.close();
			}
		}

		return DictionaryDB.state;
	}

	public static void clean() {
		DictionaryDB.state = DatabseState.missing;
		DictionaryDB.dbFullName = null;
	}

	public static boolean isReady() {
		return DictionaryDB.state == DatabseState.ok;
	}

	public static Common.DatabseState getState() {
		return DictionaryDB.state;
	}

	/**
	 * 
	 * @return All available translation languages in the dictionary database.
	 */
	public static ArrayList<String> getLangs() {

		if (DatabseState.ok != DictionaryDB.state) {
			return null;
		}
		SQLiteDatabase db = null;
		ArrayList<String> langs = null;
		synchronized (TAG) {
			try {
				db = SQLiteDatabase.openDatabase(DictionaryDB.dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				Cursor cursor = db.query(TABLE_INFO, new String[] { COL_INFO_LANG }, null, null, null, null, null);

				if (cursor.moveToFirst()) {
					langs = new ArrayList<String>(cursor.getCount());
					do {
						Common.addLangToList(cursor.getString(0), langs);
					} while (cursor.moveToNext());
				}

				cursor.close();

			} catch (SQLiteException sle) {
				langs = null;
				Log.e(TAG, sle.getMessage(), sle);
			} finally {
				if (null != db) {
					db.close();
				}
			}
		}
		return langs;
	}

	/**
	 * Delete a dictionary language
	 * 
	 * @param lang
	 *            the ISO name of the lang to delete (rus, fra, ger,...)
	 */
	public static boolean delLang(String lang) {
		SQLiteDatabase db = null;
		boolean delete = false;

		synchronized (TAG) {

			try {
				db = SQLiteDatabase.openDatabase(DictionaryDB.dbFullName, null, SQLiteDatabase.OPEN_READWRITE
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				String drop = String.format(QUERY_DROP_TABLE, lang);
				db.beginTransaction();
				db.execSQL(String.format(QUERY_DEL_LANG, lang));
				db.setTransactionSuccessful();

				db.execSQL(drop);
				db.execSQL(QUERY_VACUUM);

			} catch (SQLiteException sle) {
				delete = false;
				Log.e(TAG, sle.getMessage(), sle);
			} catch (SQLException se) {
				delete = false;
				Log.e(TAG, se.getMessage(), se);
			} finally {
				if (null != db) {
					if (db.inTransaction()) {
						db.endTransaction();
					}
					db.close();
				}
			}

		}

		return delete;
	}

	/**
	 * 
	 * @param lang
	 *            The ISO language name (fra...)
	 * @return The number of entries in the dictionary for this language, 0 if
	 *         the dictionary is not found.
	 */
	public static int getNumEntries(String lang) {
		SQLiteDatabase db = null;
		int numEntries = 0;

		synchronized (TAG) {
			try {
				db = SQLiteDatabase.openDatabase(DictionaryDB.dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				Cursor cursor = db.query(TABLE_INFO, new String[] { COL_INFO_NBENTRIES }, COL_INFO_LANG + "=\"" + lang
						+ DBL_QUOTES, null, null, null, null);

				if (cursor.moveToFirst()) {
					numEntries = cursor.getInt(0);
				}
				cursor.close();
			} catch (SQLiteException sle) {
				Log.e(TAG, sle.getMessage(), sle);
			} finally {
				if (null != db) {
					db.close();
				}
			}
		}
		return numEntries;
	}

	/**
	 * Prevent something like an SQL injection.
	 * 
	 * @param text
	 *            The text to put in the LIKE clause
	 * @return the secured text
	 */
	private static String secureTextMatch(String text) {
		int repArraySize = TEXT_REPLACE.length;
		int txtLen = text.length();
		StringBuilder result = new StringBuilder(txtLen);

		for (int i = 0; i < txtLen; i++) {
			int j = 0;
			for (; j < repArraySize; j++) {
				String txtSearch = TEXT_REPLACE[j][0];
				int txtSearchLen = txtSearch.length();
				// regionMatches -> startWith
				if (text.regionMatches(i, txtSearch, 0, txtSearchLen)) {
					result.append(TEXT_REPLACE[j][1]);
					i += txtSearchLen - 1;
					break;
				}
			}
			if (j == repArraySize) {
				// no replacement found, add the original text character
				result.append(text.charAt(i));
			}

		}
		String secText = result.toString();

		return secText;
	}

	/**
	 * 
	 * @param lang
	 *            The lang to search for
	 * @param the
	 *            text to search
	 * @return a list of entries identifiers, or null if no result
	 */
	public static ArrayList<Integer> findMatchGloss(String lang, String text) {
		String where = null;
		String[] selectionArg = null;
		// String orderby = String.format(DictionaryDB.ORDERBY, COL_GLOSS_TEXT);
		String orderby = null;
		final String tableGloss = String.format(TABLE_GLOSS, lang);
		// boolean wordMatch =
		// PreferencesDB.getParameterBoolean(PreferencesDB.PARAM_WORD_MATCH);
		boolean wordMatch = Preferences.getBoolean(R.string.prefNameDictionaryWordMatch);

		String _text = text;
		if (wordMatch) {
			// get the complicated like clause
			where = String.format(WHERE_LIKE_WORDMATCH, COL_GLOSS_TEXT, secureTextMatch(_text));
		} else {
			where = String.format(WHERE_LIKE, COL_GLOSS_TEXT);
			_text = PERCENT + text + PERCENT;
			selectionArg = new String[] { _text };
		}

		return DictionaryDB.findMatch(_text, tableGloss, where, selectionArg, orderby, null);
	}

	/**
	 * Search in the kanji table for text, where it exists translation for lang
	 * 
	 * @param lang
	 *            The translation lang to match
	 * @param text
	 *            Text to search
	 * @return
	 */
	public static ArrayList<Integer> findMatchKanji(String lang, String text) {

		String toFind = secureTextMatch(text);
		// int numResults =
		// PreferencesDB.getParameterInt(PreferencesDB.PARAM_NUM_RESULT);
		int maxResults = Preferences.getInt(R.string.prefNameDictionaryMaxResult);
		String query = String.format(Locale.US, QUERY_SEARCH_JPN, lang, TABLE_KANJI_ELEM, COL_KEB, toFind, maxResults);

		return DictionaryDB.findMatch(null, null, null, null, null, query);
	}

	public static ArrayList<Integer> findMatchKana(String lang, String text) {

		String toFind = secureTextMatch(text);
		// int numResults =
		// PreferencesDB.getParameterInt(PreferencesDB.PARAM_NUM_RESULT);
		int maxResults = Preferences.getInt(R.string.prefNameDictionaryMaxResult);
		String query = String
				.format(Locale.US, QUERY_SEARCH_JPN, lang, TABLE_READING_ELEM, COL_REB, toFind, maxResults);

		return DictionaryDB.findMatch(null, null, null, null, null, query);
	}

	/**
	 * Generic id matcher
	 * 
	 * @param text
	 *            Text to search
	 * @param table
	 *            Table to search
	 * @param where
	 *            where clause
	 * @return List of ids, or null if not found
	 */
	private static ArrayList<Integer> findMatch(String text, String table, String where, String[] selectionArg,
			String orderBy, String fullSQL) {
		ArrayList<Integer> arrayIds = null;
		SQLiteDatabase db = null;

		synchronized (TAG) {
			try {
				db = SQLiteDatabase.openDatabase(DictionaryDB.dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				Cursor cursor = null;
				if (null == fullSQL) {
					int maxResults = Preferences.getInt(R.string.prefNameDictionaryMaxResult);
					String limit = String.format(Locale.US, LIMIT, maxResults);
					cursor = db.query(false, table, ARRAY_COLS_ID, where, selectionArg, null, null, orderBy, limit);
				} else {
					cursor = db.rawQuery(fullSQL, null);
				}

				if (cursor.moveToFirst()) {
					arrayIds = new ArrayList<Integer>();
					do {
						arrayIds.add(cursor.getInt(0));
					} while (cursor.moveToNext());
				}
				cursor.close();

			} catch (SQLiteException sle) {
				Log.e(TAG, sle.getMessage(), sle);
			} finally {
				if (null != db) {
					db.close();
				}
			}
		}

		return arrayIds;
	}

	public static Entry fetchdEntry(int id) {
		Entry entry = null;
		SQLiteDatabase db = null;

		synchronized (TAG) {
			try {
				db = SQLiteDatabase.openDatabase(DictionaryDB.dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);
				entry = DictionaryDB.fetchdEntry(id, db);

			} catch (SQLiteException sle) {
				Log.e(TAG, sle.getMessage(), sle);
			} catch (AppException ae) {
				Log.e(TAG, ae.getMessage(), ae);
			} finally {
				if (null != db) {
					db.close();
				}
			}
		}

		return entry;
	}

	private static Entry fetchdEntry(int id, SQLiteDatabase db) throws AppException {
		String where = String.format(Locale.US, WHERE_ID_EQUAL, id);

		// get kanji element(s)
		ArrayList<KanjiElement> k_ele = fetchKanjiElements(db, where);

		// get reading element(s)
		ArrayList<ReadingElement> r_ele = fetchReadingElements(db, id, where);

		// get sense element(s)
		ArrayList<Sense> senses = fetchSenseElements(db, id, where);

		// Build entry:
		final Entry entry = new Entry(id);
		entry.k_ele = k_ele;
		entry.r_ele = r_ele;
		entry.senses = senses;

		return entry;
	}

	private static ArrayList<KanjiElement> fetchKanjiElements(SQLiteDatabase db, String where) {
		ArrayList<KanjiElement> k_ele = null;
		Cursor cursor = db.query(TABLE_KANJI_ELEM, ARRAY_COLS_KELE, where, null, null, null, COL_NUM);
		if (cursor.moveToFirst()) {
			k_ele = new ArrayList<KanjiElement>(cursor.getCount());
			do {
				KanjiElement ke = new KanjiElement();
				ke.keb = cursor.getString(0);

				String tmp = cursor.getString(1);
				if (null != tmp) {
					ke.ke_pri = Common.splitString(tmp, Common.SLASH);
				}

				tmp = cursor.getString(2);
				if (null != tmp) {
					ke.ke_inf = Common.splitString(tmp, Common.SLASH);
				}

				k_ele.add(ke);

			} while (cursor.moveToNext());
		}
		cursor.close();

		return k_ele;
	}

	private static ArrayList<ReadingElement> fetchReadingElements(SQLiteDatabase db, int id, String where)
			throws AppException {
		ArrayList<ReadingElement> r_ele = null;
		Cursor cursor = db.query(TABLE_READING_ELEM, ARRAY_COLS_RELE, where, null, null, null, COL_NUM);
		if (cursor.moveToFirst()) {
			r_ele = new ArrayList<ReadingElement>(cursor.getCount());

			do {
				ReadingElement re = new ReadingElement();
				re.reb = cursor.getString(0);

				String tmp = cursor.getString(1);
				if (null != tmp) {
					re.re_pri = Common.splitString(tmp, Common.SLASH);
				}

				tmp = cursor.getString(2);
				if (null != tmp) {
					re.re_inf = Common.splitString(tmp, Common.SLASH);
				}
				r_ele.add(re);

			} while (cursor.moveToNext());

		} else {
			// One reading element is mandatory
			// should not append
			cursor.close();
			throw new AppException("No reading element for id=" + id);
		}
		cursor.close();

		return r_ele;
	}

	private static ArrayList<Sense> fetchSenseElements(SQLiteDatabase db, int id, String where) {
		ArrayList<Sense> senses = null;
		Cursor cursor = db.query(TABLE_SENSE, ARRAY_COLS_SENSE, where, null, null, null, COL_NUM);
		if (cursor.moveToFirst()) {
			senses = new ArrayList<Sense>(cursor.getCount());

			// TODO: Prefetch the langs before getting all entries?
			ArrayList<String> langs = Preferences.getLangsPriorities(Common.Database.dictonary);

			do {
				Sense sense = new Sense();
				sense.stagk = cursor.getString(0);
				sense.stagr = cursor.getString(1);
				sense.pos = cursor.getString(2);
				sense.xref = cursor.getString(3);
				int num = cursor.getInt(4);

				sense.glosses = new HashMap<String, String>(langs.size());
				String whereIdNum = String.format(Locale.US, WHERE_ID_NUM_EQUAL, id, num);
				for (int i = 0; i < langs.size(); i++) {
					String lang = langs.get(i);
					String gloss = fetchGloss(db, lang, whereIdNum);
					if (null != gloss) {
						sense.glosses.put(lang, gloss);
					}
				}

				senses.add(sense);
			} while (cursor.moveToNext());
		}
		cursor.close();

		return senses;
	}

	private static String fetchGloss(SQLiteDatabase db, String lang, String where) {
		String gloss = null;
		String table = String.format(TABLE_GLOSS, lang);

		Cursor cursor = db.query(table, ARRAY_COLS_GLOSS, where, null, null, null, null);
		if (cursor.moveToFirst()) {
			gloss = cursor.getString(0);
		}
		cursor.close();

		return gloss;
	}

	/**
	 * An iterator for the KanjiToWords activity.
	 * 
	 */
	public static class KanjiToWordsIterator implements Iterator<Entry> {
		private static final String[] ARRAY_COLS_KANJI = { "id", "keb", "pri" };
		private static final String[] ARRAY_COLS_READING = { "reb" };
		// private static final String[] ARRAY_COLS_GLOSS = { "gloss" };
		private static final String SELECTION_KANJI = "keb like ? and num=?";
		private static final String SELECTION_READING = "id=?";
		private static final String SELECTION_GLOSS = "id=? and num=?";
		private static final String ORDER_READING = "num asc";
		private static final String ORDER_GLOSS = "num asc";

		private SQLiteDatabase db;
		private Cursor cursor;
		private final ArrayList<String> langs;
		private final String limit;

		public KanjiToWordsIterator() {
			this.db = null;
			this.cursor = null;
			this.limit = String.format(LIMIT, Preferences.getInt(R.string.prefNameKanjiToWordsNWTL));
			this.langs = Preferences.getLangsPriorities(Common.Database.dictonary);
		}

		protected void finalize() throws Throwable {
			this.close();
			super.finalize();
		}

		/**
		 * Open the connection to the SQLite database.
		 * 
		 * @return true if the connection is done.
		 */
		public boolean open() {
			boolean ok = true;
			try {
				this.db = SQLiteDatabase.openDatabase(DictionaryDB.dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);
			} catch (SQLiteException sle) {
				ok = false;
				Log.e(TAG, sle.getMessage());
			}
			return ok;
		}

		/**
		 * Close the SQLite connection.
		 */
		public void close() {
			if (null != this.cursor) {
				this.cursor.close();
				this.cursor = null;
			}
			if (null != this.db) {
				this.db.close();
				this.db = null;
			}
		}

		/**
		 * 
		 * @param codepoint
		 *            Set the current kanji to look words
		 * @param noLimit
		 *            If true, remove the limit of results in the query
		 * @return true if there is some words founds
		 */
		public boolean setCodepoint(int codepoint, boolean noLimit) {

			if (this.cursor != null) {
				this.cursor.close();
			}
			final String[] selectionArgs = new String[2];
			selectionArgs[0] = "%" + new String(Character.toChars(codepoint)) + "%";
			selectionArgs[1] = "0";
			String limitClause = this.limit;
			if (noLimit) {
				limitClause = null;
			}
			this.cursor = this.db.query(DictionaryDB.TABLE_KANJI_ELEM, KanjiToWordsIterator.ARRAY_COLS_KANJI,
					KanjiToWordsIterator.SELECTION_KANJI, selectionArgs, null, null, null, limitClause);
			return this.cursor.getCount() != 0;
		}

		public boolean hasNext() {
			boolean ret = false;
			if (null != this.cursor && !this.cursor.isLast()) {
				ret = true;
			}
			return ret;
		}

		/**
		 * 
		 * @return the number of elements fetch, or -1 if close.
		 */
		public int getCount() {
			if (null != this.cursor && !this.cursor.isClosed()) {
				return this.cursor.getCount();
			}
			return -1;
		}

		/**
		 * @return A minimal Dictionary Entry (id/keb/pri, common reading, one
		 *         sense).
		 */
		public Entry next() {
			if (null != this.cursor && this.cursor.moveToNext()) {

				final int id = cursor.getInt(0);
				final String keb = cursor.getString(1);
				final String pri = cursor.getString(2);

				// The kanji element
				Entry entry = new Entry(id);
				KanjiElement ke = new KanjiElement();
				ke.keb = keb;

				ke.ke_pri = Common.splitString(pri, Common.SLASH);
				entry.k_ele = new ArrayList<KanjiElement>(1);
				entry.k_ele.add(ke);

				// The reading element
				String[] selectionArgs = new String[1];
				selectionArgs[0] = Integer.toString(id); // id=

				Cursor lcursor = db.query(TABLE_READING_ELEM, KanjiToWordsIterator.ARRAY_COLS_READING,
						KanjiToWordsIterator.SELECTION_READING, selectionArgs, null, null,
						KanjiToWordsIterator.ORDER_READING);
				if (lcursor.moveToFirst()) {
					entry.r_ele = new ArrayList<ReadingElement>(lcursor.getCount());
					do {
						ReadingElement re = new ReadingElement();
						re.reb = lcursor.getString(0);
						entry.r_ele.add(re);
					} while (lcursor.moveToNext());
				}
				lcursor.close();
				// Sense
				selectionArgs = new String[2];
				selectionArgs[0] = Integer.toString(id); // id=
				selectionArgs[1] = Integer.toString(0); // num=

				final ArrayList<String> lLangs = this.langs;
				final int langCount = lLangs.size();
				entry.senses = new ArrayList<Sense>();
				for (int i = 0; i < langCount; i++) {
					final String lang = lLangs.get(i);
					final String tableGloss = String.format(DictionaryDB.TABLE_GLOSS, lang);
					lcursor = db.query(tableGloss, DictionaryDB.ARRAY_COLS_GLOSS,
							KanjiToWordsIterator.SELECTION_GLOSS, selectionArgs, null, null,
							KanjiToWordsIterator.ORDER_GLOSS);
					if (lcursor.moveToFirst()) {
						Sense sense = new Sense();
						sense.glosses = new HashMap<String, String>(1);
						sense.glosses.put(lang, lcursor.getString(0));
						entry.senses.add(sense);
						i = langCount;
					}
					lcursor.close();
				}
				return entry;
			}
			this.cursor.close();
			this.cursor = null;
			return null;
		}

		/**
		 * throw UnsupportedOperationException()
		 */
		public void remove() {
			throw new UnsupportedOperationException("Can't delete data from the dictionary databse");
		}
	}

}
