/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictionary;

import java.util.ArrayList;

import org.vono.narau.Common;
import org.vono.narau.R;
import org.vono.narau.dictionary.jmdict.Entry;
import org.vono.narau.preferences.Preferences;

import android.util.Log;
import android.util.SparseArray;

public class DictionaryThread extends Thread {
	private static final String TAG = DictionaryThread.class.getSimpleName();
	private static final int MAX_MAP_ENTRIES = 1000;

	public static final char CHAR_SPACE = ' ';
	public static final char CHAR_SLASH = '/';

	private DictionaryActivity view;
	private boolean ready;
	private String searchText;
	/**
	 * Variable which control the dictionary to search. If there is some result
	 * in one dictionary, but we want to continue the search in others
	 * dictionaries, this var is used to know the next search dictionary.<br/>
	 * Use negative value for searching in Japanese dictionary.
	 */
	private int langSearchId;

	private ArrayList<Entry> arrayLastSearch;
	private SparseArray<Entry> mapLastSearch;

	private DictionaryThread(DictionaryActivity view) {
		this.view = view;
		this.ready = false;
		this.langSearchId = 0;
		this.arrayLastSearch = new ArrayList<Entry>();
		this.mapLastSearch = new SparseArray<Entry>(MAX_MAP_ENTRIES);

		this.searchText = Common.EMPTY_STRING;
	}

	private static DictionaryThread instance = null;

	public static void init(DictionaryActivity view) {
		synchronized (TAG) {
			if (null != view) {
				if (null == DictionaryThread.instance) {
					DictionaryThread.instance = new DictionaryThread(view);
					DictionaryThread.instance.start();
					try {
						TAG.wait();
					} catch (InterruptedException ie) {
					}

					// Log.d(TAG, "running");
				} else {
					synchronized (DictionaryThread.instance) {
						DictionaryThread.instance.view = view;
					}

				}
			}
		}
	}

	/**
	 * Release resources.
	 */
	public static void clean() {
		// Log.d(TAG, "clean()");
		synchronized (TAG) {
			if (null != DictionaryThread.instance) {

				synchronized (DictionaryThread.instance) {
					DictionaryThread.instance.ready = false;
					DictionaryThread.instance.notify();
				}

				DictionaryThread.instance.view = null;
				DictionaryThread.instance.arrayLastSearch.clear();
				DictionaryThread.instance.arrayLastSearch = null;
				DictionaryThread.instance.mapLastSearch.clear();
				DictionaryThread.instance.mapLastSearch = null;
				DictionaryThread.instance.searchText = Common.EMPTY_STRING;

				while (DictionaryThread.instance.isAlive()) {
					try {
						Thread.sleep(100);
					} catch (InterruptedException ie) {
					}
				}

			}
			DictionaryThread.instance = null;
		}
		// Log.d(TAG, "clean done");
	}

	@Override
	public void run() {
		// Log.d(TAG, "thread start");

		try {
			this.loopSearch();
		} catch (Exception e) {

			StackTraceElement[] stea = e.getStackTrace();

			StringBuilder sb = new StringBuilder();
			String msg = e.getLocalizedMessage();
			if (null == msg) {
				msg = Common.getString(R.string.unknownError);
			}
			sb.append(msg);
			sb.append("\n\n:");
			for (int i = 0; i < stea.length; i++) {
				StackTraceElement ste = stea[i];
				sb.append(ste.toString());
				sb.append("\n");
			}

			this.view.sendMessage(DictionaryActivity.KEY_MESSAGE, sb.toString());

			this.run();
			this.ready = false;
			Log.e(TAG, "re-run()", e);
		}

		// Log.d(TAG, "thread end");

	}

	private synchronized void loopSearch() {

		this.ready = true;
		synchronized (TAG) {
			TAG.notify();
		}

		String dictSearchLangsMessage = Common.getString(R.string.dictSearchLangsMessage);

		while (ready) {
			// if a search text was set before loading
			try {
				this.wait();
			} catch (InterruptedException ie) {
				ready = false;
			}

			if (this.searchText.length() > 0 && ready) {

				this.arrayLastSearch.clear();
				if (this.mapLastSearch.size() > MAX_MAP_ENTRIES) {
					Log.i(TAG, "clear map search (>MAX_MAP_ENTRIES)");
					this.mapLastSearch.clear();
				}

				final ArrayList<String> langs = Preferences.getLangsPriorities(Common.Database.dictonary);
				final int numLangs = langs.size();
				ArrayList<Integer> ids = null;

				if (this.langSearchId < 0) {
					int displayLangId = -this.langSearchId - 1;
					// check bounds
					displayLangId = (displayLangId >= numLangs) ? 0 : displayLangId;
					// search in Japanese tables
					for (int i = displayLangId; i < numLangs && ids == null; i++) {
						final String lang = langs.get(i);
						final String languageName = Common.getLocale(lang).getDisplayLanguage();
						final String dictSearchMsg = String.format(dictSearchLangsMessage, Common.JPN_LANG_NAME,
								languageName);
						this.view.sendMessage(DictionaryActivity.KEY_MESSAGE, dictSearchMsg);

						// Log.d(TAG, "search " + Common.JPN_LANG_NAME + " -> "
						// + lang + " (" + (i)
						// + ")");

						ids = DictionaryDB.findMatchKana(lang, this.searchText);
						if (null == ids) {
							ids = DictionaryDB.findMatchKanji(lang, this.searchText);
						}
						if (null != ids) {
							this.langSearchId = -i - 1;
						}
					}

				} else {
					// search in non Japanese tables
					// check bounds
					this.langSearchId = this.langSearchId >= numLangs ? 0 : this.langSearchId;

					for (int i = this.langSearchId; i < numLangs; i++) {
						String lang = langs.get(i);
						// Tell the user what is going on
						String languageName = Common.getLocale(lang).getDisplayLanguage();
						String dictSearchMsg = String
								.format(dictSearchLangsMessage, languageName, Common.JPN_LANG_NAME);
						this.view.sendMessage(DictionaryActivity.KEY_MESSAGE, dictSearchMsg);

						// Searching
						ids = DictionaryDB.findMatchGloss(lang, this.searchText);
						if (null != ids) {
							this.langSearchId = i;
							break;
						}
					}
				}

				if (null != ids) {

					int numIds = ids.size();
					// Log.v(TAG, "got " + numIds + " results");
					SparseArray<Entry> mapResult = this.mapLastSearch;
					ArrayList<Entry> arrayResult = this.arrayLastSearch;
					arrayResult.ensureCapacity(numIds);

					for (int i = 0; i < numIds; i++) {
						int id = ids.get(i);
						if (mapResult.indexOfKey(id) >= 0) {
							arrayResult.add(mapResult.get(id));
						} else {
							Entry entry = DictionaryDB.fetchdEntry(ids.get(i));
							arrayResult.add(entry);
							mapResult.put(id, entry);
						}
					}

				} else {
					this.langSearchId = 0;
					this.arrayLastSearch.clear();
				}

				this.view.sendMessage(DictionaryActivity.KEY_RESULT_AVAIL, Common.EMPTY_STRING);
				this.searchText = Common.EMPTY_STRING;
			}
		}
	}

	public static ArrayList<Entry> getLastSearch() {
		if (null != DictionaryThread.instance) {
			synchronized (DictionaryThread.instance) {
				// return DictionaryThread.instance.resultLastSearch;
				return DictionaryThread.instance.arrayLastSearch;
			}
		}

		return null;
	}

	public static void setSearchText(String text) {
		if (null != DictionaryThread.instance && null != text) {
			synchronized (DictionaryThread.instance) {
				DictionaryThread.instance.searchText = text;
				DictionaryThread.instance.notify();
			}

		}
	}

	public static void setLangSearchId(int id) {
		if (null != DictionaryThread.instance) {
			DictionaryThread.instance.langSearchId = id;
		}
	}

	public static int getLangSearchId() {
		if (null != DictionaryThread.instance) {
			return DictionaryThread.instance.langSearchId;
		}
		return 0;
	}

	public static Entry getEntry(int id) {
		Entry entry = null;
		if (null != DictionaryThread.instance) {
			entry = DictionaryThread.instance.mapLastSearch.get(id);
		}
		return entry;
	}

}
