/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.download;

import java.io.IOException;
import java.io.InputStream;
import java.text.DateFormat;
import java.util.Locale;

import org.vono.narau.AppException;
import org.vono.narau.Common;
import org.vono.narau.Narau;
import org.vono.narau.QuitThread;
import org.vono.narau.R;

import android.app.Activity;
import android.app.NotificationManager;
import android.app.ProgressDialog;
import android.content.Intent;
import android.os.Bundle;
import android.os.Environment;
import android.os.StatFs;
import android.text.Html;
import android.text.Spanned;
import android.text.method.LinkMovementMethod;
import android.util.Log;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.LinearLayout;
import android.widget.ProgressBar;
import android.widget.TextView;

/**
 * Manage the download database process (first run & update).
 * 
 */
public class DownloaderActivity extends Activity implements OnClickListener {
	private static final String TAG = DownloaderActivity.class.getSimpleName();

	public static final String UPDATE_PROCESS = "upd";
	public static final String KEY_LOG_MSGS = "log";
	public static final String KEY_CURRENT_DB = "db";

	private enum CurrentScreen {
		NONE, FIRST_RUN, QUESTION, DOWNLOADING, DOWNLOAD_DONE, INIT_ERROR;
	}

	public enum LauncherSource {
		NONE, FIRST_RUN, UPDATE, NOTIFICATION
	}

	/**
	 * The current download screen.
	 */
	private static CurrentScreen currentScreen = CurrentScreen.NONE;
	private LauncherSource launcherSource;

	private ProgressDialog progressDialog;
	private TextView textViewDownloadSpeed;
	private TextView textViewLog;
	private TextView textViewDatabase;
	private ProgressBar progressBar;
	private ProgressBar progressBarTotal;

	public DownloaderActivity() {
		this.progressBar = null;
		this.progressBarTotal = null;
		this.progressDialog = null;
		this.textViewDatabase = null;
		this.textViewDownloadSpeed = null;
		this.textViewLog = null;
		this.launcherSource = LauncherSource.NONE;
	}

	/** Called when the activity is created. */
	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		if (!Common.isInitDone()) {
			try {
				Common.init(super.getApplicationContext());
			} catch (AppException ae) {
				super.setContentView(R.layout.menu_init_error);
				TextView tv = (TextView) super.findViewById(R.id.initErrorTextView);
				tv.setText(ae.getMessage());
				DownloaderActivity.currentScreen = CurrentScreen.INIT_ERROR;
			}
		}

		this.launcherSource = (LauncherSource) super.getIntent().getSerializableExtra(UPDATE_PROCESS);

		if (this.launcherSource != LauncherSource.NOTIFICATION) {
			NotificationManager nm = (NotificationManager) getSystemService(NOTIFICATION_SERVICE);
			nm.cancel(R.id.download_db_notification);
		} else {
			Narau.finishOnUpdateDB();
		}

		switch (DownloaderActivity.currentScreen) {
		case NONE:
			if (Environment.MEDIA_MOUNTED_READ_ONLY.equals(Environment.getExternalStorageState())) {
				// Error we can't write a file in a read only mounted
				// storage.
				super.setContentView(R.layout.menu_init_error);
				TextView tv = (TextView) super.findViewById(R.id.initErrorTextView);
				tv.setText(R.string.initSdcardReadOnly);
			} else if (this.launcherSource != LauncherSource.FIRST_RUN) {
				this.firstRunContinue(null);
			} else {
				this.showFirstRunScreen();
			}
			break;
		case FIRST_RUN:
			this.showFirstRunScreen();
			break;
		case QUESTION:
			this.showDownloadViewQuestion();
			break;
		case DOWNLOADING:
		case DOWNLOAD_DONE:
			this.showDownloadView();
			break;
		case INIT_ERROR:
			break;
		}
	}

	@Override
	protected void onDestroy() {
		super.onDestroy();
		DownloaderGzipDB.updateUIActivity(null);
	}

	@Override
	public void onBackPressed() {
		// Don't finish this activity, instead show the device home screen.
		final Intent setIntent = new Intent(Intent.ACTION_MAIN);
		setIntent.addCategory(Intent.CATEGORY_HOME);
		setIntent.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
		super.startActivity(setIntent);
	}

	private void showFirstRunScreen() {
		super.setContentView(R.layout.download_firstrun);

		// Read the welcome HTML file
		final InputStream inputStream = getResources().openRawResource(R.raw.welcome);
		final byte[] buffer = new byte[512];
		final StringBuilder sbWelcome = new StringBuilder(512);
		try {
			boolean eof = false;
			do {
				final int count = inputStream.read(buffer);
				if (-1 == count) {
					eof = true;
				} else {
					final String str = new String(buffer, 0, count, "UTF-8");
					sbWelcome.append(str);
				}
			} while (!eof);
		} catch (IOException ioe) {
			Log.e(TAG, ioe.getMessage(), ioe);
		}

		try {
			inputStream.close();
		} catch (IOException ioe) {
			Log.e(TAG, ioe.getMessage(), ioe);
		}

		final Spanned html = Html.fromHtml(sbWelcome.toString());
		TextView tv = (TextView) super.findViewById(R.id.downloadFirstRunTextViewWelcome);
		tv.setMovementMethod(LinkMovementMethod.getInstance());
		tv.setText(html);

		DownloaderActivity.currentScreen = CurrentScreen.FIRST_RUN;
		Button button = (Button) super.findViewById(R.id.downloadFirstRunButtonContinue);
		button.setOnClickListener(this);
		button = (Button) super.findViewById(R.id.downloadFirstRunButtonQuit);
		button.setOnClickListener(this);
	}

	private void showDownloadViewQuestion() {
		super.setContentView(R.layout.download_question);
		DownloaderActivity.currentScreen = CurrentScreen.QUESTION;

		final LinearLayout linearLayout = (LinearLayout) super.findViewById(R.id.downloadQuestionLayoutCheckboxes);

		final int numDownloads = DBInfos.arrayDBInfos.length;
		for (int i = 0; i < numDownloads; i++) {
			final DBInfos dbInfo = DBInfos.arrayDBInfos[i];
			final View partView = super.getLayoutInflater().inflate(R.layout.download_question_part, null);
			CheckBox cb = (CheckBox) partView.findViewById(R.id.downloadQuestionPartCheckBox);

			String cbText = String.format(super.getString(R.string.downloadQuestionSize),
					super.getString(dbInfo.strDBNameId), this.bytesToHumanSize(dbInfo.gzSize),
					this.bytesToHumanSize(dbInfo.dbSize));
			cb.setText(cbText);
			cb.setChecked(dbInfo.selected);
			cb.setOnClickListener(this);
			dbInfo.checkBox = cb;

			TextView tv = (TextView) partView.findViewById(R.id.downloadQuestionPartTextViewInfos);
			DateFormat df = DateFormat.getDateTimeInstance();
			tv.setText(super.getString(R.string.downloadQuestionUpdateDate, dbInfo.dbName,
					df.format(dbInfo.getCreationDate())));

			if (dbInfo.haveLatest()) {
				partView.findViewById(R.id.downloadQuestionPartTextViewWarnLatest).setVisibility(View.VISIBLE);
			}

			partView.setId(i);
			linearLayout.addView(partView);
		}

		this.onClickCheckBoxDB(null);

		Button button = (Button) super.findViewById(R.id.downloadQuestionButtonOk);
		button.setOnClickListener(this);
		button = (Button) super.findViewById(R.id.downloadQuestionButtonCancel);
		button.setOnClickListener(this);
	}

	/**
	 * Show the download view
	 */
	private void showDownloadView() {
		super.setContentView(R.layout.download);
		this.textViewDownloadSpeed = (TextView) super.findViewById(R.id.downloadTextViewSpeed);
		this.textViewLog = (TextView) super.findViewById(R.id.downloadTextViewLogs);
		this.textViewDatabase = (TextView) super.findViewById(R.id.downloadTextViewDatabase);
		this.progressBar = (ProgressBar) super.findViewById(R.id.downloadProgressBarDatabase);
		this.progressBarTotal = (ProgressBar) super.findViewById(R.id.downloadProgressBarTotal);

		DownloaderGzipDB.updateUIActivity(this);

		Button button = (Button) super.findViewById(R.id.downloadButtonCancel);
		button.setOnClickListener(this);
		button = (Button) super.findViewById(R.id.downloadButtonContinue);
		button.setOnClickListener(this);
	}

	/**
	 * menu_download_question.xml<br/>
	 * Click on OK button in the download now view.
	 * 
	 * @param v
	 *            the clicked button
	 */
	public void onClickButtonOk(View v) {
		DownloaderActivity.currentScreen = CurrentScreen.DOWNLOADING;

		this.showDownloadView();
		DownloaderGzipDB.download(this);
	}

	/**
	 * Click on the continue button in the first run layout.
	 * 
	 * @param v
	 *            the button
	 */
	public void firstRunContinue(View v) {
		super.setContentView(R.layout.menu_init_error);
		final String waitMessage = super.getString(R.string.downloadQuestionFetchURL);
		this.progressDialog = ProgressDialog.show(this, Common.EMPTY_STRING, waitMessage, true);
		DownloaderDBInfos.startDownload(this);
	}

	/**
	 * 
	 * @param v
	 *            The continue button on the download.xml
	 */
	public void showMainMenu(View v) {
		DownloaderActivity.currentScreen = CurrentScreen.NONE;
		DownloaderGzipDB.clean();
		Intent intent = new Intent(super.getApplicationContext(), Narau.class);
		super.startActivity(intent);
		super.finish();
	}

	public void dbInfoJobDone(boolean retValue, String message) {
		this.progressDialog.dismiss();
		this.progressDialog = null;
		if (retValue) {
			this.showDownloadViewQuestion();
		} else {
			TextView tv = (TextView) super.findViewById(R.id.initErrorTextView);
			tv.setText(message);
			Button button = (Button) super.findViewById(R.id.initErrorButtonQuit);
			button.setOnClickListener(this);
		}
	}

	/**
	 * menu_download_question.xml<br/>
	 * Update the total download size, when click on a checkbox.
	 * 
	 * @param v
	 *            The checkbox clicked
	 */
	public void onClickCheckBoxDB(View v) {
		final StatFs stat = new StatFs(Environment.getExternalStorageDirectory().getPath());
		final long sdAvailSize = ((long) stat.getAvailableBlocks()) * ((long) stat.getBlockSize());

		long dwnlSize = 0;
		long needSize = 0;

		final int numDownloads = DBInfos.arrayDBInfos.length;
		for (int i = 0; i < numDownloads; i++) {
			final DBInfos dbInfo = DBInfos.arrayDBInfos[i];
			if (dbInfo.checkBox.isChecked()) {
				dwnlSize += dbInfo.gzSize;
				needSize += dbInfo.dbSize;
				dbInfo.selected = true;
			} else {
				dbInfo.selected = false;
			}
		}

		String text = String.format(super.getString(R.string.downloadQuestionTotal), this.bytesToHumanSize(dwnlSize));
		TextView tv = (TextView) super.findViewById(R.id.downloadQuestionTextViewTotal);
		tv.setText(text);

		tv = (TextView) super.findViewById(R.id.downloadQuestionTextViewWarnLowSpace);
		if (sdAvailSize <= needSize) {
			text = String.format(super.getString(R.string.downloadQuestionWarnLowSpace),
					this.bytesToHumanSize(sdAvailSize), this.bytesToHumanSize(needSize));
			tv.setText(text);
			tv.setVisibility(View.VISIBLE);
		} else {
			tv.setVisibility(View.GONE);
		}

		Button btOK = (Button) super.findViewById(R.id.downloadQuestionButtonOk);
		if (0 == needSize || sdAvailSize <= needSize) {
			// If no item selected, disable the ok button
			// or if there is not enough space on the sdcard
			btOK.setEnabled(false);
		} else {
			btOK.setEnabled(true);
		}
	}

	/**
	 * Handle a press on a quit/cancel button:<br/>
	 * <ul>
	 * <li>on the first run screen</li>
	 * <li>on the question screen</li>
	 * <li>on the download screen.</li>
	 * </ul>
	 * 
	 * @param v
	 *            a button
	 */
	public void quitApplication(View v) {

		if (null != this.progressDialog) {
			this.progressDialog.dismiss();
			this.progressDialog = null;
		}

		DownloaderActivity.currentScreen = CurrentScreen.NONE;

		DownloaderGzipDB.prematureThreadStop();
		super.finish();

		if (this.launcherSource == LauncherSource.UPDATE) {
			Common.clean();
			Intent intent = new Intent(super.getApplicationContext(), Narau.class);
			super.startActivity(intent);
		} else {
			QuitThread.startQuitThread(2000);
		}
	}

	/**
	 * Update the ui with the download informations.
	 * 
	 * @param progressTotal
	 *            if of the download
	 * @param progress
	 *            the progress in percent
	 * @param speed
	 *            the speed in bytes/seconds
	 */
	public void updateDownloadProgressAndSpeed(int progressTotal, int progress, long speed) {
		if (null != this.textViewDownloadSpeed && speed >= 0) {
			String text = super.getString(R.string.downloadSpeed, this.bytesToHumanSize(speed));
			this.textViewDownloadSpeed.setText(text);
		} else {
			this.textViewDownloadSpeed.setText(Common.EMPTY_STRING);
		}

		this.progressBar.setProgress(progress);
		this.progressBarTotal.setProgress(progressTotal);
	}

	public void updateDownloadLog(StringBuilder log) {
		this.textViewLog.setText(Html.fromHtml(log.toString()));
	}

	public void updateDownloadName(String value) {
		this.textViewDatabase.setText(value);
	}

	/**
	 * The downloader thread tell the ui that is job is done.
	 */
	public void downloadComplete() {
		Button button = (Button) super.findViewById(R.id.downloadButtonCancel);
		button.setVisibility(View.GONE);
		button = (Button) super.findViewById(R.id.downloadButtonContinue);
		button.setVisibility(View.VISIBLE);
		this.textViewDownloadSpeed.setText("");
		this.updateDownloadProgressAndSpeed(100, 100, -1);
		DownloaderActivity.currentScreen = CurrentScreen.DOWNLOAD_DONE;

		Common.clean();
	}

	private String bytesToHumanSize(long value) {
		final String[] sizes = super.getResources().getStringArray(R.array.byteSizeHumanDisplay);
		float humanSize = value;
		int count = 0;

		while (humanSize > 1024 && count < sizes.length) {
			humanSize = humanSize / 1024f;
			count++;
		}

		return String.format(Locale.US, "%1$.1f%2$s", humanSize, sizes[count]);
	}

	public static boolean isActive() {
		switch (DownloaderActivity.currentScreen) {
		case DOWNLOADING:
		case DOWNLOAD_DONE:
		case FIRST_RUN:
		case INIT_ERROR:
		case QUESTION:
			return true;
		default:
			break;
		}
		return false;
	}

	public void onClick(View v) {
		switch (v.getId()) {

		case R.id.downloadFirstRunButtonContinue:
			this.firstRunContinue(v);
			break;
		case R.id.downloadQuestionButtonOk:
			this.onClickButtonOk(v);
			break;
		case R.id.downloadButtonContinue:
			this.showMainMenu(v);
			break;
		case R.id.downloadQuestionPartCheckBox:
			this.onClickCheckBoxDB(v);
			break;
		case R.id.downloadQuestionButtonCancel:
		case R.id.downloadButtonCancel:
		case R.id.initErrorButtonQuit:
		case R.id.downloadFirstRunButtonQuit:
			this.quitApplication(v);
			break;
		}

	}
}
