/**
 * Copyright (C) 2012 Yvon TANGUY
 * 
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.download;

import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Locale;

import org.vono.narau.Common;
import org.vono.narau.R;

import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.util.Log;

/**
 * 
 * Download a small text file with the downloads informations.
 */
public class DownloaderDBInfos {
	private static final String TAG = DownloaderDBInfos.class.getSimpleName();

	private static final String KEY_MSG_DBINFO_MSG = "msg";
	private static final String KEY_MSG_DBINFO_DONE = "dbi";

	private static final String VER_CODE = "NarauVersionCode:";

	/**
	 * <p>
	 * Download the narau_db_infos.txt files on launchpad.net.<br/>
	 * See DATABASE_FILEINFO_URL.
	 * </p>
	 * Fill informations in DownloaderDBInfos.dbinfos
	 */
	private static class DBInfoDownloadThread extends Thread {

		@Override
		public void run() {
			boolean success = false;

			try {
				final URL url = new URL(Common.getResourcesString(R.string.db_infos_url));

				StringBuilder sb = DBInfoDownloadThread.downloadFile(url);
				int versionCode = Common.getVersionCode();
				String searchCode = VER_CODE + versionCode;

				int startIdx = sb.indexOf(searchCode);
				if (startIdx == -1) {
					UIHandler.showMessage(Common.getString(R.string.downloadErrorNoDBInfo));
				} else {
					int endIdx = sb.indexOf(VER_CODE, startIdx + VER_CODE.length());
					if (endIdx == -1) {
						endIdx = sb.length();
					}
					success = DBInfoDownloadThread.updateDBInfos(sb.substring(startIdx, endIdx));
				}
			} catch (MalformedURLException mue) {
				String message = mue.getMessage();
				if (null == message) {
					message = "Empty URL";
				}
				UIHandler.showMessage(message);
				Log.e(TAG, message);
			}
			UIHandler.donwloadThreadEnd(success);
		}

		/**
		 * Perform the download
		 * 
		 * @param url
		 *            the url of the file to download.
		 * @return null if there is an error, or the file content.
		 */
		private static StringBuilder downloadFile(URL url) {
			HttpURLConnection connection = null;
			final StringBuilder sb = new StringBuilder();

			try {
				connection = (HttpURLConnection) url.openConnection();
				connection.connect();

				final int responseCode = connection.getResponseCode();
				if (responseCode / 100 != 2) {
					final String message = "" + responseCode + ": " + connection.getResponseMessage();
					UIHandler.showMessage(message);
					UIHandler.showMessage(Common.getString(R.string.downloadErrorGeneric,
							Common.getResourcesString(R.string.db_infos_url)));

					connection.disconnect();
					return null;
				}

				InputStream is = connection.getInputStream();
				int read = is.read();
				while (read != -1) {
					sb.append((char) read);
					read = is.read();

				}
				connection.disconnect();
			} catch (IOException ioe) {
				UIHandler.showMessage(ioe.getClass().getSimpleName() + ": " + ioe.getMessage());
				Log.e(TAG, ioe.getMessage(), ioe);
			} finally {
				if (null != connection) {
					connection.disconnect();
				}

			}
			return sb;
		}

		/**
		 * Update the dbInfos values with theses found in the downloaded file.
		 * 
		 * @param sb
		 *            The file content
		 * @return true if everything went OK.
		 */
		private static boolean updateDBInfos(String textToParse) {
			final int numDBInfos = DBInfos.arrayDBInfos.length;
			for (int i = 0; i < numDBInfos; i++) {
				final DBInfos infos = DBInfos.arrayDBInfos[i];
				if (DBInfoDownloadThread.getDBInfos(textToParse, infos)) {
					if (infos.expectedDBVersion < infos.dbVersion) {
						UIHandler.showMessage(Common.getString(R.string.downloadErrorDBNewer, infos.dbName));
					} else if (infos.expectedDBVersion > infos.dbVersion) {
						UIHandler.showMessage(Common.getString(R.string.downloadErrorDBOlder, infos.dbName));
					} else {
						infos.infosLoaded = true;
					}

				}
			}

			// Check if everything are OK
			boolean success = true;
			for (int i = 0; i < numDBInfos; i++) {
				final DBInfos infos = DBInfos.arrayDBInfos[i];
				success = success && infos.infosLoaded;
			}
			return success;
		}

		/**
		 * 
		 * @param sb
		 *            The content of the file
		 * @param infos
		 *            The DBInfos to fill in.
		 * @return true if the parsing is OK
		 */
		private static boolean getDBInfos(String text, DBInfos infos) {
			boolean parseOk = false;
			final int startIdx = text.indexOf(infos.name);
			if (startIdx > 0) {
				final int endIdx = text.indexOf("\n", startIdx);
				final String line = text.substring(startIdx, endIdx);
				final String[] values = line.split(" ");

				if (values.length < 7) {
					UIHandler.showMessage(Common.getString(R.string.downloadErrorLineFormat, line));
					return false;
				}

				try {
					// # gz_filename db_size gz_size gz_file_time gz_md5
					// db_version download_url another_download_url...
					int idx = 0;

					infos.dbSize = Long.parseLong(values[++idx]);
					infos.gzSize = Long.parseLong(values[++idx]);
					infos.setCreationDate(Long.parseLong(values[++idx]) * 1000);
					infos.md5 = values[++idx].toLowerCase(Locale.US);
					infos.dbVersion = Integer.parseInt(values[++idx]);
					idx++;
					infos.url = new URL[values.length - idx];
					for (int i = idx; i < values.length; i++) {
						infos.url[i - idx] = new URL(values[i]);
					}
					parseOk = true;

				} catch (NumberFormatException nfe) {
					Log.e(TAG, nfe.getMessage());
					UIHandler.showMessage(Common.getString(R.string.downloadErrorNumberFormat, nfe.getMessage(),
							infos.dbName));
				} catch (MalformedURLException mue) {
					Log.e(TAG, mue.getMessage());
					UIHandler.showMessage(Common.getString(R.string.downloadErrorURLFormat, mue.getMessage()));
				}

			} else {
				UIHandler.showMessage(Common.getString(R.string.downloadErrorLineNotFound, infos.dbName));
			}
			return parseOk;
		}
	}

	private static class UIHandler extends Handler {
		private static UIHandler instance;
		private DownloaderActivity activity;

		private StringBuilder messages;

		public UIHandler(DownloaderActivity activity) {
			this.activity = activity;
			this.messages = new StringBuilder();

			DBInfoDownloadThread thread = new DBInfoDownloadThread();
			thread.start();
		}

		private static void donwloadThreadEnd(boolean returnValue) {
			final Message msg = Message.obtain(UIHandler.instance);
			final Bundle bundle = msg.getData();
			bundle.clear();
			bundle.putBoolean(KEY_MSG_DBINFO_DONE, returnValue);
			UIHandler.instance.sendMessage(msg);

		}

		private static void showMessage(String message) {
			final Message msg = Message.obtain(UIHandler.instance);
			final Bundle bundle = msg.getData();
			bundle.clear();
			bundle.putString(KEY_MSG_DBINFO_MSG, message);
			UIHandler.instance.sendMessage(msg);
		}

		@Override
		public void handleMessage(Message msg) {
			final Bundle bundle = msg.getData();
			if (bundle.containsKey(KEY_MSG_DBINFO_MSG)) {
				this.messages.append(bundle.getString(KEY_MSG_DBINFO_MSG));
				this.messages.append("\n");
			} else if (bundle.containsKey(KEY_MSG_DBINFO_DONE)) {
				boolean threadReturn = bundle.getBoolean(KEY_MSG_DBINFO_DONE, false);
				this.activity.dbInfoJobDone(threadReturn, this.messages.toString());
				UIHandler.instance = null;
			}
		}
	}

	public static void startDownload(DownloaderActivity activity) {
		if (null == UIHandler.instance) {
			UIHandler.instance = new UIHandler(activity);
		}
	}
}
