/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.kanji;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeSet;

import org.vono.narau.Common;
import org.vono.narau.Common.Database;
import org.vono.narau.Common.DatabseState;
import org.vono.narau.preferences.Preferences;

import android.database.Cursor;
import android.database.SQLException;
import android.database.sqlite.SQLiteDatabase;
import android.util.Log;

/**
 * 
 * Read the kanji svg database.
 * 
 */
public class KanjiDB {
	private static final String TAG = KanjiDB.class.getSimpleName();

	public static final int VALUE_DB_VERSION = 3;

	// KanjiDict
	private static final String TABLE_DICT_KANJI = "kdict_kanji";
	private static final String TABLE_DICT_LANG_MEANINGS = "kdict_meanings_%1$s";
	// private static final String TABLE_DICT_INDEX = "kdict_index";
	// private static final String TABLE_DICT_REF = "kdict_ref";
	private static final String TABLE_DICT_LANGS = "kdict_langs";

	// KanjiVG
	private static final String TABLE_KVG_GROUP = "kvg_group";
	private static final String TABLE_KVG_PATH = "kvg_path";
	private static final String TABLE_KVG_RAD = "kvg_radicals";
	private static final String TABLE_KVG_POS = "kvg_positions";
	private static final String VIEW_KVG_SCHOOL = "kvg_school";

	public static final String WHERE_CODEPOINT = "codepoint=?";
	public static final String WHERE_CP_IS_AND_RAD_NOT_NULL = WHERE_CODEPOINT + " AND radical IS NOT NULL";

	private static final String COL_CODEPOINT = "codepoint";
	private static final String COL_KANJI_GRADE = "grade";
	private static final String COL_KANJI_FREQUENCY = "frequency";
	private static final String COL_KANJI_JLPT = "jlpt";
	private static final String COL_KANJI_STROKE_COUNT = "stroke_count";
	private static final String COL_KANJI_READING_ON = "reading_on";
	private static final String COL_KANJI_READING_KUN = "reading_kun";
	private static final String COL_KVG_GROUPS_ELEMENT = "element";
	private static final String COL_KVG_GROUPS_ORIGINAL = "original";

	private static final String[] COLS_KANJI_WITH_READING = { COL_CODEPOINT, COL_KANJI_GRADE, COL_KANJI_FREQUENCY,
			COL_KANJI_JLPT, COL_KANJI_STROKE_COUNT, COL_KANJI_READING_ON, COL_KANJI_READING_KUN };
	private static final String[] COLS_KANJI_NO_READING = { COL_CODEPOINT, COL_KANJI_GRADE, COL_KANJI_FREQUENCY,
			COL_KANJI_JLPT, COL_KANJI_STROKE_COUNT };
	private static final String[] COLS_KANJIVG_CODEPOINT = { COL_CODEPOINT };
	private static final String[] COLS_KVG_RADPOS = { "value" };
	private static final String[] COLS_DICT_LANG_MEANINGS = { "meanings" };
	private static final String[] COLS_KVG_GROUPS = { "id", "parent", "element", "position", "radical", "original" };
	private static final String[] COLS_KVG_GROUPS_RADS = { "element", "original" };
	private static final String[] COLS_KVG_PATH = { "parent", "type", "draw" };

	private static final String[] COLS_DICT_LANGS = { "lang" };

	private static final String ORDER_BY_CODEPOINT = "codepoint asc";
	private static final String ORDER_BY_ID = "id asc";
	private static final String ORDER_BY_PARENT = "parent ASC";

	private static String dbFullName = null;
	private static Common.DatabseState state;

	private static String[] radicals = null;
	private static String[] positions = null;

	/**
	 * 
	 * @return true if the db exists and have a good format version.
	 */
	public static Common.DatabseState initDB() {
		KanjiDB.dbFullName = Common.dataPath + File.separator + Common.DB_KANJI_NAME;
		KanjiDB.state = DatabseState.missing;

		// check db version
		SQLiteDatabase db = null;
		Cursor cursor = null;
		File dbFile = new File(KanjiDB.dbFullName);
		if (dbFile.canRead()) {
			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				cursor = db.query(Common.TABLE_DBINFO, Common.COLS_DBINFO, null, null, null, null, null);
				if (cursor.moveToFirst()) {
					int idxColKey = cursor.getColumnIndex(Common.COLS_DBINFO[0]);
					int idxColValueInt = cursor.getColumnIndex(Common.COLS_DBINFO[2]);
					do {
						String key = cursor.getString(idxColKey);
						if (Common.KEY_DB_VERSION.equals(key)) {
							int version = cursor.getInt(idxColValueInt);
							if (VALUE_DB_VERSION == version) {
								KanjiDB.state = DatabseState.ok;
							} else {
								KanjiDB.state = DatabseState.deprecated;
							}
						}
					} while (cursor.moveToNext());
				}
				cursor.close();

				if (DatabseState.ok == KanjiDB.state) {
					cursor = db.query(TABLE_KVG_POS, COLS_KVG_RADPOS, null, null, null, null, ORDER_BY_ID);
					if (cursor.moveToFirst()) {
						final int colIdxVal = cursor.getColumnIndex(COLS_KVG_RADPOS[0]);
						final int count = cursor.getCount();
						KanjiDB.positions = new String[count];

						for (int i = 0; i < count; i++) {
							KanjiDB.positions[i] = cursor.getString(colIdxVal);
							cursor.moveToNext();
						}
					}
					cursor.close();

					cursor = db.query(TABLE_KVG_RAD, COLS_KVG_RADPOS, null, null, null, null, ORDER_BY_ID);
					if (cursor.moveToFirst()) {
						final int colIdxVal = cursor.getColumnIndex(COLS_KVG_RADPOS[0]);
						final int count = cursor.getCount();
						KanjiDB.radicals = new String[count];

						for (int i = 0; i < count; i++) {
							KanjiDB.radicals[i] = cursor.getString(colIdxVal);
							cursor.moveToNext();
						}
					}

				}

			} catch (SQLException se) {
				Log.e(TAG, "initDB()", se);
			} finally {
				if (null != cursor) {
					cursor.close();
				}
				if (null != db) {
					db.close();
				}
			}

		}
		return KanjiDB.state;
	}

	public static void clean() {
		KanjiDB.state = DatabseState.missing;
		KanjiDB.dbFullName = null;
		KanjiDB.radicals = null;
		KanjiDB.positions = null;
	}

	public static boolean isReady() {
		return DatabseState.ok == KanjiDB.state;
	}

	public static Common.DatabseState getState() {
		return KanjiDB.state;
	}

	/**
	 * 
	 * @return a list of langs where there is meanings, or null if an error
	 *         occur.
	 */
	public static ArrayList<String> getLangs() {
		ArrayList<String> langs = null;
		if (DatabseState.ok == KanjiDB.state) {
			SQLiteDatabase db = null;

			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);
				langs = KanjiDB.getLangs(db);

			} finally {
				if (null != db) {
					db.close();
				}
			}
		}

		return langs;
	}

	private static ArrayList<String> getLangs(SQLiteDatabase db) {
		ArrayList<String> langs = null;
		Cursor cursor = null;
		try {
			cursor = db.query(true, TABLE_DICT_LANGS, COLS_DICT_LANGS, null, null, null, null, null, null);

			if (cursor.moveToFirst()) {
				langs = new ArrayList<String>(cursor.getCount());

				do {
					Common.addLangToList(cursor.getString(0), langs);
				} while (cursor.moveToNext());
			}
		} finally {
			if (null != cursor) {
				cursor.close();
			}
		}

		return langs;
	}

	/**
	 * 
	 * @param codepoint
	 *            unicode value of the kanji
	 * @return The loaded kanji (with meaning and svg path) or null if not found
	 */
	public static Kanji getKanji(int codepoint) {
		Kanji kanji = Kanji.getKanji(codepoint);
		final boolean needToLoad = !(kanji.kanjiDBLoadDictRef && kanji.kanjiDBLoadKanji && kanji.kanjiDBLoadMeaning && kanji.kanjiDBLoadSVG);

		if (needToLoad && (DatabseState.ok == KanjiDB.state)) {
			SQLiteDatabase db = null;
			Cursor cursor = null;
			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				String[] selectionArgs = new String[1];
				selectionArgs[0] = Integer.toString(codepoint);
				cursor = db.query(TABLE_DICT_KANJI, COLS_KANJI_WITH_READING, WHERE_CODEPOINT, selectionArgs, null,
						null, null);
				if (cursor.moveToFirst()) {
					kanji = KanjiDB.cursorToKanji(db, cursor, kanji, true, true, true);
				}
			} finally {
				if (null != cursor)
					cursor.close();
				if (null != db)
					db.close();
			}
		}
		return kanji;
	}

	public static boolean getKanji(Kanji kanji) {
		boolean ret = true;
		final boolean needToLoad = !(kanji.kanjiDBLoadDictRef && kanji.kanjiDBLoadKanji && kanji.kanjiDBLoadMeaning && kanji.kanjiDBLoadSVG);

		if (needToLoad && (DatabseState.ok == KanjiDB.state)) {
			SQLiteDatabase db = null;
			Cursor cursor = null;
			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				String[] selectionArgs = new String[1];
				selectionArgs[0] = Integer.toString(kanji.codepoint);
				cursor = db.query(TABLE_DICT_KANJI, COLS_KANJI_WITH_READING, WHERE_CODEPOINT, selectionArgs, null,
						null, null);
				if (cursor.moveToFirst()) {
					ret = KanjiDB.cursorToKanji(db, cursor, kanji, true, true, true) != null;
				} else {
					ret = false;
				}
			} finally {
				if (null != cursor)
					cursor.close();
				if (null != db)
					db.close();
			}
		}

		return ret;
	}

	public static void getKanjiVG(Kanji kanji) {

		if (!kanji.kanjiDBLoadSVG && (DatabseState.ok == KanjiDB.state)) {
			SQLiteDatabase db = null;
			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);
				KanjiDB.getKanjiVG(kanji, db);
			} finally {
				if (null != db)
					db.close();
			}
		}
	}

	public static void getMeaning(Kanji kanji) {
		if (!kanji.kanjiDBLoadMeaning && (DatabseState.ok == KanjiDB.state)) {
			SQLiteDatabase db = null;
			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);
				KanjiDB.getMeanings(kanji, db);
			} finally {
				if (null != db)
					db.close();
			}
		}
	}

	/**
	 * Load only the radicals from the SVG informations.
	 * 
	 * @param kanji
	 */
	public static TreeSet<Integer> getKanjiRadicals(int[] codepoints) {

		final TreeSet<Integer> setRadicals = new TreeSet<Integer>();

		if (DatabseState.ok == KanjiDB.state) {
			SQLiteDatabase db = null;
			Cursor cursor = null;

			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				final String[] selectionArgs = new String[1];

				final int numKanjis = codepoints.length;
				for (int i = 0; i < numKanjis; i++) {
					final Kanji kanji = Kanji.getKanji(codepoints[i]);
					if (kanji.kanjiDBLoadRadicals) {
						// Radical already loaded
						setRadicals.addAll(kanji.radicals);
						continue;
					}

					selectionArgs[0] = Integer.toString(kanji.codepoint);
					cursor = db.query(TABLE_KVG_GROUP, COLS_KVG_GROUPS_RADS, WHERE_CP_IS_AND_RAD_NOT_NULL,
							selectionArgs, null, null, null);

					final TreeSet<Integer> radCodePoints = new TreeSet<Integer>();

					if (cursor.moveToFirst()) {
						final Integer colElemIdx = cursor.getColumnIndex(COL_KVG_GROUPS_ELEMENT);
						final Integer colOrigIdx = cursor.getColumnIndex(COL_KVG_GROUPS_ORIGINAL);

						do {

							Integer cpElem = cursor.getInt(colElemIdx);
							Integer cpOrig = cursor.getInt(colOrigIdx);
							if (null != cpElem && !radCodePoints.contains(cpElem)) {
								radCodePoints.add(cpElem);
							}
							if (null != cpOrig && !radCodePoints.contains(cpOrig)) {
								radCodePoints.add(cpOrig);
							}
						} while (cursor.moveToNext());

					}

					kanji.radicals = new ArrayList<Integer>(radCodePoints);
					kanji.kanjiDBLoadRadicals = true;
					setRadicals.addAll(radCodePoints);
					cursor.close();
					cursor = null;
				}

			} finally {
				if (null != cursor) {
					cursor.close();
				}
				if (null != db) {
					db.close();
				}

			}
		}

		return setRadicals;
	}

	/**
	 * 
	 * @param where
	 *            The query where clause.
	 * @return An iterator for the kanjis
	 */
	public static KanjiDBIterator getKanjiIterator(String selection, String[] selectionArgs, boolean loadMeaning,
			boolean loadSvg, boolean loadReading) {
		if (DatabseState.ok != KanjiDB.state) {
			return null;
		}

		SQLiteDatabase db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
				| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

		Cursor cursor = null;

		if (loadReading) {
			cursor = db.query(TABLE_DICT_KANJI, COLS_KANJI_WITH_READING, selection, selectionArgs, null, null,
					ORDER_BY_CODEPOINT);
		} else {
			cursor = db.query(TABLE_DICT_KANJI, COLS_KANJI_NO_READING, selection, selectionArgs, null, null,
					ORDER_BY_CODEPOINT);
		}

		KanjiDB.KanjiDBIterator ite = new KanjiDB.KanjiDBIterator(db, cursor, loadMeaning, loadSvg, loadReading);
		return ite;
	}

	/**
	 * Return an iterator on kanjis codepoint where an SVG drawing is available.
	 * 
	 * @param selection
	 *            where clause
	 * @param selectionArgs
	 *            where values
	 * @return an KanjiSVGIterator or null if the database do not have the good
	 *         version.
	 */
	public static KanjiSVGIterator getKanjiSVGIterator(String selection, String[] selectionArgs) {
		if (DatabseState.ok != KanjiDB.state) {
			return null;
		}

		SQLiteDatabase db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
				| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

		Cursor cursor = db.query(VIEW_KVG_SCHOOL, COLS_KANJIVG_CODEPOINT, selection, selectionArgs, null, null,
				ORDER_BY_CODEPOINT);

		KanjiDB.KanjiSVGIterator ite = new KanjiSVGIterator(db, cursor);

		return ite;
	}

	private static void getMeanings(Kanji kanji, SQLiteDatabase db) {
		if (kanji.kanjiDBLoadMeaning) {
			return;
		}

		String[] selectionArgs = new String[1];
		selectionArgs[0] = Integer.toString(kanji.codepoint);

		final ArrayList<String> langsPri = Preferences.getLangsPriorities(Database.kanji);
		final int numLangs = langsPri.size();
		if (null == kanji.meanings) {
			kanji.meanings = new HashMap<String, String>(1);
		}
		final HashMap<String, String> map = kanji.meanings;

		for (int i = 0; i < numLangs; i++) {
			final String lang = langsPri.get(i);
			final String table = String.format(TABLE_DICT_LANG_MEANINGS, lang);
			Cursor cursor = db.query(table, COLS_DICT_LANG_MEANINGS, WHERE_CODEPOINT, selectionArgs, null, null, null);

			if (cursor.moveToFirst()) {
				final int idxColMeaning = cursor.getColumnIndex(COLS_DICT_LANG_MEANINGS[0]);

				do {
					String meanings = cursor.getString(idxColMeaning);
					map.put(lang, meanings);

				} while (cursor.moveToNext());
				i = numLangs;
			}

			cursor.close();
		}
		kanji.kanjiDBLoadMeaning = true;
	}

	private static void getKanjiVG(Kanji kanji, SQLiteDatabase db) {
		if (kanji.kanjiDBLoadSVG) {
			return;
		}

		Kanji.Group[] groups = null;
		final String[] selectionArgs = new String[1];
		selectionArgs[0] = Integer.toString(kanji.codepoint);
		Cursor cursor = db.query(TABLE_KVG_GROUP, COLS_KVG_GROUPS, WHERE_CODEPOINT, selectionArgs, null, null,
				ORDER_BY_ID);

		if (cursor.moveToFirst()) {
			final int idxColId = cursor.getColumnIndex(COLS_KVG_GROUPS[0]);
			final int idxColParent = cursor.getColumnIndex(COLS_KVG_GROUPS[1]);
			final int idxColElement = cursor.getColumnIndex(COL_KVG_GROUPS_ELEMENT);
			final int idxColPosition = cursor.getColumnIndex(COLS_KVG_GROUPS[3]);
			final int idxColRadical = cursor.getColumnIndex(COLS_KVG_GROUPS[4]);
			final int idxColOriginal = cursor.getColumnIndex(COL_KVG_GROUPS_ORIGINAL);

			groups = new Kanji.Group[cursor.getCount()];

			final TreeSet<Integer> radCodePoints = new TreeSet<Integer>();
			do {
				final int id = cursor.getInt(idxColId);
				final int parent = cursor.getInt(idxColParent);

				Kanji.Group group = new Kanji.Group(id);
				group.element = cursor.getInt(idxColElement);
				group.position = cursor.isNull(idxColPosition) ? null
						: KanjiDB.positions[cursor.getInt(idxColPosition)];
				group.radical = cursor.isNull(idxColRadical) ? null : KanjiDB.radicals[cursor.getInt(idxColRadical)];
				group.original = cursor.getInt(idxColOriginal);

				if (null != group.element)
					radCodePoints.add(group.element);
				if (null != group.original)
					radCodePoints.add(group.original);

				if (-1 != parent) {
					// parent==-1 => Root element
					if (null == groups[parent].elements) {
						groups[parent].elements = new ArrayList<Kanji.Element>(1);
					}
					groups[parent].elements.add(group);
				}
				groups[id] = group;

			} while (cursor.moveToNext());

			kanji.group = groups[0];
			kanji.radicals = new ArrayList<Integer>(radCodePoints.size());
			kanji.radicals.addAll(radCodePoints);
		}

		cursor.close();

		// Path
		// codepoint INTEGER NOT NULL, parent INTEGER NOT NULL, type TEXT, draw
		// TEXT NOT NULL
		cursor = db.query(TABLE_KVG_PATH, COLS_KVG_PATH, WHERE_CODEPOINT, selectionArgs, null, null, ORDER_BY_PARENT);
		if (cursor.moveToFirst()) {
			int idxColParent = cursor.getColumnIndex(COLS_KVG_PATH[0]);
			int idxColType = cursor.getColumnIndex(COLS_KVG_PATH[1]);
			int idxColDraw = cursor.getColumnIndex(COLS_KVG_PATH[2]);

			if (null == kanji.group) {
				kanji.group = new Kanji.Group(0);
			}
			if (null == groups) {
				groups = new Kanji.Group[1];
				groups[0] = kanji.group;
			}
			final int numGroups = groups.length;
			int id = 1;

			do {
				int parent = cursor.getInt(idxColParent);
				String type = cursor.getString(idxColType);
				String draw = cursor.getString(idxColDraw);

				if (parent < numGroups) {
					Kanji.Path path = new Kanji.Path(id, type, draw);
					if (null == groups[parent].elements) {
						groups[parent].elements = new ArrayList<Kanji.Element>(1);
					}
					groups[parent].elements.add(path);
				} else {
					Log.e(TAG, "Error in parenting path for kanji " + kanji.codepoint + ", no parent group id "
							+ parent);
				}

			} while (cursor.moveToNext());
		}
		cursor.close();

		kanji.kanjiDBLoadSVG = true;
		kanji.kanjiDBLoadRadicals = true;
	}

	private static Kanji cursorToKanji(SQLiteDatabase db, Cursor cursor, Kanji kanji, boolean loadMeaning,
			boolean loadSVG, boolean loadReading) { // boolean loadDictRef
		if (cursor.isClosed() || cursor.isAfterLast() || cursor.isBeforeFirst()) {
			return null;
		}

		Kanji lKanji = kanji;

		if (null == lKanji || !lKanji.kanjiDBLoadKanji) {
			int idxColCodepoint = cursor.getColumnIndex(COL_CODEPOINT);
			int idxColGrade = cursor.getColumnIndex(COL_KANJI_GRADE);
			int idxColFrequency = cursor.getColumnIndex(COL_KANJI_FREQUENCY);
			int idxColJLPT = cursor.getColumnIndex(COL_KANJI_JLPT);
			int idxColStrokeCount = cursor.getColumnIndex(COL_KANJI_STROKE_COUNT);
			int idxColReadingOn = cursor.getColumnIndex(COL_KANJI_READING_ON);
			int idxColReadingKun = cursor.getColumnIndex(COL_KANJI_READING_KUN);

			final int codepoint = cursor.getInt(idxColCodepoint);
			if (null == lKanji) {
				lKanji = Kanji.getKanji(codepoint);
			} else if (codepoint != lKanji.codepoint) {
				Log.w(TAG, "Codepoint mismatch cursor: " + codepoint + ", kanji: " + lKanji.codepoint);
				return null;
			}

			if (!cursor.isNull(idxColGrade))
				lKanji.grade = cursor.getInt(idxColGrade);

			if (!cursor.isNull(idxColFrequency))
				lKanji.frequency = cursor.getInt(idxColFrequency);

			if (!cursor.isNull(idxColJLPT))
				lKanji.jlpt = cursor.getInt(idxColJLPT);

			if (!cursor.isNull(idxColStrokeCount))
				lKanji.strokeCount = cursor.getInt(idxColStrokeCount);

			if (loadReading && idxColReadingKun != -1 && idxColReadingOn != -1) {
				lKanji.readingsOn = cursor.getString(idxColReadingOn);
				lKanji.readingsKun = cursor.getString(idxColReadingKun);
			}

			lKanji.kanjiDBLoadKanji = true;
		}

		if (loadMeaning && !lKanji.kanjiDBLoadMeaning) {
			KanjiDB.getMeanings(lKanji, db);
		}
		if (loadSVG && !lKanji.kanjiDBLoadSVG) {
			KanjiDB.getKanjiVG(lKanji, db);
		}

		// TODO LoadDictRef
		lKanji.kanjiDBLoadDictRef = true;

		return lKanji;
	}

	/**
	 * Iterator on a kanji db
	 * 
	 */
	public static class KanjiDBIterator implements Iterator<Kanji> {
		private final Cursor cursor;
		private final SQLiteDatabase db;
		private int current;
		private int count;
		private boolean loadMeaning;
		private boolean loadSVG;
		private boolean loadReading;

		private KanjiDBIterator(SQLiteDatabase db, Cursor cursor, boolean loadMeaning, boolean loadSVG,
				boolean loadReading) {
			this.db = db;
			this.cursor = cursor;
			this.current = 0;
			this.count = cursor.getCount();
			this.loadMeaning = loadMeaning;
			this.loadSVG = loadSVG;
			this.loadReading = loadReading;
		}

		/**
		 * 
		 * @return The current kanji number or -1 if this iterator is closed or
		 *         0 the first() method was not called.
		 */
		public int getCurrent() {
			return this.current;
		}

		/**
		 * 
		 * @return The number of resuls in the query.
		 */
		public int getCount() {
			return this.count;
		}

		/**
		 * 
		 * @return the first kanji, or null if there is no kanji in the list.
		 */
		public Kanji first() {
			Kanji kanji = null;
			if (this.cursor.moveToFirst()) {
				kanji = KanjiDB.cursorToKanji(db, cursor, null, this.loadMeaning, this.loadSVG, this.loadReading);
				this.current = 0;
			}
			return kanji;
		}

		/**
		 * 
		 * @return The next kanji.
		 */
		public Kanji next() {
			Kanji kanji = null;
			if (!this.cursor.isClosed() && this.cursor.moveToNext()) {
				kanji = KanjiDB.cursorToKanji(db, cursor, null, this.loadMeaning, this.loadSVG, this.loadReading);
				this.current++;
			}
			return kanji;
		}

		/**
		 * 
		 * @return The previous kanji
		 */
		public Kanji previous() {
			Kanji kanji = null;
			if (this.cursor.moveToPrevious()) {
				kanji = KanjiDB.cursorToKanji(db, cursor, null, this.loadMeaning, this.loadSVG, this.loadReading);
				this.current--;
			}
			return kanji;
		}

		/**
		 * 
		 * @return The last kanji
		 */
		public Kanji last() {
			Kanji kanji = null;
			if (this.cursor.moveToLast()) {
				kanji = KanjiDB.cursorToKanji(db, cursor, null, this.loadMeaning, this.loadSVG, this.loadReading);
				this.current = this.count;
			}
			return kanji;
		}

		/**
		 * Move the the iterator to the given position
		 * 
		 * @param newPosition
		 *            The new position (>=0 && < count)
		 * @return The kanji, or null if the cursor is close or the new position
		 *         is out of range.
		 */
		public Kanji position(int newPosition) {
			Kanji kanji = null;
			if (!this.cursor.isClosed() && newPosition >= 0 && newPosition < this.count) {
				if (this.cursor.moveToPosition(newPosition)) {
					this.current = newPosition;
					kanji = KanjiDB.cursorToKanji(db, cursor, null, this.loadMeaning, this.loadSVG, this.loadReading);
				}
			}
			return kanji;
		}

		/**
		 * Get the specific kanji at index position
		 * 
		 * @param index
		 *            The kanji position
		 * @return the kanji at index position or null if index is invalid.
		 */
		public Kanji get(int index) {
			Kanji kanji = null;
			if (this.cursor.moveToPosition(index)) {
				kanji = KanjiDB.cursorToKanji(db, cursor, null, this.loadMeaning, this.loadSVG, this.loadReading);
				this.cursor.moveToPosition(this.current);
			}
			return kanji;
		}

		/**
		 * Must be call, to close the db cursor and the database.
		 */
		public void close() {
			this.cursor.close();
			this.db.close();
			this.current = -1;
			this.count = 0;
		}

		public boolean hasNext() {
			return !this.cursor.isClosed() && !this.cursor.isLast();
		}

		public boolean hasPrevious() {
			return !this.cursor.isClosed() && !this.cursor.isFirst();
		}

		/**
		 * throw UnsupportedOperationException.
		 */
		public void remove() {
			throw new UnsupportedOperationException("Can't remove a Kanji from the database");

		}
	}

	/**
	 * Iterator on codepoint which have an SVG path available.
	 * 
	 */
	public static class KanjiSVGIterator implements Iterator<Integer> {
		private int codepoint;

		private final Cursor cursor;
		private final SQLiteDatabase db;
		private final int count;
		private int current;

		private KanjiSVGIterator(SQLiteDatabase db, Cursor cursor) {
			this.db = db;
			this.cursor = cursor;
			this.count = cursor.getCount();
			this.codepoint = -1;
			this.current = 0;
		}

		/**
		 * 
		 * @return The number of resuls in the query.
		 */
		public int getCount() {
			return this.count;
		}

		/**
		 * 
		 * @return The current kanji number or -1 if this iterator is closed or
		 *         0 the first() method was not called.
		 */
		public int getCurrent() {
			return this.current;
		}

		public boolean hasNext() {
			return !this.cursor.isClosed() && !this.cursor.isLast();
		}

		/**
		 * Get the next kanji codepoint.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer next() {
			this.codepoint = -1;
			if (!this.cursor.isAfterLast() && !this.cursor.isClosed()) {
				if (cursor.moveToNext()) {
					this.codepoint = cursor.getInt(0);
					this.current++;
				}
			}
			return this.codepoint;
		}

		/**
		 * Get the previous kanji codepoint.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer previous() {
			this.codepoint = -1;
			if (!this.cursor.isBeforeFirst() && !this.cursor.isClosed()) {
				if (cursor.moveToPrevious()) {
					this.codepoint = cursor.getInt(0);
					this.current--;
				}
			}
			return this.codepoint;
		}

		/**
		 * Get the first kanji.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer first() {
			this.codepoint = -1;
			if (!this.cursor.isClosed()) {
				if (cursor.moveToFirst()) {
					this.codepoint = cursor.getInt(0);
					this.current = 0;
				}
			}
			return this.codepoint;
		}

		/**
		 * Get the last kanji.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer last() {
			this.codepoint = -1;
			if (!this.cursor.isClosed()) {
				if (cursor.moveToLast()) {
					this.codepoint = cursor.getInt(0);
					this.current = this.count;
				}
			}
			return this.codepoint;
		}

		/**
		 * Move the the iterator to the given position
		 * 
		 * @param newPosition
		 *            The new position (>=0 && < count)
		 * @return The kanji codepoint, or -1 if the cursor is close or the new
		 *         position is out of range.
		 */
		public Integer position(int newPosition) {
			this.codepoint = -1;
			if (!this.cursor.isClosed() && newPosition >= 0 && newPosition < this.count) {
				if (this.cursor.moveToPosition(newPosition)) {
					this.current = newPosition;
					this.codepoint = cursor.getInt(0);
				}
			}
			return this.codepoint;
		}

		/**
		 * Must be call, to close the db cursor and the database.
		 */
		public void close() {
			this.cursor.close();
			this.db.close();
			this.current = -1;
			this.codepoint = -1;
		}

		/**
		 * UnsupportedOperationException.
		 */
		public void remove() {
			throw new UnsupportedOperationException("Can't remove a Kanji from the database");

		}

	}
}
