/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.kanji;

import java.util.ArrayList;
import java.util.Locale;

import org.vono.narau.R;
import org.vono.narau.preferences.Preferences;

import android.content.Context;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.DashPathEffect;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.PathMeasure;
import android.graphics.PointF;
import android.util.AttributeSet;
import android.util.Log;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.View;

/**
 * Class to draw kanji, and allow the user to draw them too. TODO Implement
 * mirror view.
 */
public class KanjiView extends View {
	private static final String TAG = KanjiView.class.getSimpleName();
	private static final int K_SIZE = 109;

	private enum DrawState {
		NONE, STATIC, PLAY
	}

	private final HumanDrawing userDrawing;
	private final SVGDrawing svgDrawing;

	private int currentStroke;
	private int maxStrokes;
	private DrawState drawState;
	private boolean showUserDrawing;

	public KanjiView(Context context) {
		this(context, null, 0);
	}

	public KanjiView(Context context, AttributeSet attrs) {
		this(context, attrs, 0);
	}

	public KanjiView(Context context, AttributeSet attrs, int defStyle) {
		super(context, attrs, defStyle);
		super.setFocusable(true);

		this.showUserDrawing = attrs.getAttributeBooleanValue(null, "showUserDrawing", false);
		this.drawState = DrawState.NONE;
		this.svgDrawing = new SVGDrawing();
		this.userDrawing = new HumanDrawing();
	}

	public void showKanji() {
		if (this.svgDrawing != null) {
			this.drawState = DrawState.STATIC;
		} else {
			this.drawState = DrawState.NONE;
		}
		super.invalidate();
	}

	public void setShowGrid(boolean visible) {
		this.svgDrawing.showGrid = visible;
		this.userDrawing.showGrid = visible;
		super.invalidate();
	}

	public void setShowStrokesNumbers(boolean visible) {
		this.svgDrawing.showNumbers = visible;
		this.userDrawing.showNumbers = visible;
		super.invalidate();
	}

	public void setKanji(Kanji kanji) {
		if (null == kanji) {
			this.drawState = DrawState.NONE;
			this.svgDrawing.clear();
			this.userDrawing.clear();
		} else {
			this.currentStroke = this.maxStrokes = Integer.MAX_VALUE;
			this.drawState = DrawState.STATIC;
			this.svgDrawing.setKanji(kanji);
			this.userDrawing.clear();
			// Log.i(TAG, "set kanji " + (char) kanji.codepoint);
		}
		super.invalidate();
	}

	public void playKanji() {
		if (this.drawState != DrawState.PLAY) {
			this.drawState = DrawState.PLAY;
			this.currentStroke = 0;

			this.svgDrawing.startPlay();
			this.svgDrawing.setCurrentDrawingStroke(0);

			if (null != this.userDrawing) {
				this.userDrawing.startPlay();
				this.userDrawing.setCurrentDrawingStroke(0);
				this.maxStrokes = Math.max(this.svgDrawing.getNumStrokes(), this.userDrawing.getNumStrokes());
			} else {
				this.maxStrokes = this.svgDrawing.getNumStrokes();
			}
			super.invalidate();
		}
	}

	public void clearUserDrawing() {
		this.userDrawing.clear();
		if (this.drawState == DrawState.PLAY) {
			this.drawState = DrawState.STATIC;
		}
		super.invalidate();

	}

	@Override
	protected void onSizeChanged(int w, int h, int oldw, int oldh) {
		super.onSizeChanged(w, h, oldw, oldh);

		if (this.showUserDrawing) {
			int drX;
			int drY;
			int posSvgX = 0;
			int posSvgY = 0;
			int posHumanX;
			int posHumanY;
			if (w > h) {
				drX = w / 2;
				drY = h;
				posHumanX = drX;
				posHumanY = 0;
			} else {
				drX = w;
				drY = h / 2;
				posHumanX = 0;
				posHumanY = drY;
			}

			this.svgDrawing.onSizeChange(drX, drY, posSvgX, posSvgY);
			this.userDrawing.onSizeChange(drX, drY, posHumanX, posHumanY);
		} else {
			this.svgDrawing.onSizeChange(w, h, 0, 0);
		}
		super.invalidate();
	}

	@Override
	protected void onDraw(Canvas canvas) {
		// Log.d(TAG, "onDraw");
		canvas.drawColor(0);
		if (DrawState.PLAY == this.drawState && (this.currentStroke >= this.maxStrokes)) {
			this.drawState = DrawState.STATIC;
		}

		if (DrawState.PLAY == this.drawState) {
			this.svgDrawing.onDraw(canvas, this.drawState, this.currentStroke);
			if (this.showUserDrawing)
				this.userDrawing.onDraw(canvas, this.drawState, this.currentStroke);

			if (this.svgDrawing.isDrawStrokeEnded() && this.userDrawing.isDrawStrokeEnded()) {
				this.currentStroke++;
				this.svgDrawing.setCurrentDrawingStroke(this.currentStroke);
				if (this.showUserDrawing)
					this.userDrawing.setCurrentDrawingStroke(this.currentStroke);
			}

			super.invalidate();
		} else {
			this.svgDrawing.onDraw(canvas, this.drawState, Integer.MAX_VALUE);
			if (this.showUserDrawing)
				this.userDrawing.onDraw(canvas, this.drawState, Integer.MAX_VALUE);
		}
	}

	private KDrawable drawable = null;

	@Override
	public boolean onTouchEvent(MotionEvent event) {

		if (this.drawState == DrawState.PLAY) {
			// don't accept touch event while drawing a kanji
			return true;
		}
		float x = event.getX();
		float y = event.getY();

		switch (event.getAction()) {
		case MotionEvent.ACTION_DOWN:
			if (this.svgDrawing.isPointInDrawing(x, y)) {
				drawable = this.svgDrawing;
			} else if (this.userDrawing.isPointInDrawing(x, y)) {
				drawable = this.userDrawing;
			}
			if (null != this.drawable) {
				this.drawable.touchDown(x, y);
			}

			break;
		case MotionEvent.ACTION_MOVE:
			if (null != this.drawable) {
				this.drawable.touchMove(x, y);
			}
			break;
		case MotionEvent.ACTION_UP:
			if (null != this.drawable) {
				this.drawable.touchUp(x, y);
				this.drawable = null;
			}
			break;
		}
		super.invalidate();

		return true;
	}

	/**
	 * Remove the last users drawing path, if press the back button, and there
	 * is at least one path draw.
	 */
	@Override
	public boolean onKeyDown(int keyCode, KeyEvent event) {
		boolean ret = false;

		if (KeyEvent.KEYCODE_BACK == keyCode) {
			ret = this.userDrawing.removeLastPath();
			if (ret) {
				super.invalidate();
			}
		}
		if (!ret) {
			ret = super.onKeyDown(keyCode, event);
		}
		return ret;
	}

	private static abstract class KDrawable {
		final Path gridPath;
		final ArrayList<Path> paths;
		final ArrayList<PointF> strokesNumberPos;
		final Paint paint;
		final Matrix matrix;
		final PathMeasure pathMeasure;
		final float[] intervals;

		//
		float posX;
		float posY;
		float scale;
		/**
		 * The size without the scale
		 */
		float size;
		// Optional displays
		boolean showGrid;
		boolean showNumbers;

		float drawSpeed;

		public KDrawable() {
			this.gridPath = new Path();
			this.matrix = new Matrix();
			this.paths = new ArrayList<Path>(4);
			this.strokesNumberPos = new ArrayList<PointF>(2);
			this.pathMeasure = new PathMeasure();
			this.intervals = new float[] { 0f, 0f };
			this.drawSpeed = 1f;

			this.posX = this.posY = this.size = this.scale = 0f;
			this.showGrid = false;
			this.showNumbers = false;

			this.paint = new Paint();
			this.paint.setAntiAlias(true);
			this.paint.setDither(true);
			this.paint.setColor(0xFFFFFFFF);
			this.paint.setStyle(Paint.Style.STROKE);
			this.paint.setStrokeJoin(Paint.Join.ROUND);
			this.paint.setStrokeCap(Paint.Cap.ROUND);
			this.paint.setStrokeWidth(Preferences.getFloat(R.string.prefNameKanjiStrokesWidth));
		}

		/**
		 * 
		 * @param w
		 *            Width
		 * @param h
		 *            Height
		 * @param pX
		 *            new pos X
		 * @param pY
		 *            new pos Y
		 */
		public void onSizeChange(int w, int h, int pX, int pY) {
			this.posX = pX;
			this.posY = pY;
			this.scale = Math.min(w, h) / this.size;
			this.matrix.reset();
			this.matrix.setScale(this.scale, this.scale);
			this.matrix.postTranslate(pX, pY);

			this.clear();
			this.buildGridPath();
		}

		/**
		 * Clear path and strokesNumberPos.
		 */
		public void clear() {
			this.paths.clear();
			this.strokesNumberPos.clear();
		}

		public int getNumStrokes() {
			return this.paths.size();
		}

		public void onDraw(Canvas canvas, DrawState state, int strokeDrawing) {
			Paint lPaint = this.paint;

			if (this.showGrid && state != DrawState.NONE) {
				lPaint.setColor(Color.GRAY);
				lPaint.setStrokeWidth(0f);
				canvas.drawPath(this.gridPath, lPaint);
				lPaint.setColor(Color.WHITE);
				lPaint.setStrokeWidth(Preferences.getFloat(R.string.prefNameKanjiStrokesWidth));
			}

			ArrayList<Path> lPaths = this.paths;
			int count = lPaths.size();

			switch (state) {
			case NONE:
				break;
			case PLAY:
				if (strokeDrawing < count) {
					float[] lIntervals = this.intervals;

					// Draw the previous strokes
					for (int i = 0; i < strokeDrawing; i++) {
						canvas.drawPath(lPaths.get(i), lPaint);
					}

					// Draw the current stroke, partially
					lIntervals[0] += this.drawSpeed;
					lIntervals[1] -= this.drawSpeed;
					lPaint.setPathEffect(new DashPathEffect(lIntervals, 0f));
					canvas.drawPath(lPaths.get(strokeDrawing), lPaint);
					lPaint.setPathEffect(null);

					break;
				}
				//$FALL-THROUGH$
			case STATIC:
				for (int i = 0; i < count; i++) {
					canvas.drawPath(lPaths.get(i), lPaint);
				}
				break;
			}

			if (this.showNumbers && DrawState.STATIC == state
					&& Preferences.getBoolean(R.string.prefNameKanjiShowStrokesNumber)) {
				// Draw strokes numbers
				ArrayList<PointF> positions = this.strokesNumberPos;
				count = positions.size();

				for (int i = 0; i < count; i++) {
					PointF point = positions.get(i);
					String number = Integer.toString(i + 1);

					lPaint.setColor(0xFF5555FF);
					lPaint.setStrokeWidth(0f);
					canvas.drawText(number, point.x, point.y, lPaint);
				}
				lPaint.setColor(Color.WHITE);
				lPaint.setStrokeWidth(Preferences.getFloat(R.string.prefNameKanjiStrokesWidth));
			}
		}

		public boolean isDrawStrokeEnded() {
			return this.intervals[1] <= 0f;
		}

		public void setCurrentDrawingStroke(int value) {
			float[] lIntervals = this.intervals;
			ArrayList<Path> lPaths = this.paths;
			if (lIntervals[1] <= 0f && value < lPaths.size()) {
				// Start drawing the next stroke
				this.pathMeasure.setPath(lPaths.get(value), false);
				lIntervals[0] = 0;
				lIntervals[1] = this.pathMeasure.getLength() + this.drawSpeed;
			}
		}

		public void startPlay() {
			this.intervals[0] = this.intervals[1] = 0f;
			this.drawSpeed = Preferences.getFloat(R.string.prefNameKanjiDrawSpeed);
		}

		private void buildGridPath() {
			Path lGridPath = this.gridPath;
			lGridPath.reset();

			float x = this.posX;
			float y = this.posY;
			float _size = this.size * this.scale;

			lGridPath.moveTo(x, y);
			lGridPath.lineTo(x + _size, y);
			lGridPath.lineTo(x + _size, y + _size);
			lGridPath.lineTo(x, y + _size);
			lGridPath.lineTo(x, y);

			float midSize = _size / 2f;
			lGridPath.moveTo(x + midSize, y);
			lGridPath.lineTo(x + midSize, y + _size);
			lGridPath.moveTo(x, y + midSize);
			lGridPath.lineTo(x + _size, y + midSize);
		}

		// Touch event:
		public boolean isPointInDrawing(float x, float y) {
			boolean ret = (this.posX <= x) && (x <= this.posX + this.size * this.scale);
			ret = ret && ((this.posY <= y) && (y <= this.posY + this.size * this.scale));

			return ret;
		}

		public abstract void touchDown(float x, float y);

		public abstract void touchMove(float x, float y);

		public abstract void touchUp(float x, float y);

	}

	/**
	 * What the user do with his little fingers.
	 * 
	 */
	private static class HumanDrawing extends KDrawable {
		private float touchX;
		private float touchY;
		private float touchTolerance;
		private Path path;

		public HumanDrawing() {
			super();
			this.touchTolerance = this.touchX = this.touchY = 0f;
			this.path = null;
		}

		@Override
		public void onSizeChange(int w, int h, int pX, int pY) {
			super.size = Math.min(w, h);
			super.onSizeChange(w, h, pX, pY);
			this.touchTolerance = super.size / (2f * K_SIZE);
		}

		@Override
		public void touchDown(float x, float y) {
			this.touchX = x;
			this.touchY = y;
			this.path = new Path();
			this.paths.add(path);
			this.path.moveTo(x, y);
			super.strokesNumberPos.add(new PointF(x, y));
		}

		@Override
		public void touchMove(float x, float y) {
			float dx = Math.abs(x - this.touchX);
			float dy = Math.abs(y - this.touchY);

			if (dx > this.touchTolerance || dy > this.touchTolerance) {
				this.path.quadTo(this.touchX, this.touchY, (this.touchX + x) / 2f, (this.touchY + y) / 2f);

				this.touchX = x;
				this.touchY = y;
			}

		}

		@Override
		public void touchUp(float x, float y) {
			float dx = Math.abs(x - this.touchX);
			float dy = Math.abs(y - this.touchY);
			if (dx > 0 || dy > 0) {
				this.path.quadTo(this.touchX, this.touchY, (this.touchX + x) / 2f, (this.touchY + y) / 2f);
			}
			this.path = null;
		}

		public boolean removeLastPath() {
			int count = this.paths.size();
			if (count > 0) {
				this.paths.remove(count - 1);
				this.strokesNumberPos.remove(count - 1);
				return true;
			}
			return false;
		}
	}

	/**
	 * The kanji with the path representation.
	 * 
	 */
	private static class SVGDrawing extends KDrawable {
		final ArrayList<SVGStroke> strokes;
		private Kanji kanji;

		public SVGDrawing() {
			super();
			this.strokes = new ArrayList<KanjiView.SVGStroke>(2);
			this.kanji = null;
			super.size = K_SIZE;
		}

		public void setKanji(Kanji kanji) {
			this.clear();
			this.kanji = kanji;
			if (null != kanji && null != kanji.group) {
				this.buildStrokes(this.kanji.group);
				this.buildPaths();
			}

		}

		@Override
		public void clear() {
			super.clear();
			this.kanji = null;
			this.strokes.clear();
		}

		@Override
		public void onSizeChange(int w, int h, int pX, int pY) {
			Kanji _kanji = this.kanji; // not pretty
			super.onSizeChange(w, h, pX, pY);

			this.setKanji(_kanji);
		}

		private void buildStrokes(Kanji.Element kElem) {
			if (kElem instanceof Kanji.Group) {
				Kanji.Group kGroup = (Kanji.Group) kElem;

				if (null != kGroup.elements) {
					int numElem = kGroup.elements.size();
					for (int i = 0; i < numElem; i++) {
						this.buildStrokes(kGroup.elements.get(i));
					}
				}
			} else if (kElem instanceof Kanji.Path) {
				Kanji.Path kPath = (Kanji.Path) kElem;
				SVGStroke stroke = new SVGStroke(kPath.draw);
				this.strokes.add(stroke);
			}
		}

		private void buildPaths() {
			ArrayList<SVGStroke> lStrokes = this.strokes;
			int numStrokes = lStrokes.size();
			ArrayList<Path> lPaths = super.paths;

			for (int i = 0; i < numStrokes; i++) {
				ArrayList<SVGCommand> svgCommands = lStrokes.get(i).svgCommands;
				lPaths.add(this.buildPath(svgCommands));
			}
		}

		private Path buildPath(ArrayList<SVGCommand> svgCommands) {
			// c:780, M:222, s:212, C:198, S:49, m:2, z:2
			Path path = new Path();

			int numCmd = svgCommands.size();

			SVGCommand command = svgCommands.get(0);
			float firstX = command.params.get(0);
			float firstY = command.params.get(1);
			float lastX = firstX;
			float lastY = firstY;

			PointF point = new PointF(firstX * super.scale + super.posX, firstY * super.scale + super.posY - 3f);
			super.strokesNumberPos.add(point);

			for (int i = 0; i < numCmd; i++) {
				command = svgCommands.get(i);
				if (!command.valid()) {
					Log.w(TAG, "invalid command " + i + ": " + command.toString());
					continue;
				}
				ArrayList<Float> params = command.params;

				switch (command.cmd) {
				case 'c':
					float x1 = params.get(0) + lastX;
					float y1 = params.get(1) + lastY;
					float x2 = params.get(2) + lastX;
					float y2 = params.get(3) + lastY;
					float x = params.get(4) + lastX;
					float y = params.get(5) + lastY;
					path.cubicTo(x1, y1, x2, y2, x, y);
					lastX = x;
					lastY = y;
					break;
				case 'M':
					x = params.get(0);
					y = params.get(1);
					path.moveTo(x, y);
					lastX = x;
					lastY = y;
					break;
				case 's':
					x1 = lastX;
					y1 = lastY;
					x2 = params.get(0) + lastX;
					y2 = params.get(1) + lastY;
					x = params.get(2) + lastX;
					y = params.get(3) + lastY;
					path.cubicTo(x1, y1, x2, y2, x, y);
					lastX = x;
					lastY = y;
					break;
				case 'C':
					x1 = params.get(0);
					y1 = params.get(1);
					x2 = params.get(2);
					y2 = params.get(3);
					x = params.get(4);
					y = params.get(5);
					path.cubicTo(x1, y1, x2, y2, x, y);
					lastX = x;
					lastY = y;
					break;
				case 'S':
					x1 = lastX;
					y1 = lastY;
					x2 = params.get(0);
					y2 = params.get(1);
					x = params.get(2);
					y = params.get(3);
					path.cubicTo(x1, y1, x2, y2, x, y);
					lastX = x;
					lastY = y;
					break;
				case 'm':
					x = params.get(0) + lastX;
					y = params.get(1) + lastY;
					path.moveTo(x, y);
					lastX = x;
					lastY = y;
					break;
				case 'z':
				case 'Z':
					path.lineTo(firstX, firstY);
					break;
				}

			}

			path.transform(super.matrix);

			return path;
		}

		@Override
		public void touchDown(float x, float y) {
		}

		@Override
		public void touchMove(float x, float y) {
		}

		@Override
		public void touchUp(float x, float y) {
		}

	}

	private static class SVGStroke {
		final ArrayList<SVGCommand> svgCommands;

		// final String svgPath;

		SVGStroke(String svgPath) {
			// this.svgPath = svgPath;
			this.svgCommands = new ArrayList<KanjiView.SVGCommand>(2);
			this.buildSVGCommands(svgPath);
		}

		/**
		 * Parse a svg path, creating a list of SVGCommand. The float parsing is
		 * done here, since the float number we have are very simple.
		 * 
		 * @param svgPath
		 *            The svg path text
		 * @return a list of parsed command
		 */
		@SuppressWarnings("null")
		private void buildSVGCommands(String svgPath) {
			final ArrayList<SVGCommand> listCommands = this.svgCommands;
			final int lastPos = svgPath.length() - 1;
			ArrayList<Float> params = null;
			SVGCommand svgCmd = null;

			int intPart = 0;
			int intPartTen = 1;
			int floatPart = 0;
			int floatPartTen = 0;
			float sign = 1f;

			for (int i = 0; i <= lastPos; i++) {
				char c = svgPath.charAt(i);
				boolean isLetter = Character.isLetter(c);

				if (((intPartTen > 1 && c == '-') || c == ',' || (i > 0 && isLetter))) {
					// end of the number, start of the new one
					float f = intPart;
					if (0 != floatPartTen) {
						f += floatPart / (float) floatPartTen;
					}
					f *= sign;
					params.add(f);

					intPart = 0;
					intPartTen = 1;
					floatPart = 0;
					floatPartTen = 0;
					sign = 1f;
				}

				if (isLetter) {
					// A new command start
					params = new ArrayList<Float>(2);
					svgCmd = new SVGCommand(c, params);
					listCommands.add(svgCmd);
					continue;
				}

				// inside a number
				switch (c) {
				case '-':
					sign = -1f;
					break;
				case '.':
					floatPartTen = 1;
					break;
				case ',':
					break;
				default:
					int val = (c - '0');

					if (val < 0 || val > 9) {
						System.err.println("char error '" + c + "'");
					}

					if (0 == floatPartTen) {
						intPart = intPart * intPartTen + val;
						intPartTen *= 10;
					} else {
						floatPart = floatPart * floatPartTen + val;
						floatPartTen *= 10;
					}
				}

				if (i == lastPos) {
					// The last number
					float f = intPart;
					if (0 != floatPartTen) {
						f += floatPart / (float) floatPartTen;
					}
					f *= sign;

					params.add(f);
				}
			}
		}
	}

	/**
	 * Represent a simple command of the SVGPath.
	 * 
	 */
	private static class SVGCommand {
		final char cmd;
		final ArrayList<Float> params;
		int numSameCmd;

		SVGCommand(char cmd, ArrayList<Float> params) {
			this.cmd = cmd;
			this.params = params;
			this.numSameCmd = 1;
		}

		@Override
		public String toString() {
			StringBuilder sb = new StringBuilder();

			sb.append(cmd);
			int numF = this.params.size();

			for (int i = 0; i < numF; i++) {
				float val = this.params.get(i);
				if (i > 0 && 0 < val)
					sb.append(',');
				if (val == 0)
					sb.append('0');
				else
					sb.append(String.format(Locale.ENGLISH, "%1$.2f", val));
			}

			return sb.toString();
		}

		boolean valid() {
			boolean ret = true;
			int numParams = params.size();

			switch (this.cmd) {
			case 'c':
			case 'C':
				ret = numParams % 6 == 0;
				this.numSameCmd = numParams / 6;
				break;
			case 'm':
			case 'M':
				ret = numParams % 2 == 0;
				this.numSameCmd = numParams / 2;
				break;
			case 's':
			case 'S':
				ret = numParams % 4 == 0;
				this.numSameCmd = numParams / 4;
				break;
			case 'Z':
			case 'z':
				ret = numParams == 0;
				break;
			default:
				// You should never see me:
				Log.e(TAG, "Missing svg path parse command '" + this.cmd + "'");
				ret = false;
				break;
			}

			if (1 < this.numSameCmd) {
				Log.w(TAG, "numSameCmd > 1");
			}
			return ret;
		}
	}
}
