/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictdb;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Set;

import org.tmatesoft.sqljet.core.SqlJetEncoding;
import org.tmatesoft.sqljet.core.SqlJetException;
import org.tmatesoft.sqljet.core.SqlJetTransactionMode;
import org.tmatesoft.sqljet.core.table.ISqlJetCursor;
import org.tmatesoft.sqljet.core.table.ISqlJetTable;
import org.tmatesoft.sqljet.core.table.SqlJetDb;
import org.vono.narau.Common;
import org.vono.narau.Log;
import org.vono.narau.dictdb.entry.Entry;
import org.vono.narau.dictdb.entry.KanjiElement;
import org.vono.narau.dictdb.entry.ReadingElement;
import org.vono.narau.dictdb.entry.Sense;
import org.vono.narau.dictdb.entry.SenseGloss;

public class DictDBWriter {
	private static final String TAG = DictDBWriter.class.getSimpleName();

	private static final String DB_NAME_TEMPLATE = "dict_%1$s-%2$d.db";

	private static final int NUM_LANGS = 5;
	private static final int NUM_GLOSS_MAX = 177493;
	private static final int NUM_KANJI = 155422;
	private static final int NUM_READING = 177097;
	public static final String EMPTY_STRING = "";
	public static final String SLASH = "/";

	private static final int VALUE_DB_VERSION = 2;

	private static final String TABLE_DICT_LANG = "dict_lang";
	private static final String TABLE_KANJI_ELEM = "k_ele";
	private static final String TABLE_READING_ELEM = "r_ele";
	private static final String TABLE_SENSE = "sense";
	private static final String TABLE_GLOSS = "gloss";

	private static final String CREATE_TABLE_DICT_LANG = "CREATE TABLE " + TABLE_DICT_LANG
			+ " (lang TEXT PRIMARY KEY, nb_entries INTEGER, table_name TEXT);";
	private static final String CREATE_TABLE_KANJI_ELEM = "CREATE TABLE "
			+ TABLE_KANJI_ELEM
			+ " (id INTEGER, num INTEGER, keb TEXT NOT NULL, pri TEXT, inf TEXT) CONSTRAINT kele_pk PRIMARY_KEY (id,num);";
	private static final String CREATE_TABLE_READING_ELEM = "CREATE TABLE "
			+ TABLE_READING_ELEM
			+ " (id INTEGER , num INTEGER, reb TEXT NOT NULL, pri TEXT, inf TEXT) CONSTRAINT rele_pk PRIMARY_KEY (id,num);";
	private static final String CREATE_TABLE_SENSE = "CREATE TABLE "
			+ TABLE_SENSE
			+ " (id INTEGER, num INTEGER, stagk TEXT, stagr TEXT, pos TEXT, xref TEXT) CONSTRAINT sense_pk PRIMARY_KEY (id,num);";
	private static final String CREATE_TABLE_GLOSS = "CREATE TABLE " + TABLE_GLOSS
			+ " (id INTEGER, num INTEGER, gloss TEXT NOT NULL) CONSTRAINT gloss_pk PRIMARY_KEY (id,num);";

	private static final String CREATE_INDEX_ID = "CREATE INDEX idx_%1$s_id ON %1$s (id);";

	final private Dictionary dictionary;
	final private File dbDirectory;
	final HashMap<String, ArrayList<EntryComparable>> mapSortedLang;

	public DictDBWriter(Dictionary dict, File directory) {
		this.dictionary = dict;
		this.dbDirectory = directory;
		this.mapSortedLang = new HashMap<String, ArrayList<EntryComparable>>(NUM_LANGS);
	}

	public void createDatabases() throws SqlJetException, IOException {
		final SqlJetDb mainDB = this.createJpnDatabase();

		try {
			for (String lang : this.mapSortedLang.keySet()) {
				this.createLangDatabase(lang, mainDB);
			}
		} finally {
			mainDB.commit();
			mainDB.close();
		}
	}

	private SqlJetDb createJpnDatabase() throws SqlJetException, IOException {
		final File dbFile = new File(this.dbDirectory, getDBName(Common.LANG_JAPANESE));

		final SqlJetDb db = DictDBWriter.createDatabase(dbFile);

		try {
			DictDBWriter.createTablesJpn(db);

			final ArrayList<EntryComparable> readingElements;
			final ArrayList<EntryComparable> kanjiElements;

			readingElements = new ArrayList<EntryComparable>(NUM_READING);
			kanjiElements = new ArrayList<EntryComparable>(NUM_KANJI);

			ArrayList<Entry> entries = new ArrayList<Entry>(this.dictionary.getEntries());
			int numEntries = entries.size();

			db.beginTransaction(SqlJetTransactionMode.WRITE);
			for (int i = 0; i < numEntries; i++) {
				this.writeSense(db, entries.get(i), readingElements, kanjiElements);
			}
			db.commit();

			Log.i(TAG, "Sorting kanji elements...");
			Collections.sort(kanjiElements);
			int numKanjis = kanjiElements.size();
			db.beginTransaction(SqlJetTransactionMode.WRITE);
			ISqlJetTable table = db.getTable(TABLE_KANJI_ELEM);
			Log.i(TAG, "Writing kanji elements...");
			for (int i = 0; i < numKanjis; i++) {
				EntryComparable ec = kanjiElements.get(i);
				table.insert(ec.id, ec.num, ec.text, ec.pri, ec.inf);
			}
			db.commit();

			Log.i(TAG, "Sorting reading elements...");
			Collections.sort(readingElements);
			int numReading = readingElements.size();
			db.beginTransaction(SqlJetTransactionMode.WRITE);
			table = db.getTable(TABLE_READING_ELEM);
			Log.i(TAG, "Writing reading elements...");
			for (int i = 0; i < numReading; i++) {
				EntryComparable ec = readingElements.get(i);
				table.insert(ec.id, ec.num, ec.text, ec.pri, ec.inf);
			}
			db.commit();

		} finally {
			db.commit();

			db.createIndex(String.format(CREATE_INDEX_ID, TABLE_KANJI_ELEM));
			db.createIndex(String.format(CREATE_INDEX_ID, TABLE_READING_ELEM));
			db.createIndex(String.format(CREATE_INDEX_ID, TABLE_SENSE));

		}

		return db;
	}

	private void createLangDatabase(String lang, SqlJetDb mainDB) throws SqlJetException, IOException {

		final File dbFile = new File(this.dbDirectory, getDBName(lang));
		final SqlJetDb db = DictDBWriter.createDatabase(dbFile);

		try {
			DictDBWriter.createTablesLang(db, lang);

			db.beginTransaction(SqlJetTransactionMode.WRITE);

			final ArrayList<EntryComparable> alEC = this.mapSortedLang.get(lang);
			String langName = Common.getLangName(lang);
			Log.i(TAG, "Sorting " + langName + " glosses...");
			Collections.sort(alEC);
			final int numGloss = alEC.size();
			Log.i(TAG, "Writing " + numGloss + " " + langName + " glosses...");
			final ISqlJetTable tableGloss = db.getTable(TABLE_GLOSS);
			for (int i = 0; i < numGloss; i++) {
				EntryComparable sg = alEC.get(i);
				tableGloss.insert(sg.id, sg.num, sg.text);
			}
			db.commit();

			Log.i(TAG, "Create index on " + TABLE_GLOSS);
			db.createIndex(String.format(CREATE_INDEX_ID, TABLE_GLOSS));

			db.beginTransaction(SqlJetTransactionMode.READ_ONLY);
			final ISqlJetCursor cursor = tableGloss.open();
			cursor.last();
			long count = cursor.getRowIndex();
			final ISqlJetTable tableMainDictInfo = mainDB.getTable(TABLE_DICT_LANG);
			tableMainDictInfo.insert(lang, count, dbFile.getName());
			cursor.close();

			Log.i(TAG, String.format("%1$s: %2$d entries", langName == null ? lang : langName, count));

		} finally {
			db.commit();
			db.close();
		}
	}

	private static SqlJetDb createDatabase(File dbFile) throws SqlJetException, IOException {

		if (dbFile.isFile()) {
			Log.i(TAG, dbFile.getName() + " exists, deleting...");
			if (!dbFile.delete()) {
				final String msg = "can't delete " + dbFile.getName();
				throw new IOException(msg);
			}
		}

		final SqlJetDb db = new SqlJetDb(dbFile, true);
		db.open();
		db.getOptions().setAutovacuum(true);
		// utf16 ??? check phone speed and db size
		db.getOptions().setEncoding(SqlJetEncoding.UTF8);

		return db;
	}

	private static void createTablesJpn(SqlJetDb db) throws SqlJetException {
		Common.createTablesCommon(db, VALUE_DB_VERSION, Common.LANG_JAPANESE);

		db.createTable(CREATE_TABLE_DICT_LANG);
		db.createTable(CREATE_TABLE_KANJI_ELEM);
		db.createTable(CREATE_TABLE_READING_ELEM);
		db.createTable(CREATE_TABLE_SENSE);
	}

	private static void createTablesLang(SqlJetDb db, String lang) throws SqlJetException {
		Common.createTablesCommon(db, VALUE_DB_VERSION, lang);
		db.createTable(CREATE_TABLE_GLOSS);
	}

	private void writeSense(SqlJetDb db, Entry entry, ArrayList<EntryComparable> readingElements,
			ArrayList<EntryComparable> kanjiElements) throws SqlJetException {

		int id = entry.sequence;

		// Find kanji elements
		for (int i = 0; i < entry.k_ele.size(); i++) {
			KanjiElement k_ele = entry.k_ele.get(i);
			String pri = DictDBWriter.listToString(k_ele.ke_pri);
			String inf = DictDBWriter.listToString(k_ele.ke_inf);

			EntryComparable ec = new EntryComparable();
			ec.id = id;
			ec.num = i;
			ec.text = k_ele.keb;
			ec.pri = pri;
			ec.inf = inf;

			kanjiElements.add(ec);
		}

		// Find reading elements
		for (int i = 0; i < entry.r_ele.size(); i++) {
			ReadingElement r_ele = entry.r_ele.get(i);
			String pri = DictDBWriter.listToString(r_ele.re_pri);
			String inf = DictDBWriter.listToString(r_ele.re_inf);

			EntryComparable ec = new EntryComparable();
			ec.id = id;
			ec.num = i;
			ec.text = r_ele.reb;
			ec.pri = pri;
			ec.inf = inf;

			readingElements.add(ec);
		}

		// Write sense
		ISqlJetTable table = db.getTable(TABLE_SENSE);
		final ArrayList<Sense> senses = entry.senses;
		final int numSenses = senses.size();
		for (int senseNum = 0; senseNum < numSenses; senseNum++) {
			final Sense sense = senses.get(senseNum);
			String stagk = DictDBWriter.listToString(sense.stagk);
			String stagr = DictDBWriter.listToString(sense.stagr);
			String pos = DictDBWriter.listToString(sense.pos);
			String xref = DictDBWriter.listToString(sense.xref);

			table.insert(id, senseNum, stagk, stagr, pos, xref);

			// Fetch glosses, and put them into mapSortedLang, for sorting and
			// writing after
			HashMap<String, String> glossTextMap = new HashMap<String, String>();
			final ArrayList<SenseGloss> glosses = sense.glosses;
			final int numGlosses = glosses.size();
			for (int j = 0; j < numGlosses; j++) {
				final SenseGloss gloss = glosses.get(j);

				// Merge gloss from the same language into one string
				String glossText = null;
				if (glossTextMap.containsKey(gloss.lang)) {
					glossText = glossTextMap.get(gloss.lang) + SLASH + gloss.text;
				} else {
					glossText = gloss.text;
				}
				glossTextMap.put(gloss.lang, glossText);
			}

			// Put the glosses translation for this sense
			// into the global glosses map.
			final Set<String> kSet = glossTextMap.keySet();
			for (String lang : kSet) {
				final EntryComparable ec = new EntryComparable();
				ec.text = glossTextMap.get(lang);
				ec.id = id;
				ec.num = senseNum;

				final String fixLang = Common.fixLang(lang);
				ArrayList<EntryComparable> alEC = this.mapSortedLang.get(fixLang);
				if (null == alEC) {
					Log.d(TAG, "new gloss lang " + lang);
					alEC = new ArrayList<EntryComparable>(NUM_GLOSS_MAX);
					mapSortedLang.put(fixLang, alEC);
				}
				alEC.add(ec);
			}
		}

	}

	private static String listToString(ArrayList<String> list) {
		if (null == list || 0 == list.size()) {
			return null;
		}

		String ret = list.get(0);
		for (int i = 1; i < list.size(); i++) {
			ret = ret + SLASH + list.get(i);
		}

		return ret;
	}

	private static class EntryComparable implements Comparable<EntryComparable> {
		private String text;
		private int id;
		private int num;
		private String pri;
		private String inf;

		@Override
		public int compareTo(EntryComparable ec) {
			int d = text.length() - ec.text.length();
			d = d > 0 ? 1 : d;
			d = d < 0 ? -1 : d;
			if (d == 0)
				d = text.compareTo(ec.text);

			return d;
		}
	}

	private static String getDBName(String lang) {
		return String.format(DB_NAME_TEMPLATE, lang, VALUE_DB_VERSION);
	}
}
