/*
 *  Copyright (C) 2012 Yvon TANGUY

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictdb;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

import org.vono.narau.Log;
import org.vono.narau.dictdb.entry.Entry;
import org.vono.narau.dictdb.entry.Info;
import org.vono.narau.dictdb.entry.KanjiElement;
import org.vono.narau.dictdb.entry.ReadingElement;
import org.vono.narau.dictdb.entry.Sense;
import org.vono.narau.dictdb.entry.SenseGloss;
import org.vono.narau.dictdb.entry.SenseLangSource;
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

/**
 * <p>
 * <a href="http://www.csse.monash.edu.au/~jwb/j_jmdict.html">[www.csse.monash.
 * edu.au]</a>
 * </p>
 * <p>
 * The JMdict (Japanese-Multilingual Dictionary) project has at its aim the
 * compilation of a multilingual lexical database with Japanese as the pivot
 * language. The project began in 1999 as an offshoot of the EDICT
 * Japanese-English Electronic Dictionary project. It involved a major rebuild
 * of the main files, with a more complex structure using XML.
 * 
 * The first release of the JMdict file occurred in May 1999, with occasional
 * releases after that. From mid-2006 the JMdict file has been generated daily
 * from the source database (along with the matching EDICT and EDICT2 files.)
 * The source database is updated almost every day. The version linked below
 * will always be the most recent version.
 * </p>
 * <p>
 * The JMdict file is copyright. It may be used quite freely subject to
 * conditions stated in the <a
 * href="http://www.edrdg.org/edrdg/licence.html">licence</a> statement.<br/>
 * The dictionary files are made available under a <a
 * href="https://creativecommons.org/licenses/by-sa/3.0/">Creative Commons
 * Attribution-ShareAlike Licence (V3.0)</a>.<br/>
 * </p>
 */
public class JMDictReader {
	private static final String TAG = JMDictReader.class.getSimpleName();

	public static final String[][] ENTITY_REPLACE = { { "MA", "martial arts term" },
			{ "X", "rude or X-rated term (not displayed in educational software)" }, { "abbr", "abbreviation" },
			{ "adj-i", "adjective (keiyoushi)" }, { "adj-na", "adjectival nouns or quasi-adjectives (keiyodoshi)" },
			{ "adj-no", "nouns which may take the genitive case particle `no'" },
			{ "adj-pn", "pre-noun adjectival (rentaishi)" }, { "adj-t", "`taru' adjective" },
			{ "adj-f", "noun or verb acting prenominally" },
			{ "adj", "former adjective classification (being removed)" }, { "adv", "adverb (fukushi)" },
			{ "adv-to", "adverb taking the `to' particle" }, { "anat", "anatomical thingos" }, { "arch", "archaism" },
			{ "ateji", "ateji (phonetic) reading" }, { "aux", "auxiliary" }, { "aux-v", "auxiliary verb" },
			{ "aux-adj", "auxiliary adjective" }, { "Buddh", "Buddhist term" }, { "chem", "chemistry term" },
			{ "chn", "children's language" }, { "col", "colloquialism" }, { "comp", "computer terminology" },
			{ "conj", "conjunction" }, { "ctr", "counter" }, { "derog", "derogatory" }, { "eK", "exclusively kanji" },
			{ "engr", "engineering term" }, { "ek", "exclusively kana" },
			{ "exp", "Expressions (phrases, clauses, etc.)" }, { "fam", "familiar language" },
			{ "fem", "female term or language" }, { "finc", "Finnish word" }, { "food", "food term" },
			{ "geom", "geometry term" },
			{ "gikun", "gikun (meaning as reading)  or jukujikun (special kanji reading)" },
			{ "hon", "honorific or respectful (sonkeigo) language" }, { "hum", "humble (kenjougo) language" },
			{ "iK", "word containing irregular kanji usage" }, { "id", "idiomatic expression" },
			{ "ik", "word containing irregular kana usage" }, { "int", "interjection (kandoushi)" },
			{ "io", "irregular okurigana usage" }, { "iv", "irregular verb" }, { "ling", "linguistics terminology" },
			{ "m-sl", "manga slang" }, { "male", "male term or language" }, { "male-sl", "male slang" },
			{ "math", "mathematics" }, { "mil", "military" }, { "n", "noun (common) (futsuumeishi)" },
			{ "n-adv", "adverbial noun (fukushitekimeishi)" }, { "n-suf", "noun, used as a suffix" },
			{ "n-pref", "noun, used as a prefix" }, { "n-t", "noun (temporal) (jisoumeishi)" }, { "num", "numeric" },
			{ "oik", "old or irregular kana form" }, { "oK", "word containing out-dated kanji" },
			{ "obs", "obsolete term" }, { "obsc", "obscure term" }, { "ok", "out-dated or obsolete kana usage" },
			{ "on-mim", "onomatopoeic or mimetic word" }, { "pn", "pronoun" }, { "poet", "poetical term" },
			{ "pol", "polite (teineigo) language" }, { "pref", "prefix" }, { "proverb", "proverb" },
			{ "prt", "particle" }, { "physics", "physics terminology" }, { "rare", "rare" }, { "sens", "sensitive" },
			{ "sl", "slang" }, { "suf", "suffix" }, { "uK", "word usually written using kanji alone" },
			{ "uk", "word usually written using kana alone" }, { "v1", "Ichidan verb" },
			{ "v2a-s", "Nidan verb with 'u' ending (archaic)" }, { "v4h", "Yodan verb with `hu/fu' ending (archaic)" },
			{ "v4r", "Yodan verb with `ru' ending (archaic)" }, { "v5", "Godan verb (not completely classified)" },
			{ "v5aru", "Godan verb - -aru special class" }, { "v5b", "Godan verb with `bu' ending" },
			{ "v5g", "Godan verb with `gu' ending" }, { "v5k", "Godan verb with `ku' ending" },
			{ "v5k-s", "Godan verb - Iku/Yuku special class" }, { "v5m", "Godan verb with `mu' ending" },
			{ "v5n", "Godan verb with `nu' ending" }, { "v5r", "Godan verb with `ru' ending" },
			{ "v5r-i", "Godan verb with `ru' ending (irregular verb)" }, { "v5s", "Godan verb with `su' ending" },
			{ "v5t", "Godan verb with `tsu' ending" }, { "v5u", "Godan verb with `u' ending" },
			{ "v5u-s", "Godan verb with `u' ending (special class)" },
			{ "v5uru", "Godan verb - Uru old class verb (old form of Eru)" },
			{ "vz", "Ichidan verb - zuru verb (alternative form of -jiru verbs)" }, { "vi", "intransitive verb" },
			{ "vk", "Kuru verb - special class" }, { "vn", "irregular nu verb" },
			{ "vr", "irregular ru verb, plain form ends with -ri" },
			{ "vs", "noun or participle which takes the aux. verb suru" },
			{ "vs-c", "su verb - precursor to the modern suru" }, { "vs-s", "suru verb - special class" },
			{ "vs-i", "suru verb - irregular" }, { "kyb", "Kyoto-ben" }, { "osb", "Osaka-ben" },
			{ "ksb", "Kansai-ben" }, { "ktb", "Kantou-ben" }, { "tsb", "Tosa-ben" }, { "thb", "Touhoku-ben" },
			{ "tsug", "Tsugaru-ben" }, { "kyu", "Kyuushuu-ben" }, { "rkb", "Ryuukyuu-ben" }, { "nab", "Nagano-ben" },
			{ "hob", "Hokkaido-ben" }, { "vt", "transitive verb" }, { "vulg", "vulgar expression or word" },
			{ "adj-kari", "`kari' adjective (archaic)" }, { "adj-ku", "`ku' adjective (archaic)" },
			{ "adj-shiku", "`shiku' adjective (archaic)" }, { "adj-nari", "archaic/formal form of na-adjective" },
			{ "n-pr", "proper noun" }, { "v-unspec", "verb unspecified" },
			{ "v4k", "Yodan verb with `ku' ending (archaic)" }, { "v4g", "Yodan verb with `gu' ending (archaic)" },
			{ "v4s", "Yodan verb with `su' ending (archaic)" }, { "v4t", "Yodan verb with `tsu' ending (archaic)" },
			{ "v4n", "Yodan verb with `nu' ending (archaic)" }, { "v4b", "Yodan verb with `bu' ending (archaic)" },
			{ "v4m", "Yodan verb with `mu' ending (archaic)" },
			{ "v2k-k", "Nidan verb (upper class) with `ku' ending (archaic)" },
			{ "v2g-k", "Nidan verb (upper class) with `gu' ending (archaic)" },
			{ "v2t-k", "Nidan verb (upper class) with `tsu' ending (archaic)" },
			{ "v2d-k", "Nidan verb (upper class) with `dzu' ending (archaic)" },
			{ "v2h-k", "Nidan verb (upper class) with `hu/fu' ending (archaic)" },
			{ "v2b-k", "Nidan verb (upper class) with `bu' ending (archaic)" },
			{ "v2m-k", "Nidan verb (upper class) with `mu' ending (archaic)" },
			{ "v2y-k", "Nidan verb (upper class) with `yu' ending (archaic)" },
			{ "v2r-k", "Nidan verb (upper class) with `ru' ending (archaic)" },
			{ "v2k-s", "Nidan verb (lower class) with `ku' ending (archaic)" },
			{ "v2g-s", "Nidan verb (lower class) with `gu' ending (archaic)" },
			{ "v2s-s", "Nidan verb (lower class) with `su' ending (archaic)" },
			{ "v2z-s", "Nidan verb (lower class) with `zu' ending (archaic)" },
			{ "v2t-s", "Nidan verb (lower class) with `tsu' ending (archaic)" },
			{ "v2d-s", "Nidan verb (lower class) with `dzu' ending (archaic)" },
			{ "v2n-s", "Nidan verb (lower class) with `nu' ending (archaic)" },
			{ "v2h-s", "Nidan verb (lower class) with `hu/fu' ending (archaic)" },
			{ "v2b-s", "Nidan verb (lower class) with `bu' ending (archaic)" },
			{ "v2m-s", "Nidan verb (lower class) with `mu' ending (archaic)" },
			{ "v2y-s", "Nidan verb (lower class) with `yu' ending (archaic)" },
			{ "v2r-s", "Nidan verb (lower class) with `ru' ending (archaic)" },
			{ "v2w-s", "Nidan verb (lower class) with `u' ending and `we' conjugation (archaic)" },
			{ "archit", "architecture term" }, { "astron", "astronomy, etc. term" }, { "baseb", "baseball term" },
			{ "biol", "biology term" }, { "bot", "botany term" }, { "bus", "business term" },
			{ "econ", "economics term" }, { "eng", "engineering term" }, { "fin", "finance term" },
			{ "geol", "geology, etc. term" }, { "law", "law, etc. term" }, { "med", "medicine, etc. term" },
			{ "music", "music term" }, { "Shinto", "Shinto term" }, { "sports", "sports term" },
			{ "sumo", "sumo term" }, { "zool", "zoology term" }, { "joc", "jocular, humorous term" }, };

	// private static final String NAMESPACE_XML = "xml";
	private static final String ELEM_ENTRY = "entry";
	private static final String ELEM_JMDICT = "JMdict";
	private static final String ELEM_SEQUENCE = "ent_seq";
	private static final String ELEM_KANJI = "k_ele";
	private static final String ELEM_KANJI_KEB = "keb";
	private static final String ELEM_KANJI_INF = "ke_inf";
	private static final String ELEM_KANJI_PRI = "ke_pri";
	private static final String ELEM_READING = "r_ele";
	private static final String ELEM_READING_REB = "reb";
	private static final String ELEM_READING_NOK = "re_nokanji";
	private static final String ELEM_READING_RES = "re_restr";
	private static final String ELEM_READING_INF = "re_inf";
	private static final String ELEM_READING_PRI = "re_pri";
	private static final String ELEM_INFO = "info";
	private static final String ELEM_INFO_LINKS = "links";
	private static final String ELEM_INFO_LINK_TAG = "link_tag";
	private static final String ELEM_INFO_LINK_DESC = "link_desc";
	private static final String ELEM_INFO_LINK_URI = "link_uri";
	private static final String ELEM_INFO_BIBL = "bibl";
	private static final String ELEM_INFO_BIBL_TAG = "bib_tag";
	private static final String ELEM_INFO_BIBL_TXT = "bib_txt";
	private static final String ELEM_INFO_ETYM = "etym";
	private static final String ELEM_INFO_AUDIT = "audit";
	private static final String ELEM_INFO_AUDIT_DATE = "upd_date";
	private static final String ELEM_INFO_AUDIT_DETL = "upd_detl";
	private static final String ELEM_SENSE = "sense";
	private static final String ELEM_SENSE_STAGK = "stagk";
	private static final String ELEM_SENSE_STAGR = "stagr";
	private static final String ELEM_SENSE_POS = "pos";
	private static final String ELEM_SENSE_XREF = "xref";
	private static final String ELEM_SENSE_ANT = "ant";
	private static final String ELEM_SENSE_FIELD = "field";
	private static final String ELEM_SENSE_MISC = "misc";
	private static final String ELEM_SENSE_SINF = "s_inf";
	private static final String ELEM_SENSE_LSOURCE = "lsource";
	private static final String ELEM_SENSE_LSOURCE_WASEI = "ls_wasei";
	private static final String ELEM_SENSE_LSOURCE_LANG = "lang";
	private static final String ELEM_SENSE_LSOURCE_TYPE = "type";
	private static final String ELEM_SENSE_DIAL = "dial";
	private static final String ELEM_SENSE_GLOSS = "gloss";
	private static final String ELEM_SENSE_GLOSS_LANG = "lang";
	private static final String ELEM_SENSE_GLOSS_GEND = "g_gend";
	private static final String ELEM_SENSE_EXAMPLE = "example";

	private static final String EXPECT_TEXT = "TEXT";

	private File file;
	private XmlPullParser xmlPullParser;

	int[] holderForStartAndLength;
	private int errors;
	private Dictionary dictionary;

	public JMDictReader(File jmdictFile, Dictionary dictionary) {
		this.file = jmdictFile;
		this.dictionary = dictionary;
		this.holderForStartAndLength = new int[2];
		this.errors = 0;
	}

	public void read() throws XmlPullParserException, IOException {
		Log.i(TAG, "start reading " + this.file.getName());

		XmlPullParserFactory factory = XmlPullParserFactory.newInstance();
		factory.setNamespaceAware(true);
		XmlPullParser xpp = factory.newPullParser();
		this.xmlPullParser = xpp;
		xpp.setInput(new FileReader(this.file));

		for (int i = 0; i < ENTITY_REPLACE.length; i++) {
			xpp.defineEntityReplacementText(ENTITY_REPLACE[i][0], ENTITY_REPLACE[i][0]);
		}
		int eventType = xpp.getEventType();
		String name = null;
		int count = 0;
		do {
			switch (eventType) {
			case XmlPullParser.START_DOCUMENT:
				break;
			case XmlPullParser.END_DOCUMENT:
				break;
			case XmlPullParser.START_TAG:
				name = xpp.getName();
				if (ELEM_ENTRY.equals(name)) {
					count++;
					Entry entry = this.processEntryElement(xpp);
					if (null != entry) {
						this.dictionary.addEntry(entry);
					} else {
						this.errors++;
					}
				} else if (ELEM_JMDICT.equals(name)) {
					// ok ignore
				} else {
					Log.e(TAG, "Error line " + xpp.getLineNumber() + " not an entry elem " + name);
				}
				break;
			case XmlPullParser.END_TAG:
				name = xpp.getName();
				if (!ELEM_JMDICT.equals(name)) {
					JMDictReader.displayExpectedError(ELEM_JMDICT, name, xpp.getLineNumber());
					JMDictReader.loopUntilEndTag(xpp, ELEM_JMDICT);
				}
				break;
			case XmlPullParser.TEXT:
				break;
			}
			eventType = xpp.next();
		} while (XmlPullParser.END_DOCUMENT != eventType);

		Log.i(TAG, "Readed " + count + " entries, " + this.errors + " error(s)");
	}

	/**
	 * entry (ent_seq, k_ele*, r_ele+, info?, sense+)
	 * 
	 * @param xpp
	 * @throws XmlPullParserException
	 * @throws IOException
	 */
	private Entry processEntryElement(XmlPullParser xpp) throws XmlPullParserException, IOException {
		Entry entry = null;

		int eventType = xpp.next();

		READ: {
			// ent_seq
			String str = this.readNextTextTag(xpp, ELEM_SEQUENCE, true);

			if (null == str) {
				entry = null;
				break READ;
			}
			try {
				entry = new Entry(Integer.parseInt(str));
			} catch (NumberFormatException nfe) {
				nfe.printStackTrace();
				entry = null;
				break READ;
			}

			// k_ele*
			eventType = xpp.getEventType();
			while (XmlPullParser.START_TAG == eventType && ELEM_KANJI.equals(xpp.getName())) {
				KanjiElement ke = this.readKanjiElem(xpp);
				if (null != ke) {
					entry.k_ele.add(ke);
				}
				eventType = JMDictReader.skipTextElem(xpp);
			}

			// r_ele+
			if (!(XmlPullParser.START_TAG == eventType && ELEM_READING.equals(xpp.getName()))) {
				JMDictReader.displayExpectedError(ELEM_READING, xpp.getName(), xpp.getLineNumber());
				entry = null;
				break READ;
			}
			do {
				ReadingElement re = this.readReadingElem(xpp);
				if (null != re) {
					entry.r_ele.add(re);
				} else {
					entry = null;
					break READ;
				}
				eventType = JMDictReader.skipTextElem(xpp);
			} while (XmlPullParser.START_TAG == eventType && ELEM_READING.equals(xpp.getName()));

			// info?
			if (XmlPullParser.START_TAG == eventType && ELEM_INFO.equals(xpp.getName())) {
				Info info = this.readInfoElem(xpp);
				if (null != info) {
					entry.info = info;
				}
				eventType = JMDictReader.skipTextElem(xpp);
			}

			// sense+
			if (!(XmlPullParser.START_TAG == eventType && ELEM_SENSE.equals(xpp.getName()))) {
				JMDictReader.displayExpectedError(ELEM_SENSE, xpp.getName(), xpp.getLineNumber());
				entry = null;
				break READ;
			}
			do {
				Sense sense = this.readSenseElem(xpp);
				if (null != sense) {
					entry.senses.add(sense);
				}
				eventType = JMDictReader.skipTextElem(xpp);
			} while (XmlPullParser.START_TAG == eventType && ELEM_SENSE.equals(xpp.getName()));
		}

		// check end tag
		if (!(XmlPullParser.END_TAG == eventType && ELEM_ENTRY.equals(xpp.getName()))) {
			JMDictReader.displayExpectedError(ELEM_ENTRY, xpp.getName(), xpp.getLineNumber());
			JMDictReader.loopUntilEndTag(xpp, ELEM_ENTRY);
			entry = null;
		}

		return entry;
	}

	/**
	 * sense (stagk*, stagr*, pos*, xref*, ant*, field*, misc*, s_inf*,
	 * lsource*, dial*, gloss*, example*)
	 * 
	 * @param xpp
	 * @return
	 * @throws XmlPullParserException
	 * @throws IOException
	 */
	private Sense readSenseElem(XmlPullParser xpp) throws XmlPullParserException, IOException {
		Sense sense = new Sense();
		xpp.next();

		READ: {
			// stagk*
			String stagk = this.readNextTextTag(xpp, ELEM_SENSE_STAGK, false);
			while (null != stagk) {
				if (null == sense.stagk) {
					sense.stagk = new ArrayList<String>(1);
				}
				sense.stagk.add(stagk);
				stagk = this.readNextTextTag(xpp, ELEM_SENSE_STAGK, false);
			}

			// stagr*
			String stagr = this.readNextTextTag(xpp, ELEM_SENSE_STAGR, false);
			while (null != stagr) {
				if (null == sense.stagr) {
					sense.stagr = new ArrayList<String>(1);
				}
				sense.stagr.add(stagr);
				stagr = this.readNextTextTag(xpp, ELEM_SENSE_STAGR, false);
			}

			// pos*
			String pos = this.readNextTextTag(xpp, ELEM_SENSE_POS, false);
			while (null != pos) {
				if (null == sense.pos) {
					sense.pos = new ArrayList<String>(1);
				}
				sense.pos.add(pos);
				pos = this.readNextTextTag(xpp, ELEM_SENSE_POS, false);
			}

			// xref*
			String xref = this.readNextTextTag(xpp, ELEM_SENSE_XREF, false);
			while (null != xref) {
				if (null == sense.xref) {
					sense.xref = new ArrayList<String>(1);
				}
				sense.xref.add(xref);
				xref = this.readNextTextTag(xpp, ELEM_SENSE_XREF, false);
			}

			// ant*
			String ant = this.readNextTextTag(xpp, ELEM_SENSE_ANT, false);
			while (null != ant) {
				if (null == sense.ant) {
					sense.ant = new ArrayList<String>(1);
				}
				sense.ant.add(ant);
				ant = this.readNextTextTag(xpp, ELEM_SENSE_ANT, false);
			}

			// field*
			String field = this.readNextTextTag(xpp, ELEM_SENSE_FIELD, false);
			while (null != field) {
				if (null == sense.field) {
					sense.field = new ArrayList<String>(1);
				}
				sense.field.add(field);
				field = this.readNextTextTag(xpp, ELEM_SENSE_FIELD, false);
			}

			// misc*
			String misc = this.readNextTextTag(xpp, ELEM_SENSE_MISC, false);
			while (null != misc) {
				if (null == sense.misc) {
					sense.misc = new ArrayList<String>(1);
				}
				sense.misc.add(misc);
				misc = this.readNextTextTag(xpp, ELEM_SENSE_MISC, false);
			}

			// s_inf*
			String s_inf = this.readNextTextTag(xpp, ELEM_SENSE_SINF, false);
			while (null != s_inf) {
				if (null == sense.s_inf) {
					sense.s_inf = new ArrayList<String>(1);
				}
				sense.s_inf.add(s_inf);
				s_inf = this.readNextTextTag(xpp, ELEM_SENSE_SINF, false);
			}

			// lsource*
			SenseLangSource lSource = this.readLSource(xpp);
			while (null != lSource) {
				if (null == sense.lsource) {
					sense.lsource = new ArrayList<SenseLangSource>();
				}
				sense.lsource.add(lSource);
				lSource = this.readLSource(xpp);
			}

			// dial*
			String dial = this.readNextTextTag(xpp, ELEM_SENSE_DIAL, false);
			while (null != dial) {
				if (null == sense.dial) {
					sense.dial = new ArrayList<String>(1);
				}
				sense.dial.add(dial);
				dial = this.readNextTextTag(xpp, ELEM_SENSE_DIAL, false);
			}

			// gloss*
			SenseGloss gloss = this.readGloss(xpp);
			while (null != gloss) {
				sense.glosses.add(gloss);
				gloss = this.readGloss(xpp);
			}

			// example*
			String example = this.readNextTextTag(xpp, ELEM_SENSE_EXAMPLE, false);
			while (null != example) {
				if (null == sense.examples) {
					sense.examples = new ArrayList<String>(1);
				}
				sense.examples.add(example);
				example = this.readNextTextTag(xpp, ELEM_SENSE_EXAMPLE, false);
			}

			int eventType = JMDictReader.skipTextElem(xpp);
			if (!(XmlPullParser.END_TAG == eventType && ELEM_SENSE.equals(xpp.getName()))) {
				JMDictReader.displayExpectedError("END " + ELEM_SENSE, xpp.getName(), xpp.getLineNumber());
				JMDictReader.loopUntilEndTag(xpp, ELEM_SENSE);
				sense = null;
				break READ;
			}
		}

		xpp.next();
		return sense;
	}

	/**
	 * gloss (#PCDATA | pri)*
	 * 
	 * @param xpp
	 * @return
	 * @throws XmlPullParserException
	 * @throws IOException
	 */
	private SenseGloss readGloss(XmlPullParser xpp) throws XmlPullParserException, IOException {
		SenseGloss gloss = null;
		int eventType = JMDictReader.skipTextElem(xpp);
		if (XmlPullParser.START_TAG == eventType && ELEM_SENSE_GLOSS.equals(xpp.getName())) {
			READ: {
				gloss = new SenseGloss();
				int attrCount = xpp.getAttributeCount();
				for (int i = 0; i < attrCount; i++) {
					String name = xpp.getAttributeName(i);
					if (ELEM_SENSE_GLOSS_LANG.equals(name)) {
						gloss.lang = xpp.getAttributeValue(i);
					} else if (ELEM_SENSE_GLOSS_GEND.equals(name)) {
						gloss.g_gend = xpp.getAttributeValue(i);
					} else {
						// error
						gloss = null;
						Log.e(TAG, "unknown attribute " + name + " for tag " + ELEM_SENSE_GLOSS);
						break READ;
					}
				}

				eventType = xpp.next();
				if (XmlPullParser.TEXT == eventType) {
					char[] ch = xpp.getTextCharacters(this.holderForStartAndLength);
					gloss.text = new String(ch, this.holderForStartAndLength[0], this.holderForStartAndLength[1]);
					eventType = xpp.next();
				}

				if (XmlPullParser.END_TAG != eventType && ELEM_SENSE_GLOSS.equals(xpp.getName())) {
					// error
					gloss = null;
					JMDictReader.displayExpectedError("END " + ELEM_SENSE_GLOSS, xpp.getName(), xpp.getLineNumber());
					break READ;
				}

			}

			if (null == gloss) {
				// found error, skip this tag
				JMDictReader.loopUntilEndTag(xpp, ELEM_SENSE_GLOSS);
			}
			xpp.next();

		}

		return gloss;
	}

	private SenseLangSource readLSource(XmlPullParser xpp) throws XmlPullParserException, IOException {
		SenseLangSource lSource = null;
		int eventType = JMDictReader.skipTextElem(xpp);
		if (XmlPullParser.START_TAG == eventType && ELEM_SENSE_LSOURCE.equals(xpp.getName())) {
			READ: {
				lSource = new SenseLangSource();

				int attrCount = xpp.getAttributeCount();
				for (int i = 0; i < attrCount; i++) {
					String name = xpp.getAttributeName(i);
					if (ELEM_SENSE_LSOURCE_LANG.equals(name)) {
						// lang first: 2107 occurrences in my JMdict file
						lSource.lang = xpp.getAttributeValue(i);
					} else if (ELEM_SENSE_LSOURCE_WASEI.equals(name)) {
						// wasei second: 799 occurrences in my JMdict file
						lSource.wasei = true;
					} else if (ELEM_SENSE_LSOURCE_TYPE.equals(name)) {
						// type last: no occurrence
						lSource.type = xpp.getAttributeValue(i);
					} else {
						// error
						lSource = null;
						Log.e(TAG, "unknown attribute " + name + " for tag " + ELEM_SENSE_LSOURCE);
						break READ;

					}
				}

				eventType = xpp.next();
				if (XmlPullParser.TEXT == eventType) {
					char[] ch = xpp.getTextCharacters(this.holderForStartAndLength);
					lSource.text = new String(ch, this.holderForStartAndLength[0], this.holderForStartAndLength[1]);
					eventType = xpp.next();
				}

				if (XmlPullParser.END_TAG != eventType && ELEM_SENSE_LSOURCE.equals(xpp.getName())) {
					// error
					lSource = null;
					JMDictReader.displayExpectedError("END " + ELEM_SENSE_LSOURCE, xpp.getName(), xpp.getLineNumber());
					break READ;
				}
			}

			if (null == lSource) {
				// found error, skip this tag
				JMDictReader.loopUntilEndTag(xpp, ELEM_SENSE_LSOURCE);
			}
			xpp.next();
		}

		return lSource;
	}

	/**
	 * links*, bibl*, etym*, audit*
	 * 
	 * @param xpp
	 * @return
	 * @throws XmlPullParserException
	 * @throws IOException
	 */
	private Info readInfoElem(XmlPullParser xpp) throws XmlPullParserException, IOException {
		Info info = new Info();

		xpp.next();
		int eventType = JMDictReader.skipTextElem(xpp);

		READ: {
			// links* : (link_tag, link_desc, link_uri)
			while (XmlPullParser.START_TAG == eventType && ELEM_INFO_LINKS.equals(xpp.getName())) {
				if (null == info.links) {
					info.links = new ArrayList<String[]>();
				}
				String[] link = new String[3];
				xpp.next();
				link[Info.LINK_TAG] = this.readNextTextTag(xpp, ELEM_INFO_LINK_TAG, true);
				link[Info.LINK_DESC] = this.readNextTextTag(xpp, ELEM_INFO_LINK_DESC, true);
				link[Info.LINK_URI] = this.readNextTextTag(xpp, ELEM_INFO_LINK_URI, true);

				eventType = xpp.getEventType();
				if (XmlPullParser.END_TAG == eventType && ELEM_INFO_LINKS.equals(xpp.getName())) {
					info.links.add(link);
					xpp.next();
					eventType = JMDictReader.skipTextElem(xpp);
				} else {
					info = null;
					break READ;
				}
			}

			// bibl* : (bib_tag?, bib_txt?)
			while (XmlPullParser.START_TAG == eventType && ELEM_INFO_BIBL.equals(xpp.getName())) {
				if (null == info.bibls) {
					info.bibls = new ArrayList<String[]>();
				}

				String[] bibl = new String[2];
				xpp.next();

				bibl[Info.BIBL_TAG] = this.readNextTextTag(xpp, ELEM_INFO_BIBL_TAG, false);
				bibl[Info.BIBL_TXT] = this.readNextTextTag(xpp, ELEM_INFO_BIBL_TXT, false);

				eventType = JMDictReader.skipTextElem(xpp);
				if (XmlPullParser.END_TAG == eventType && ELEM_INFO_BIBL.equals(xpp.getName())) {
					info.bibls.add(bibl);
					xpp.next();
					eventType = JMDictReader.skipTextElem(xpp);
				} else {
					info = null;
					break READ;
				}
			}

			// etym*
			while (XmlPullParser.START_TAG == eventType && ELEM_INFO_ETYM.equals(xpp.getName())) {
				if (null == info.etym) {
					info.etym = new ArrayList<String>(1);
				}
				xpp.next();
				String etym = this.readNextTextTag(xpp, ELEM_INFO_ETYM, true);
				if (null != etym) {
					info.etym.add(etym);
				}

				eventType = xpp.getEventType();
			}

			// audit* : (upd_date, upd_detl)
			while (XmlPullParser.START_TAG == eventType && ELEM_INFO_AUDIT.equals(xpp.getName())) {
				if (null == info.audits) {
					info.audits = new ArrayList<String[]>();
				}

				String[] audit = new String[2];
				xpp.next();
				audit[Info.AUDIT_DATE] = this.readNextTextTag(xpp, ELEM_INFO_AUDIT_DATE, true);
				if (null == audit[Info.AUDIT_DATE]) {
					info = null;
					break READ;
				}
				audit[Info.AUDIT_DETL] = this.readNextTextTag(xpp, ELEM_INFO_AUDIT_DETL, true);
				if (null == audit[Info.AUDIT_DETL]) {
					info = null;
					break READ;
				}

				eventType = xpp.getEventType();
				if (XmlPullParser.END_TAG == eventType && ELEM_INFO_AUDIT.equals(xpp.getName())) {
					info.audits.add(audit);
					xpp.next();
					eventType = JMDictReader.skipTextElem(xpp);
				} else {
					info = null;
					break READ;
				}

			}
		}

		eventType = xpp.getEventType();
		if (info == null || !(XmlPullParser.END_TAG == eventType && ELEM_INFO.equals(xpp.getName()))) {
			JMDictReader.displayExpectedError("END " + ELEM_INFO, xpp.getName(), xpp.getLineNumber());
			JMDictReader.loopUntilEndTag(xpp, ELEM_INFO);
			info = null;
		}

		xpp.next();
		return info;
	}

	/**
	 * r_ele (reb, re_nokanji?, re_restr*, re_inf*, re_pri*)
	 * 
	 * @param xpp
	 * @return
	 * @throws XmlPullParserException
	 * @throws IOException
	 */
	private ReadingElement readReadingElem(XmlPullParser xpp) throws XmlPullParserException, IOException {
		ReadingElement re = new ReadingElement();

		xpp.next();

		READ: {
			// reb
			re.reb = this.readNextTextTag(xpp, ELEM_READING_REB, true);
			if (null == re.reb) {
				re = null;
				break READ;
			}

			// re_nokanji?
			re.re_nokanji = this.readNextTextTag(xpp, ELEM_READING_NOK, false);

			// re_restr*
			String restr = this.readNextTextTag(xpp, ELEM_READING_RES, false);
			if (null != restr) {
				re.re_restr = new ArrayList<String>(1);
				do {
					re.re_restr.add(restr);
					restr = this.readNextTextTag(xpp, ELEM_READING_RES, false);
				} while (restr != null);
			}

			// re_inf*
			String inf = this.readNextTextTag(xpp, ELEM_READING_INF, false);
			if (null != inf) {
				re.re_inf = new ArrayList<String>(1);
				do {
					re.re_inf.add(inf);
					inf = this.readNextTextTag(xpp, ELEM_READING_INF, false);
				} while (inf != null);
			}

			// re_pri*
			String pri = this.readNextTextTag(xpp, ELEM_READING_PRI, false);
			if (null != pri) {
				re.re_pri = new ArrayList<String>(1);
				do {
					re.re_pri.add(pri);
					pri = this.readNextTextTag(xpp, ELEM_READING_PRI, false);
				} while (pri != null);
			}

		}

		int eventType = xpp.getEventType();
		if (!(null != re && XmlPullParser.END_TAG == eventType && ELEM_READING.equals(xpp.getName()))) {
			JMDictReader.loopUntilEndTag(xpp, ELEM_READING);

		}
		xpp.next();
		return re;
	}

	/**
	 * k_ele (keb, ke_inf*, ke_pri*)
	 * 
	 * @param xpp
	 * @throws XmlPullParserException
	 * @throws IOException
	 */
	private KanjiElement readKanjiElem(XmlPullParser xpp) throws XmlPullParserException, IOException {
		KanjiElement k_ele = new KanjiElement();
		xpp.next();

		READ: {
			// keb
			k_ele.keb = this.readNextTextTag(xpp, ELEM_KANJI_KEB, true);
			if (null == k_ele.keb) {
				k_ele = null;
				break READ;
			}

			// ke_inf
			String inf = null;
			do {
				inf = this.readNextTextTag(xpp, ELEM_KANJI_INF, false);
				if (null != inf) {
					if (null == k_ele.ke_inf) {
						k_ele.ke_inf = new ArrayList<String>(1);
					}
					k_ele.ke_inf.add(inf);
				}
			} while (null != inf);

			// ke_pri
			String pri = null;
			do {
				pri = this.readNextTextTag(xpp, ELEM_KANJI_PRI, false);
				if (null != pri) {
					if (null == k_ele.ke_pri) {
						k_ele.ke_pri = new ArrayList<String>(1);
					}
					k_ele.ke_pri.add(pri);
				}
			} while (null != pri);
		}

		int eventType = xpp.getEventType();
		if (XmlPullParser.END_TAG == eventType && ELEM_KANJI.equals(xpp.getName())) {
			xpp.next();
		} else {
			k_ele = null;
			JMDictReader.displayExpectedError("END " + ELEM_KANJI, xpp.getName(), xpp.getLineNumber());
			// this.loopUntilEndTag(xpp, ELEM_KANJI);
		}

		return k_ele;

	}

	/**
	 * Read a text tab: &lt;toto&gt;text&lt;/toto&gt;
	 * 
	 * @param xpp
	 * @param tagName
	 * @param mandatory
	 * @return
	 * @throws XmlPullParserException
	 * @throws IOException
	 */
	private String readNextTextTag(XmlPullParser xpp, String tagName, boolean mandatory) throws XmlPullParserException,
			IOException {
		String text = null;
		String expectError = null;
		READ: {
			int eventType = JMDictReader.skipTextElem(xpp);
			if (!(eventType == XmlPullParser.START_TAG && tagName.equals(xpp.getName()))) {
				expectError = mandatory ? tagName : null;
				break READ;
			}

			eventType = xpp.next();
			if (XmlPullParser.TEXT != eventType) {
				expectError = mandatory ? EXPECT_TEXT : null;
				break READ;
			}
			char[] ch = xpp.getTextCharacters(this.holderForStartAndLength);
			text = new String(ch, this.holderForStartAndLength[0], this.holderForStartAndLength[1]);

			eventType = xpp.next();
			if (!(XmlPullParser.END_TAG == eventType && tagName.equals(xpp.getName()))) {
				expectError = mandatory ? tagName : null;
				break READ;
			}

			eventType = xpp.next();
			eventType = JMDictReader.skipTextElem(xpp);
		}

		if (mandatory && null != expectError) {
			JMDictReader.displayExpectedError(expectError, xpp.getName(), xpp.getLineNumber());
			text = null;
		}
		return text;
	}

	private static void displayExpectedError(String expected, String found, int line) {
		String error = "Error:" + line + ":Expected element '" + expected + "' found '" + found + "'";
		Log.e(TAG, error);
	}

	/**
	 * To skip the '\n' text tag mostly
	 * 
	 * @param xpp
	 * @return
	 * @throws XmlPullParserException
	 * @throws IOException
	 */
	private static int skipTextElem(XmlPullParser xpp) throws XmlPullParserException, IOException {
		int evenType = xpp.getEventType();
		while (XmlPullParser.TEXT == evenType) {
			evenType = xpp.next();
		}

		return evenType;
	}

	/**
	 * Used when an error is found in one tag
	 * 
	 * @param xpp
	 * @param tagName
	 * @throws XmlPullParserException
	 * @throws IOException
	 */
	private static void loopUntilEndTag(XmlPullParser xpp, String tagName) throws XmlPullParserException, IOException {
		boolean loop = true;

		while (loop) {
			int evenType = xpp.next();
			switch (evenType) {
			case XmlPullParser.END_TAG:
				String name = xpp.getName();
				if (tagName.equals(name)) {
					loop = false;
				}
				break;
			case XmlPullParser.END_DOCUMENT:
				loop = false;
				break;
			}
		}
	}

	public int getLineNumber() {
		if (null != this.xmlPullParser) {
			return this.xmlPullParser.getLineNumber();
		}

		return 0;
	}

}
