/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictdb;

import java.io.File;
import java.io.IOException;

import org.tmatesoft.sqljet.core.SqlJetException;
import org.vono.narau.Common;
import org.vono.narau.Log;
import org.xmlpull.v1.XmlPullParserException;

/**
 * 
 * Read JMdict/Edict files, and create the sqlite databases files.
 */
public class Main {
	private static final String TAG = Main.class.getSimpleName();

	private static void usage() {
		System.out.println("usage:");
		System.out.println("{java cmd line} <db_creation_dir> <JMdict_file> [lang:edict_files]");
		System.out.println();
		System.out.println("Example:");
		System.out.println("{java cmd line} . JMDict es:hispadic.utf8");
		System.out.println();
		System.out.println("JMDict file direct download link: ftp://ftp.monash.edu.au/pub/nihongo/JMdict.gz");
		System.out.println("Spanish EDict website: http://hispadic.byethost3.com/index.php");
		System.out.println("Spanish EDict download: "
				+ "http://sites.google.com/site/hispadic/download/hispadic.zip?attredirects=0");
		System.out.println();
		System.out.println("Accepted langs code:");
		final String[][] langArray = Common.LANGS_ARRAY;
		final int numLangs = langArray.length;
		for (int i = 0; i < numLangs; i++) {
			final String[] ll = langArray[i];
			System.out.println(ll[1] + ": " + ll[2]);
		}
	}

	private static class EDictFile {
		String lang;
		File file;
	}

	public static void main(String[] args) {
		if (args.length < 2) {
			Main.usage();
			return;
		}

		File dbFilesDir = new File(args[0]);
		if (!dbFilesDir.isDirectory()) {
			Log.e(TAG, args[0] + " is not a directory");
			return;
		}

		File jmDictFile = new File(args[1]);
		if (!(jmDictFile.isFile() && jmDictFile.canRead())) {
			Log.e(TAG, args[1] + " can't be read");
			return;
		}

		EDictFile[] edDictFiles = new EDictFile[args.length - 2];
		if (args.length > 2) {

			for (int i = 2; i < args.length; i++) {
				EDictFile dictFile = new EDictFile();
				String[] arg = args[i].split(":");
				if (arg.length == 2) {
					String lang = arg[0];
					dictFile.file = new File(arg[1]);
					if (!dictFile.file.isFile()) {
						Log.e(TAG, "Invalid edict parameter: " + args[i]);
						Log.e(TAG, "File " + arg[1] + " do not exists");
						return;
					}

					dictFile.lang = Common.getLang3(lang);
					if (dictFile.lang.equals(lang)) {
						Log.e(TAG, "Invalid edict parameter: " + args[i]);
						Log.e(TAG, "Lang tag " + dictFile.lang + " is unknown");
						return;
					}

					edDictFiles[i - 2] = dictFile;
				} else {
					Log.e(TAG, "Invalid edict parameter: " + args[i]);
					return;
				}
			}
		}

		long startTime = System.currentTimeMillis();

		final Dictionary dictionary = new Dictionary();

		JMDictReader reader = new JMDictReader(jmDictFile, dictionary);

		try {
			reader.read();
		} catch (XmlPullParserException xppe) {
			Log.e(TAG, "line " + reader.getLineNumber(), xppe);
			return;
		} catch (IOException ioe) {
			Log.e(TAG, ioe.getMessage(), ioe);
			return;
		}

		EDictReader eDictReader = new EDictReader(dictionary);
		try {
			for (int i = 0; i < edDictFiles.length; i++) {
				final EDictFile edictFile = edDictFiles[i];
				eDictReader.read(edictFile.lang, edictFile.file);
			}
		} catch (IOException ioe) {
			Log.e(TAG, ioe.getMessage(), ioe);
			return;
		}

		DictDBWriter sqliteWriter = new DictDBWriter(dictionary, dbFilesDir);
		try {
			sqliteWriter.createDatabases();
		} catch (SqlJetException sje) {
			Log.e(TAG, sje.getMessage(), sje);
			return;
		} catch (IOException ioe) {
			Log.e(TAG, ioe.getMessage(), ioe);
			return;
		}

		long endTime = System.currentTimeMillis();
		long mili = (endTime - startTime) % 1000;
		long sec = (endTime - startTime) / 1000;
		long min = 0;
		min = sec / 60;
		sec = sec % 60;
		Log.i(TAG, "creation time: " + min + "m " + sec + "." + mili + "s");
	}
}
