/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.dictdb.entry;

import java.util.ArrayList;
import java.util.HashMap;

/**
 * <p>
 * sense (stagk*, stagr*, pos*, xref*, ant*, field*, misc*, s_inf*, lsource*,
 * dial*, gloss*, example*)
 * </p>
 * The sense element will record the translational equivalent of the Japanese
 * word, plus other related information. Where there are several distinctly
 * different meanings of the word, multiple sense elements will be employed.
 * 
 */
public class Sense {

	/**
	 * These elements, if present, indicate that the sense is restricted to the
	 * lexeme represented by the keb.
	 */
	public ArrayList<String> stagk;

	/**
	 * These elements, if present, indicate that the sense is restricted to the
	 * lexeme represented by the reb.
	 */
	public ArrayList<String> stagr;

	/**
	 * Part-of-speech information about the entry/sense. Should use appropriate
	 * entity codes. In general where there are multiple senses in an entry, the
	 * part-of-speech of an earlier sense will apply to later senses unless
	 * there is a new part-of-speech indicated.
	 */
	public ArrayList<String> pos;

	/**
	 * This element is used to indicate a cross-reference to another entry with
	 * a similar or related meaning or sense. The content of this element is
	 * typically a keb or reb element in another entry. In some cases a keb will
	 * be followed by a reb and/or a sense number to provide a precise target
	 * for the cross-reference. Where this happens, a JIS "centre-dot" (0x2126)
	 * is placed between the components of the cross-reference.
	 */
	public ArrayList<String> xref;

	/**
	 * This element is used to indicate another entry which is an antonym of the
	 * current entry/sense. The content of this element must exactly match that
	 * of a keb or reb element in another entry.
	 */
	public ArrayList<String> ant;

	/**
	 * Information about the field of application of the entry/sense. When
	 * absent, general application is implied. Entity coding for specific fields
	 * of application.
	 */
	public ArrayList<String> field;

	/**
	 * This element is used for other relevant information about the
	 * entry/sense. As with part-of-speech, information will usually apply to
	 * several senses.
	 */
	public ArrayList<String> misc;

	/**
	 * The sense-information elements provided for additional information to be
	 * recorded about a sense. Typical usage would be to indicate such things as
	 * level of currency of a sense, the regional variations, etc.
	 */
	public ArrayList<String> s_inf;

	/**
	 * This element records the information about the source language(s) of a
	 * loan-word/gairaigo. If the source language is other than English, the
	 * language is indicated by the xml:lang attribute. The element value (if
	 * any) is the source word or phrase.
	 */
	public ArrayList<SenseLangSource> lsource;

	/**
	 * For words specifically associated with regional dialects in Japanese, the
	 * entity code for that dialect, e.g. ksb for Kansaiben.
	 */
	public ArrayList<String> dial;

	/**
	 * Within each sense will be one or more "glosses", i.e. target-language
	 * words or phrases which are equivalents to the Japanese word. This element
	 * would normally be present, however it may be omitted in entries which are
	 * purely for a cross-reference.
	 */
	public final ArrayList<SenseGloss> glosses;

	/**
	 * The example elements provide for pairs of short Japanese and
	 * target-language phrases or sentences which exemplify the usage of the
	 * Japanese head-word and the target-language gloss. Words in example fields
	 * would typically not be indexed by a dictionary application.
	 */
	public ArrayList<String> examples;

	public Sense() {
		this.glosses = new ArrayList<SenseGloss>();
	}

	@Override
	public String toString() {
		// Useful when debugging
		final ArrayList<SenseGloss> lGlosses = this.glosses;
		final int nGlosses = lGlosses.size();
		final StringBuilder sb = new StringBuilder();
		final HashMap<String, Integer> glossPerLang = new HashMap<String, Integer>();
		for (int i = 0; i < nGlosses; i++) {
			SenseGloss sg = lGlosses.get(i);
			int val = 1;
			if (glossPerLang.containsKey(sg.lang)) {
				val = glossPerLang.get(sg.lang) + 1;
			}
			glossPerLang.put(sg.lang, val);
		}

		sb.append('[');
		for (String lang : glossPerLang.keySet()) {
			int val = glossPerLang.get(lang);
			sb.append(lang + ":" + val);
			sb.append(",");
		}
		sb.append(']');

		return sb.toString();
	}

}
