/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.kanjidb;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import org.vono.narau.Log;

/**
 * Informations on a kanji<br/>
 * Datas from kanjivg & kanjidict.
 */
public class Kanji implements Comparable<Kanji> {
	private static final String TAG = Kanji.class.getSimpleName();

	public static final String[] POSITIONS = { "bottom", "kamae", "kamae1", "kamae2", "left", "nyo", "right", "tare",
			"top" };
	public static final String[] RADICAL = { "general", "nelson", "tradit" };

	public int codepoint;
	public Group group;

	/**
	 * G&lt;num&gt; -- the "grade" of the kanji. At most one per line.
	 * <ul>
	 * <li>G1 to G6 indicates the grade level as specified by the Japanese
	 * Ministry of Education for kanji that are to be taught in elementary
	 * school (1006 Kanji). These are sometimes called the "Kyouiku" (education)
	 * kanji and are part of the set of Jouyou (daily use) kanji;</li>
	 * <li>G8 indicates the remaining Jouyou kanji that are to be taught in
	 * secondary school (additional 1130 Kanji);</li>
	 * <li>G9 and G10 indicate Jinmeiyou ("for use in names") kanji which in
	 * addition to the Jouyou kanji are approved for use in family name
	 * registers and other official documents. G9 (774 kanji, of which 628 are
	 * in KANJIDIC) indicates the kanji is a "regular" name kanji, and G10 (209
	 * kanji of which 128 are in KANJIDIC) indicates the kanji is a variant of a
	 * Jouyou kanji;</li>
	 * </ul>
	 */
	public Integer grade;

	/**
	 * F&lt;num&gt; -- the frequency-of-use ranking. At most one per line. The
	 * 2,501 most-used characters have a ranking; those characters that lack
	 * this field are not ranked. The frequency is a number from 1 to 2,501 that
	 * expresses the relative frequency of occurrence of a character in modern
	 * Japanese. The data is based on an analysis of word frequencies in the
	 * Mainichi Shimbun over 4 years by Alexandre Girardi. From this the
	 * relative frequencies have been derived. Note:
	 * <ul>
	 * <li>these frequencies are biassed towards words and kanji used in
	 * newspaper articles,</li>
	 * <li>the relative frequencies for the last few hundred kanji so graded is
	 * quite imprecise.</li>
	 * </ul>
	 * 
	 * (Earlier editions of the KANJIDIC file used a frequency-of-use ranking
	 * from the National Language Research Institute (Tokyo), interpreted and
	 * adapted by Jack Halpern.)
	 */
	public Integer frequency;

	/**
	 * J&lt;num&gt; -- the level of the Japanese Language Proficiency Test
	 * (JLPT) in which the kanji occurs. (1-4)
	 */
	public Integer jlpt;

	/**
	 * S&lt;num&gt; -- the stroke count. At least one per line. If more than
	 * one, the first is considered the accepted count.
	 */
	public Integer strokeCount;

	/**
	 * ja_on - the "on" Japanese reading of the kanji, in katakana. Another
	 * attribute r_status, if present, will indicate with a value of "jy"
	 * whether the reading is approved for a "Jouyou kanji". A further attribute
	 * on_type, if present, will indicate with a value of kan, go, tou or
	 * kan'you the type of on-reading.
	 */
	public String readingsOn;

	/**
	 * ja_kun - the "kun" Japanese reading of the kanji, usually in hiragana. <br/>
	 * Where relevant the okurigana is also included separated by a ".".
	 * Readings associated with prefixes and suffixes are marked with a "-". A
	 * second attribute r_status, if present, will indicate with a value of "jy"
	 * whether the reading is approved for a "Jouyou kanji".
	 */
	public String readingsKun;

	/**
	 * Meaning map: lang for key, meanings '|' separated.
	 */
	public HashMap<String, String> meanings;

	/**
	 * <p>
	 * This element contains the index numbers and similar unstructured
	 * information such as page numbers in a number of published dictionaries,
	 * and instructional books on kanji.
	 * </p>
	 * The dr_type defines the dictionary or reference book, etc. to which
	 * dic_ref element applies. The initial allocation is:<br/>
	 * <ul>
	 * <li>nelson_c - "Modern Reader's Japanese-English Character Dictionary",
	 * edited by Andrew Nelson (now published as the "Classic" Nelson).
	 * <li>nelson_n - "The New Nelson Japanese-English Character Dictionary",
	 * edited by John Haig.</li>
	 * <li>halpern_njecd - "New Japanese-English Character Dictionary", edited
	 * by Jack Halpern.</li>
	 * <li>halpern_kkld - "Kanji Learners Dictionary" (Kodansha) edited by Jack
	 * Halpern.</li>
	 * <li>heisig - "Remembering The  Kanji" by James Heisig.</li>
	 * <li>gakken - "A  New Dictionary of Kanji Usage" (Gakken)</li>
	 * <li>oneill_names - "Japanese Names", by P.G. O'Neill.</li>
	 * <li>oneill_kk - "Essential Kanji" by P.G. O'Neill.</li>
	 * <li>moro - "Daikanwajiten" compiled by Morohashi. For some kanji two
	 * additional attributes are used: m_vol: the volume of the dictionary in
	 * which the kanji is found, and m_page: the page number in the volume.</li>
	 * <li>henshall - "A Guide To Remembering Japanese Characters" by Kenneth G.
	 * Henshall.</li>
	 * <li>sh_kk - "Kanji and Kana" by Spahn and Hadamitzky.</li>
	 * <li>sakade - "A Guide To Reading and Writing Japanese" edited by Florence
	 * Sakade.</li>
	 * <li>jf_cards - Japanese Kanji Flashcards, by Max Hodges and Tomoko
	 * Okazaki. (Series 1)</li>
	 * <li>henshall3 - "A Guide To Reading and Writing Japanese" 3rd edition,
	 * edited by Henshall, Seeley and De Groot.</li>
	 * <li>tutt_cards - Tuttle Kanji Cards, compiled by Alexander Kask.</li>
	 * <li>crowley - "The Kanji Way to Japanese Language Power" by Dale Crowley.
	 * </li>
	 * <li>kanji_in_context - "Kanji in Context" by Nishiguchi and Kono.</li>
	 * <li>busy_people - "Japanese For Busy People" vols I-III, published by the
	 * AJLT. The codes are the volume.chapter.</li>
	 * <li>kodansha_compact - the "Kodansha Compact Kanji Guide".</li>
	 * <li>maniette - codes from Yves Maniette's "Les Kanjis dans la tete"
	 * French adaptation of Heisig.</li>
	 * </ul>
	 */
	public HashMap<String, String> dicNumber;

	public static final HashMap<String, String> dictNames;

	static {
		dictNames = new HashMap<String, String>(20);
		dictNames.put("nelson_c", "\"Modern Reader's Japanese-English Character Dictionary\""
				+ " edited by Andrew Nelson (now published as the \"Classic\" Nelson).");
		dictNames.put("nelson_n", "\"The New Nelson Japanese-English Character Dictionary\", edited by John Haig.");
		dictNames.put("halpern_njecd", "\"New Japanese-English Character Dictionary\", edited by Jack Halpern.");
		dictNames.put("halpern_kkld", "\"Kanji Learners Dictionary\" (Kodansha) edited by Jack Halpern.");
		dictNames.put("heisig", "\"Remembering The  Kanji\" by James Heisig.");
		dictNames.put("gakken", "\"A  New Dictionary of Kanji Usage\" (Gakken)");
		dictNames.put("oneill_names", "\"Japanese Names\", by P.G. O'Neill.");
		dictNames.put("oneill_kk", "\"Essential Kanji\" by P.G. O'Neill.");
		dictNames.put("moro", "\"Daikanwajiten\" compiled by Morohashi.");
		dictNames.put("henshall", "\"A Guide To Remembering Japanese Characters\" by Kenneth G. Henshall.");
		dictNames.put("sh_kk", "\"Kanji and Kana\" by Spahn and Hadamitzky.");
		dictNames.put("sakade", "\"A Guide To Reading and Writing Japanese\" edited by Florence Sakade.");
		dictNames.put("jf_cards", "\"Japanese Kanji Flashcards\", by Max Hodges and Tomoko Okazaki. (Series 1)");
		dictNames.put("henshall3", "\"A Guide To Reading and Writing Japanese\" "
				+ "3rd edition, edited by Henshall, Seeley and De Groot.");
		dictNames.put("tutt_cards", "Tuttle Kanji Cards, compiled by Alexander Kask.");
		dictNames.put("crowley", "\"The Kanji Way to Japanese Language Power\" by Dale Crowley.");
		dictNames.put("kanji_in_context", "\"Kanji in Context\" by Nishiguchi and Kono.");
		dictNames.put("busy_people", "\"Japanese For Busy People\" vols I-III, "
				+ "published by the AJLT. The codes are the volume.chapter");
		dictNames.put("kodansha_compact", "the \"Kodansha Compact Kanji Guide\".");
		dictNames.put("maniette", "codes from Yves Maniette's \"Les Kanjis dans la tête\""
				+ "French adaptation of Heisig.");

	}

	public Kanji(int codepoint) {
		this.codepoint = codepoint;
		this.group = null;
		this.frequency = null;
		this.grade = null;
		this.jlpt = null;
		this.strokeCount = null;
		this.readingsKun = null;
		this.readingsOn = null;
		this.dicNumber = null;
	}

	public void addReadingsOn(String text) {
		if (null == this.readingsOn) {
			this.readingsOn = text;
		} else {
			this.readingsOn += "/" + text;
		}
	}

	public void addReadingsKun(String text) {
		if (null == this.readingsKun) {
			this.readingsKun = text;
		} else {
			this.readingsKun += "/" + text;
		}
	}

	public void addMeaning(String lang, String value) {
		if (null == this.meanings) {
			this.meanings = new HashMap<String, String>(1);
		}
		String oldValue = this.meanings.get(lang);
		if (null == oldValue) {
			this.meanings.put(lang, value);
		} else {
			this.meanings.put(lang, oldValue + "/" + value);
		}
	}

	@Override
	public int compareTo(Kanji o) {
		return this.codepoint - o.codepoint;
	}

	public static interface Element {
	}

	public static class Group implements Element {
		public final int id;
		public Integer element;
		public Integer position;
		public Integer radical;
		public Integer original;
		public String part;
		public boolean variant;
		public boolean partial;
		public boolean radicalForm;
		public int number;
		public boolean tradForm;
		/**
		 * TBD as of 20120505
		 */
		public String phon;

		public List<Element> elements;

		public Group(String id) {
			int idxChar = id.indexOf('-');
			if (idxChar > 0) {
				this.id = Integer.parseInt(id.substring(idxChar + 2));
			} else {
				this.id = 0;
			}
			this.element = 0;
			this.position = null;
			this.radical = null;
			this.original = 0;
			this.part = null;
			this.variant = false;
			this.partial = false;
			this.radicalForm = false;
			this.number = 0;
			this.tradForm = false;
		}

		public void setPosition(String value) {
			if (null != value) {
				this.position = Arrays.binarySearch(POSITIONS, value);

				if (0 > this.position) {
					System.err.println("Unknown position type: '" + value + "'");
				}
			}
		}

		public void setRadical(String value) {
			if (null != value) {
				this.radical = Arrays.binarySearch(RADICAL, value);

				if (0 > this.radical) {
					Log.e(TAG, "Unknown radical type: " + value);
				}
			}
		}

	}

	public static class Path implements Element {
		public final String id;
		public final String type;
		public final String draw;

		public Path(String id, String type, String draw) {
			this.type = type;
			this.draw = draw;
			this.id = id.substring(id.indexOf('-') + 1);
		}
	}

}
