/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.kanjidb;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

import org.tmatesoft.sqljet.core.SqlJetEncoding;
import org.tmatesoft.sqljet.core.SqlJetException;
import org.tmatesoft.sqljet.core.SqlJetTransactionMode;
import org.tmatesoft.sqljet.core.table.ISqlJetTable;
import org.tmatesoft.sqljet.core.table.SqlJetDb;
import org.vono.narau.Common;
import org.vono.narau.Log;
import org.vono.narau.kanjidb.Kanji.Element;

/**
 * 
 */
public class KanjiDBWriter {
	private static final String TAG = KanjiDBWriter.class.getSimpleName();

	private static final int VALUE_DB_VERSION = 4;

	private static final String DB_NAME_TEMPLATE = "kanjis_%1$s-%2$d.db";

	// KanjiDict
	private static final String TABLE_DICT_KANJI = "kdict_kanji";
	private static final String TABLE_DICT_MEANINGS = "kdict_meanings";
	private static final String TABLE_DICT_INDEX = "kdict_index";
	private static final String TABLE_DICT_REF = "kdict_ref";
	private static final String TABLE_DICT_LANGS = "kdict_langs";

	// KanjiVG
	private static final String TABLE_KVG_GROUP = "kvg_group";
	private static final String TABLE_KVG_PATH = "kvg_path";
	private static final String TABLE_KVG_RAD = "kvg_radicals";
	private static final String TABLE_KVG_POS = "kvg_positions";
	private static final String VIEW_KVG_SCHOOL = "kvg_school";

	private static final String CREATE_TABLE = "CREATE TABLE ";

	/**
	 * The kanji table, merge the primary group. I took only the 2 first more
	 * used elements of the main group:<br/>
	 * <ul>
	 * <li>6404 kvg:element</li>
	 * <li>38 kvg:original</li>
	 * <li>3 kvg:partial</li>
	 * <li>1 kvg:phon</li>
	 * <li>4 kvg:position</li>
	 * <li>247 kvg:radical</li>
	 * <li>1 kvg:radicalForm</li>
	 * <li>31 kvg:variant</li>
	 * </ul>
	 * Unix cmd:
	 * 
	 * <pre>grep '^&lt;g id="kvg:....."' kanjivg.xml | sed 's/&lt;g
	 * id="kvg:....."//' | sed 's/="[^"]*"//g' | sed 's/&gt;//g' | sed 's/ /\n/g' |
	 * sort | uniq -c</pre>
	 */
	private static final String CREATE_TABLE_DICT_KANJI = CREATE_TABLE + TABLE_DICT_KANJI
			+ " (codepoint INTEGER PRIMARY KEY," + "grade INTEGER," + "frequency INTEGER," + "jlpt INTEGER,"
			+ "stroke_count INTEGER, reading_on TEXT, reading_kun TEXT);";

	private static final String CREATE_TABLE_DICT_MEANINGS = CREATE_TABLE + TABLE_DICT_MEANINGS
			+ " (codepoint INTEGER NOT NULL," + "meanings TEXT NOT NULL)"
			+ " CONSTRAINT meanings_pk PRIMARY_KEY(codepoint);";

	private static final String CREATE_TABLE_DICT_LANGS = CREATE_TABLE + TABLE_DICT_LANGS
			+ "(lang TEXT PRIMARY KEY, nb_entries INTEGER, table_name TEXT)";

	/**
	 * This element contains the index numbers and similar unstructured
	 * information such as page numbers in a number of published dictionaries,
	 * and instructional books on kanji.
	 */
	private static final String CREATE_TABLE_DICT_IDX = CREATE_TABLE + TABLE_DICT_INDEX
			+ " (codepoint INTEGER, dictionary INTEGER, index_number TEXT)"
			+ " CONSTRAINT dict_idx_pk PRIMARY_KEY (codepoint, dictionary);";

	private static final String CREATE_TABLE_DICT_REF = CREATE_TABLE + TABLE_DICT_REF
			+ " (id INTEGER, name TEXT, label TEXT);";

	/**
	 * <ul>
	 * <li>5 kvg:radicalForm</li>
	 * <li>7 kvg:tradForm</li>
	 * <li>19 kvg:number</li>
	 * <li>468 kvg:partial</li>
	 * <li>1361 kvg:phon</li>
	 * <li>3425 kvg:part</li>
	 * <li>4487 kvg:original</li>
	 * <li>5892 kvg:variant</li>
	 * <li>35906 kvg:element</li>
	 * <li>7040 kvg:radical</li>
	 * <li>25228 kvg:position</li>
	 * </ul>
	 * 
	 * <pre>
	 * grep -E '<g id="kvg:.....-g[0-9]+"' kanjivg.xml | sed -E 's/.*<g id="kvg:.....-g[0-9]+"//'| sed 's/="[^"]*"//g' | sed 's/>//g' | sed 's/ /\n/g' | sort | uniq -c
	 * </pre>
	 */
	private static final String CREATE_TABLE_KVG_GROUP = CREATE_TABLE
			+ TABLE_KVG_GROUP
			+ " (codepoint INTEGER, id INTEGER, parent INTEGER NOT NULL, element INTEGER, position INTEGER, radical INTEGER, original INTEGER)"
			+ " CONSTRAINT group_pk PRIMARY_KEY (codepoint, id);";
	private static final String CREATE_TABLE_KVG_PATH = CREATE_TABLE + TABLE_KVG_PATH
			+ " (codepoint INTEGER NOT NULL, parent INTEGER NOT NULL, type TEXT, draw TEXT NOT NULL)"
			+ " CONSTRAINT path_pk PRIMARY_KEY (codepoint, parent);";
	/**
	 * radical: 'r'+id ("bottom", "kamae", "kamae1", "kamae2", "left", "nyo",
	 * "right", "tare", "top")
	 */
	private static final String CREATE_TABLE_KVG_RAD = CREATE_TABLE + TABLE_KVG_RAD
			+ " (id INTEGER NOT NULL, value TEXT NOT NULL) " + "CONSTRAINT " + TABLE_KVG_RAD + "_pk PRIMARY_KEY(id);";
	/**
	 * position: 'p'+id ("general", "nelson", "tradit")<br/>
	 */
	private static final String CREATE_TABLE_KVG_POS = CREATE_TABLE + TABLE_KVG_POS
			+ " (id INTEGER  NOT NULL, value TEXT NOT NULL) " + "CONSTRAINT " + TABLE_KVG_POS + "_pk PRIMARY_KEY(id);";

	private static final String CREATE_VIEW_KVG_SCHOOL = "CREATE VIEW " + VIEW_KVG_SCHOOL + " AS " + "SELECT DISTINCT "
			+ TABLE_DICT_KANJI + ".codepoint AS codepoint, " + TABLE_DICT_KANJI + ".jlpt AS jlpt, " + TABLE_DICT_KANJI
			+ ".grade AS grade " + "FROM " + TABLE_DICT_KANJI + "," + TABLE_KVG_PATH + " " + "WHERE "
			+ TABLE_DICT_KANJI + ".codepoint=" + TABLE_KVG_PATH + ".codepoint";
	// and (" + TABLE_DICT_KANJI+ ".jlpt is not null or
	// " + TABLE_DICT_KANJI + ".grade is not null)

	private static final String CREATE_INDEX = "CREATE INDEX idx_%1$s_id ON %1$s (codepoint);";

	private final ArrayList<Kanji> listKanji;
	/**
	 * Ugly hack to get the dict ref id in the database
	 */
	private HashMap<String, Integer> dictRef;

	private File dbDirectory;

	private final HashMap<String, SqlJetDb> langsDB;

	public KanjiDBWriter(ArrayList<Kanji> kanjis) {
		this.listKanji = kanjis;
		this.dictRef = null;
		this.dbDirectory = null;
		this.langsDB = new HashMap<String, SqlJetDb>();
	}

	public void createDB(File dbDir) throws SqlJetException, IOException {
		this.dbDirectory = dbDir;

		this.createDBJpn();

		final String createIndexCmd = String.format(CREATE_INDEX, TABLE_DICT_MEANINGS);
		for (String lang : this.langsDB.keySet()) {
			SqlJetDb db = this.langsDB.get(lang);

			db.createIndex(createIndexCmd);

			db.commit();
			db.close();
		}
		this.langsDB.clear();
	}

	private void createDBJpn() throws SqlJetException, IOException {

		final SqlJetDb db = createDBLang(Common.LANG_JAPANESE);

		try {
			KanjiDBWriter.createTablesJpn(db);

			final ISqlJetTable table = db.getTable(TABLE_DICT_REF);
			this.writeDictRef(db, table);

			ISqlJetTable tableKanji = db.getTable(TABLE_DICT_KANJI);
			ISqlJetTable tableGroup = db.getTable(TABLE_KVG_GROUP);
			ISqlJetTable tablePath = db.getTable(TABLE_KVG_PATH);
			ISqlJetTable tableDictIdx = db.getTable(TABLE_DICT_INDEX);

			final HashMap<String, ISqlJetTable> tablesMeanings = new HashMap<String, ISqlJetTable>(4);
			final HashMap<String, Integer> meaningCount = new HashMap<String, Integer>(4);

			db.beginTransaction(SqlJetTransactionMode.WRITE);
			final ArrayList<Kanji> lList = this.listKanji;
			final int count = lList.size();
			for (int i = 0; i < count; i++) {

				final Kanji kanji = lList.get(i);
				final int codepoint = kanji.codepoint;

				tableKanji.insert(codepoint, kanji.grade, kanji.frequency, kanji.jlpt, kanji.strokeCount,
						kanji.readingsOn, kanji.readingsKun);

				this.writeElement(tableGroup, tablePath, codepoint, -1, kanji.group);

				if (null != kanji.dicNumber) {
					this.writeDictIndex(tableDictIdx, codepoint, kanji.dicNumber);
				}

				if (null != kanji.meanings) {
					Set<Entry<String, String>> meanings = kanji.meanings.entrySet();
					for (Entry<String, String> meaning : meanings) {
						final String lang = meaning.getKey();

						ISqlJetTable tableMeanings = tablesMeanings.get(lang);
						if (null == tableMeanings) {
							SqlJetDb langDB = this.createDBLang(lang);
							this.langsDB.put(lang, langDB);

							KanjiDBWriter.createTablesLang(langDB, lang);
							langDB.beginTransaction(SqlJetTransactionMode.WRITE);

							tableMeanings = langDB.getTable(TABLE_DICT_MEANINGS);
							tablesMeanings.put(lang, tableMeanings);
							meaningCount.put(lang, 1);
						} else {
							// Not optimise, but for this program, it is not
							// really important
							meaningCount.put(lang, meaningCount.get(lang) + 1);
						}
						tableMeanings.insert(codepoint, meaning.getValue());
					}
				}
			}

			final ISqlJetTable tableDictLangs = db.getTable(TABLE_DICT_LANGS);
			final Set<Entry<String, Integer>> setLangsCount = meaningCount.entrySet();
			for (Entry<String, Integer> entry : setLangsCount) {
				final String lang = entry.getKey();
				tableDictLangs.insert(lang, entry.getValue(), getDBName(lang));
			}

			final ISqlJetTable tableKPos = db.getTable(TABLE_KVG_POS);
			final int lenPos = Kanji.POSITIONS.length;
			for (int i = 0; i < lenPos; i++) {
				tableKPos.insert(i, Kanji.POSITIONS[i]);
			}
			final ISqlJetTable tableKRad = db.getTable(TABLE_KVG_RAD);
			final int lenRad = Kanji.RADICAL.length;
			for (int i = 0; i < lenRad; i++) {
				tableKRad.insert(i, Kanji.RADICAL[i]);
			}

			db.createIndex(String.format(CREATE_INDEX, TABLE_DICT_KANJI));
			db.createIndex(String.format(CREATE_INDEX, TABLE_KVG_GROUP));
			db.createIndex(String.format(CREATE_INDEX, TABLE_KVG_PATH));
			db.createIndex(String.format(CREATE_INDEX, TABLE_DICT_INDEX));

			db.createView(CREATE_VIEW_KVG_SCHOOL);

		} finally {
			db.commit();
			db.close();
		}
	}

	private void writeElement(ISqlJetTable tableGroup, ISqlJetTable tablePath, int codepoint, int parentId,
			Kanji.Element element) throws SqlJetException {
		if (element instanceof Kanji.Group) {
			final Kanji.Group group = (Kanji.Group) element;
			tableGroup.insert(codepoint, group.id, parentId, group.element, group.position, group.radical,
					group.original);

			final List<Element> elements = group.elements;
			final int numElem = elements.size();
			if (0 == numElem) {
				Log.w(TAG, "codepoint " + Integer.toHexString(codepoint) + " group " + group.id
						+ " with no sub element ");
			}

			for (int i = 0; i < numElem; i++) {
				writeElement(tableGroup, tablePath, codepoint, group.id, elements.get(i));
			}
		} else if (element instanceof Kanji.Path) {
			Kanji.Path path = (Kanji.Path) element;
			tablePath.insert(codepoint, parentId, path.type, path.draw);
		}
	}

	private void writeDictIndex(ISqlJetTable tableDictIdx, int codepoint, HashMap<String, String> dicNumber)
			throws SqlJetException {
		for (Entry<String, String> es : dicNumber.entrySet()) {

			final Integer id = this.dictRef.get(es.getKey());
			final String indexNumber = es.getValue();
			tableDictIdx.insert(codepoint, id, indexNumber);

		}
	}

	private void writeDictRef(SqlJetDb db, ISqlJetTable tableDictRef) throws SqlJetException {
		int pos = 0;
		this.dictRef = new HashMap<String, Integer>();

		db.beginTransaction(SqlJetTransactionMode.WRITE);

		for (String key : Kanji.dictNames.keySet()) {
			String val = Kanji.dictNames.get(key);
			tableDictRef.insert(pos, key, val);
			this.dictRef.put(key, pos);
			pos++;
		}
		db.commit();
	}

	private SqlJetDb createDBLang(String lang) throws SqlJetException, IOException {
		final String dbFilename = getDBName(lang);

		Log.i(TAG, "Creating " + dbFilename);

		final File dbFile = new File(this.dbDirectory, dbFilename);
		if (dbFile.isFile()) {
			Log.i(TAG, dbFile.getName() + " exists, deleting...");
			if (!dbFile.delete()) {
				final String msg = "can't delete " + dbFile.getName();
				throw new IOException(msg);
			}
		}

		final SqlJetDb db = new SqlJetDb(dbFile, true);

		db.open();
		db.getOptions().setAutovacuum(true);
		db.getOptions().setEncoding(SqlJetEncoding.UTF8);

		return db;
	}

	private static void createTablesJpn(SqlJetDb db) throws SqlJetException {
		Common.createTablesCommon(db, VALUE_DB_VERSION, Common.LANG_JAPANESE);

		db.createTable(CREATE_TABLE_DICT_LANGS);
		db.createTable(CREATE_TABLE_DICT_KANJI);
		db.createTable(CREATE_TABLE_DICT_IDX);
		db.createTable(CREATE_TABLE_DICT_REF);
		db.createTable(CREATE_TABLE_KVG_GROUP);
		db.createTable(CREATE_TABLE_KVG_PATH);
		db.createTable(CREATE_TABLE_KVG_POS);
		db.createTable(CREATE_TABLE_KVG_RAD);

	}

	private static void createTablesLang(SqlJetDb db, String lang) throws SqlJetException {
		Common.createTablesCommon(db, VALUE_DB_VERSION, lang);

		db.createTable(CREATE_TABLE_DICT_MEANINGS);
	}

	private static String getDBName(String lang) {
		return String.format(DB_NAME_TEMPLATE, lang, VALUE_DB_VERSION);
	}

}
