/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.kanjidb;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.vono.narau.Common;
import org.vono.narau.Log;
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

/**
 * Read kanjidic2.xml files.
 * 
 */
public class KanjiDictReader {
	private static final String TAG = KanjiDictReader.class.getSimpleName();

	/**
	 * The xml file format version this class can read.
	 */
	private static final String FILE_VERSION = "4";

	private XmlPullParser xmlPullParser;
	private Map<Integer, Kanji> mapKanjis;
	private File file;

	public KanjiDictReader(File xmlFile, Map<Integer, Kanji> kanjis) throws XmlPullParserException,
			FileNotFoundException {
		XmlPullParserFactory factory = XmlPullParserFactory.newInstance();
		this.xmlPullParser = factory.newPullParser();
		this.xmlPullParser.setInput(new FileReader(xmlFile));
		this.mapKanjis = kanjis;
		this.file = xmlFile;
	}

	public void read() throws XmlPullParserException, IOException {
		final XmlPullParser xpp = this.xmlPullParser;

		int eventType = 0;
		int lastEventType = -1;
		do {
			eventType = xpp.getEventType();

			switch (eventType) {
			case XmlPullParser.START_DOCUMENT:
				Log.i(TAG, "Start reading " + this.file.getName());
				break;
			case XmlPullParser.ENTITY_REF:
				String name = xpp.getName();
				Log.i(TAG, "Entity ref " + name);
				break;
			case XmlPullParser.COMMENT:
				// Log.i(TAG, "Comment:\n");
				break;
			case XmlPullParser.TEXT:
				if (XmlPullParser.COMMENT == lastEventType || XmlPullParser.DOCDECL == lastEventType) {
					// Log.i(TAG, xpp.getText());
				}
				// String text = xpp.getText();
				// if (!text.trim().equals("\n"))
				// Log.i(TAG, "TEXT" + text);
				break;
			case XmlPullParser.START_TAG:
				String tagName = xpp.getName();
				if ("character".equals(tagName)) {
					this.readCharacter();
				} else if ("header".equals(tagName)) {
					this.readHeader();
				} else if ("kanjidic2".equals(tagName)) {
					Log.i(TAG, "Start kanjidic2");
				}

				break;
			case XmlPullParser.END_TAG:
				tagName = xpp.getName();
				if ("kanjidic2".equals(tagName)) {
					Log.i(TAG, "End kanjidic2");
				} else {
					Log.e(TAG, String.format(Common.ERR_UNEXPECTED_END_TAG, "KanjiDictReader.read()", tagName,
							xpp.getLineNumber()));
				}
				break;
			case XmlPullParser.END_DOCUMENT:
				Log.i(TAG, "End reading");
				break;
			case XmlPullParser.DOCDECL:
				// Log.i(TAG, "DOCDECL:" + xpp.getText());

				break;
			case XmlPullParser.IGNORABLE_WHITESPACE:
				break;
			default:
				Log.e(TAG, String.format(Common.ERR_UNKNOWN_EVENTTYPE, "KanjiDictReader.read()", eventType,
						xpp.getLineNumber()));
			}

			lastEventType = eventType;
			eventType = xpp.nextToken();
		} while (XmlPullParser.END_DOCUMENT != eventType);
	}

	private void readHeader() throws XmlPullParserException, IOException {
		final XmlPullParser xpp = this.xmlPullParser;
		boolean loop = true;
		while (loop) {
			int eventType = xpp.next();
			String tagName = xpp.getName();
			if (eventType == XmlPullParser.START_TAG && "file_version".equals(tagName)) {
				eventType = xpp.next();
				if (XmlPullParser.TEXT == eventType) {
					String version = xpp.getText();
					if (!FILE_VERSION.equals(version)) {
						Log.e(TAG, "Warning: the file version is " + version + " this program is for version "
								+ FILE_VERSION);
					}
				}
			} else if (eventType == XmlPullParser.END_TAG && "header".equals(tagName)) {
				loop = false;
			} else if (eventType == XmlPullParser.END_DOCUMENT) {
				// to be sure to avoid an infinite loop
				loop = false;
			}

		} // while (loop)
	}

	/**
	 * Read the character tag
	 * 
	 * @throws IOException
	 * @throws XmlPullParserException
	 */
	private void readCharacter() throws XmlPullParserException, IOException {
		final XmlPullParser xpp = this.xmlPullParser;

		boolean loop = true;
		Kanji kanji = null;

		String tagName = null;
		while (loop) {
			int eventType = xpp.next();

			tagName = xpp.getName();

			if (XmlPullParser.START_TAG == eventType) {
				if ("codepoint".equals(tagName)) {
					int codepoint = this.getUnicodeCodepoint();
					if (codepoint > 0) {
						kanji = this.mapKanjis.get(codepoint);
						if (null == kanji) {
							kanji = new Kanji(codepoint);
							this.mapKanjis.put(codepoint, kanji);
						}
					}
				} else if ("misc".equals(tagName)) {
					this.readMisc(kanji);
				} else if ("reading_meaning".equals(tagName)) {
					this.readReadingMeaning(kanji);
				} else if ("dic_number".equals(tagName)) {
					this.readDicNumber(kanji);
				}

			} else if (XmlPullParser.END_TAG == eventType && "character".equals(tagName)) {
				loop = false;
			} else if (eventType == XmlPullParser.END_DOCUMENT) {
				// to be sure to avoid an infinite loop
				loop = false;
			}
		}

	}

	private int getUnicodeCodepoint() throws XmlPullParserException, IOException {
		final XmlPullParser xpp = this.xmlPullParser;
		boolean loop = true;
		int codepoint = 0;
		while (loop) {
			int eventType = xpp.next();
			String tagName = xpp.getName();
			if (XmlPullParser.START_TAG == eventType && "cp_value".equals(tagName)) {
				String value = xpp.getAttributeValue(null, "cp_type");
				if ("ucs".equals(value)) {
					eventType = xpp.next();
					if (XmlPullParser.TEXT == eventType) {
						value = xpp.getText();
						codepoint = Integer.parseInt(value, 16);
					}
				}
			} else if (XmlPullParser.END_TAG == eventType && "codepoint".equals(tagName)) {
				loop = false;
			}

		}

		return codepoint;
	}

	/**
	 * Read the misc tag (grade, frequency, jlpt, stroke_count)
	 * 
	 * @param kanji
	 * @throws XmlPullParserException
	 * @throws IOException
	 */
	private void readMisc(Kanji kanji) throws XmlPullParserException, IOException {
		final XmlPullParser xpp = this.xmlPullParser;
		if (null == kanji) {
			Log.e(TAG, "Can't read a null kanji at line " + xpp.getLineNumber());
			return;
		}

		boolean loop = true;
		while (loop) {
			int eventType = xpp.next();
			String tagName = xpp.getName();
			if (eventType == XmlPullParser.START_TAG) {
				if ("grade".equals(tagName)) {
					eventType = xpp.next();
					if (XmlPullParser.TEXT == eventType) {
						kanji.grade = Integer.parseInt(xpp.getText());
					} else {
						Log.e(TAG, "Expected TEXT \"grade\" value at line " + xpp.getLineNumber());
					}

				} else if ("stroke_count".equals(tagName)) {
					eventType = xpp.next();
					if (XmlPullParser.TEXT == eventType) {
						kanji.strokeCount = Integer.parseInt(xpp.getText());
					} else {
						Log.e(TAG, "Expected TEXT \"stroke_count\" value at line " + xpp.getLineNumber());
					}

				} else if ("freq".equals(tagName)) {
					eventType = xpp.next();
					if (XmlPullParser.TEXT == eventType) {
						kanji.frequency = Integer.parseInt(xpp.getText());
					} else {
						Log.e(TAG, "Expected TEXT \"freq\" value at line " + xpp.getLineNumber());
					}
				} else if ("jlpt".equals(tagName)) {
					eventType = xpp.next();
					if (XmlPullParser.TEXT == eventType) {
						kanji.jlpt = Integer.parseInt(xpp.getText());
					} else {
						Log.e(TAG, "Expected TEXT \"jlpt\" value at line " + xpp.getLineNumber());
					}
				}
			} else if (eventType == XmlPullParser.END_TAG) {
				if ("misc".equals(tagName)) {
					loop = false;
				}
			}
		} // while (loop)

	} // readMisc

	private void readReadingMeaning(Kanji kanji) throws XmlPullParserException, IOException {
		final XmlPullParser xpp = this.xmlPullParser;
		if (null == kanji) {
			Log.e(TAG, "Can't read a null kanji at line " + xpp.getLineNumber());
			return;
		}

		boolean loop = true;
		while (loop) {
			int eventType = xpp.next();
			String tagName = xpp.getName();
			if (eventType == XmlPullParser.START_TAG) {
				if ("rmgroup".equals(tagName)) {
					this.readRmgroup(kanji);
				}
			} else if (eventType == XmlPullParser.END_TAG) {
				if ("reading_meaning".equals(tagName)) {
					loop = false;
				}
			}
		}

	}

	private void readRmgroup(Kanji kanji) throws XmlPullParserException, IOException {
		final XmlPullParser xpp = this.xmlPullParser;
		boolean loop = true;
		while (loop) {
			int eventType = xpp.next();
			String tagName = xpp.getName();
			if (eventType == XmlPullParser.START_TAG) {
				if ("reading".equals(tagName)) {
					String rType = xpp.getAttributeValue(null, "r_type");
					if ("ja_on".equals(rType)) {
						eventType = xpp.next();
						if (XmlPullParser.TEXT == eventType) {
							kanji.addReadingsOn(xpp.getText());
						}
					} else if ("ja_kun".equals(rType)) {
						eventType = xpp.next();
						if (XmlPullParser.TEXT == eventType) {
							kanji.addReadingsKun(xpp.getText());
						}
					}
				} else if ("meaning".equals(tagName)) {
					String mLang = xpp.getAttributeValue(null, "m_lang");
					if (null == mLang) {
						mLang = Common.LANG_DEFAULT;
					} else {
						mLang = Common.getLang3(mLang);
					}

					eventType = xpp.next();
					if (XmlPullParser.TEXT == eventType) {
						kanji.addMeaning(mLang, xpp.getText());
					}
				}
			} else if (eventType == XmlPullParser.END_TAG) {
				if ("rmgroup".equals(tagName)) {
					loop = false;
				}
			}
		}

	}

	private void readDicNumber(Kanji kanji) throws XmlPullParserException, IOException {
		final XmlPullParser xpp = this.xmlPullParser;
		boolean loop = true;
		kanji.dicNumber = new HashMap<String, String>();
		while (loop) {
			int eventType = xpp.next();
			String tagName = xpp.getName();
			if (eventType == XmlPullParser.START_TAG) {
				if ("dic_ref".equals(tagName)) {
					String dictName = xpp.getAttributeValue(null, "dr_type");
					String vol = xpp.getAttributeValue(null, "m_vol");
					String page = xpp.getAttributeValue(null, "m_page");
					eventType = xpp.next();
					if (XmlPullParser.TEXT == eventType) {
						String value = xpp.getText();
						if (null != page) {
							value = "page " + page + " #" + value;
						}
						if (null != vol) {
							value = "vol " + vol + " " + value;
						}
						kanji.dicNumber.put(dictName, value);
					}
				}
			} else if (eventType == XmlPullParser.END_TAG) {
				if ("dic_number".equals(tagName)) {
					loop = false;
				}
			}
		}

	}

}
