/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.kanjidb;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Map;

import org.vono.narau.Common;
import org.vono.narau.Log;
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

/**
 * http://kanjivg.tagaini.net
 * 
 */
public class KanjiVGReader {
	private static final String TAG = KanjiVGReader.class.getSimpleName();

	private XmlPullParser xmlPullParser;
	private Map<Integer, Kanji> mapKanjis;

	public KanjiVGReader(Map<Integer, Kanji> kanjis) throws XmlPullParserException {
		XmlPullParserFactory factory = XmlPullParserFactory.newInstance();
		this.xmlPullParser = factory.newPullParser();
		this.mapKanjis = kanjis;
	}

	public boolean read(File xmlFile) throws XmlPullParserException, IOException {
		final XmlPullParser xpp = this.xmlPullParser;
		xpp.setInput(new FileReader(xmlFile));

		int eventType = 0;
		int lastEventType = -1;
		boolean ret = true;
		do {
			eventType = xpp.getEventType();
			switch (eventType) {
			case XmlPullParser.START_DOCUMENT:
				// Log.i(TAG, "Start reading " + xmlFile.getName());
				break;
			case XmlPullParser.ENTITY_REF:
				String name = xpp.getName();
				Log.i(TAG, "Entity ref " + name);
				break;
			case XmlPullParser.COMMENT:
				// Log.i(TAG, "Comment:\n");
				break;
			case XmlPullParser.TEXT:
				if (XmlPullParser.COMMENT == lastEventType || XmlPullParser.DOCDECL == lastEventType) {
					// Log.i(TAG, xpp.getText());
				}
				// String text = xpp.getText();
				// if (!text.trim().equals("\n"))
				// Log.i(TAG, "TEXT" + text);
				break;
			case XmlPullParser.START_TAG:
				String tagName = xpp.getName();
				if ("g".equals(tagName)) {
					final int count = xpp.getAttributeCount();
					for (int i = 0; i < count; i++) {
						String attrName = xpp.getAttributeName(i);
						String attrValue = xpp.getAttributeValue(i);
						if ("id".equals(attrName) && attrValue.contains("StrokePaths")) {
							i = count;
							Kanji kanji = this.readKanjiTag();
							if (null != kanji) {
								this.mapKanjis.put(kanji.codepoint, kanji);
							} else {
								ret = false;
							}
						} else if ("id".equals(attrName) && attrValue.contains("StrokeNumbers")) {
							this.readStrokeNumbers();
							i = count;
						}
					}

				} else if ("svg".equals(tagName)) {
					// Log.i(TAG, "Begin kanjivg");
				} else {
					Log.e(TAG,
							String.format(Common.ERR_UNEXPECTED_START_TAG, "KanjiVGReader.read()", tagName,
									xpp.getLineNumber()));
				}
				break;
			case XmlPullParser.END_TAG:
				tagName = xpp.getName();
				if ("svg".equals(tagName) || "g".equals(tagName)) {
					// Log.i(TAG, "End kanjivg");
				} else {
					Log.e(TAG,
							String.format(Common.ERR_UNEXPECTED_END_TAG, "KanjiVGReader.read()", tagName,
									xpp.getLineNumber()));
				}
				break;
			case XmlPullParser.END_DOCUMENT:
				Log.i(TAG, "End reading");
				break;
			case XmlPullParser.DOCDECL:
				// Log.i(TAG, "DOCDECL:" + xpp.getText());

				break;
			case XmlPullParser.IGNORABLE_WHITESPACE:
				break;
			default:
				Log.e(TAG,
						String.format(Common.ERR_UNKNOWN_EVENTTYPE, "KanjiVGReader.read()", eventType,
								xpp.getLineNumber()));
			}
			lastEventType = eventType;
			eventType = xpp.nextToken();
		} while (XmlPullParser.END_DOCUMENT != eventType && ret);

		return ret;
	}

	private void readStrokeNumbers() throws XmlPullParserException, IOException {
		XmlPullParser xpp = this.xmlPullParser;
		boolean loop = true;
		while (loop) {
			final int eventType = xpp.nextToken();
			if (XmlPullParser.END_TAG == eventType && "g".equals(xpp.getName())) {
				loop = false;
			}
		}

	}

	private Kanji readKanjiTag() throws XmlPullParserException, IOException {
		XmlPullParser xpp = this.xmlPullParser;

		final int count = xpp.getAttributeCount();
		int id = -1;
		for (int i = 0; i < count; i++) {
			String name = xpp.getAttributeName(i);
			String value = xpp.getAttributeValue(i);
			if ("id".equals(name) && value.contains("StrokePaths")) {
				id = Integer.decode(value.replace("kvg:StrokePaths_", "0x"));
			} else if ("id".equals(name)) {
				return null;
			} else if ("style".equals(name)) {
				// ignore
			} else {
				Log.e(TAG,
						String.format(Common.ERR_UNKNOWN_ATTRIBUTE, "KanjiVGReader.readKanjiTag", name,
								xpp.getLineNumber()));
			}
		}
		if (id == -1) {
			Log.e(TAG, "kanji tag without id, line " + xpp.getLineNumber());
			return null;
		}

		Kanji.Group group = null;

		int eventType = 0;
		boolean loop = true;
		do {
			eventType = xpp.nextTag();
			switch (eventType) {
			case XmlPullParser.START_TAG:
				if (group == null && "g".equals(xpp.getName())) {
					group = this.readGroupTag();
				} else {
					Log.e(TAG, String.format(Common.ERR_UNEXPECTED_START_TAG, "KanjiVGReader.readKanjiTag",
							xpp.getName(), xpp.getLineNumber()));
				}
				break;
			case XmlPullParser.END_TAG:
				if ("g".equals(xpp.getName())) {
					loop = false;
				} else {
					Log.e(TAG, String.format(Common.ERR_UNEXPECTED_END_TAG, "KanjiVGReader.readKanjiTag",
							xpp.getName(), xpp.getLineNumber()));
				}
				break;
			default:
				Log.e(TAG,
						String.format(Common.ERR_UNKNOWN_EVENTTYPE, "KanjiVGReader.readKanjiTag", eventType,
								xpp.getLineNumber()));
				break;
			}
		} while (loop);

		Kanji kanji = null;
		if (null != group) {
			if (this.mapKanjis.containsKey(id)) {
				kanji = this.mapKanjis.get(id);
			} else {
				kanji = new Kanji(id);
			}
			kanji.group = group;
		}

		return kanji;
	}

	private Kanji.Group readGroupTag() throws XmlPullParserException, IOException {
		XmlPullParser xpp = this.xmlPullParser;

		String id = null;
		String element = null;
		String position = null;
		String radical = null;
		String original = null;
		String phon = null; // phonetique?
		String part = null;
		String variant = null;
		String partial = null;
		String radicalForm = null;
		String number = null;
		String tradForm = null;

		int count = xpp.getAttributeCount();
		for (int i = 0; i < count; i++) {
			String name = xpp.getAttributeName(i);
			String value = xpp.getAttributeValue(i);
			if ("id".equals(name)) {
				id = value;
			} else if ("kvg:element".equals(name)) {
				element = value;
			} else if ("kvg:position".equals(name)) {
				position = value;
			} else if ("kvg:radical".equals(name)) {
				radical = value;
			} else if ("kvg:original".equals(name)) {
				original = value;
			} else if ("kvg:phon".equals(name)) {
				phon = value;
			} else if ("kvg:part".equals(name)) {
				part = value;
			} else if ("kvg:variant".equals(name)) {
				variant = value;
			} else if ("kvg:partial".equals(name)) {
				partial = value;
			} else if ("kvg:radicalForm".equals(name)) {
				radicalForm = value;
			} else if ("kvg:number".equals(name)) {
				number = value;
			} else if ("kvg:tradForm".equals(name)) {
				tradForm = value;
			} else {
				Log.e(TAG,
						String.format(Common.ERR_UNKNOWN_ATTRIBUTE, "KanjiVGReader.readGroupTag", name,
								xpp.getLineNumber()));
			}
		}

		ArrayList<Kanji.Element> elements = new ArrayList<Kanji.Element>(2);
		int eventType = 0;
		boolean loop = true;
		do {
			eventType = xpp.nextTag();
			switch (eventType) {
			case XmlPullParser.START_TAG:
				String tagName = xpp.getName();
				if ("g".equals(tagName)) {
					Kanji.Group group = this.readGroupTag();
					elements.add(group);
				} else if ("path".equals(tagName)) {
					Kanji.Path path = this.readPathTag();
					elements.add(path);
				} else {
					Log.e(TAG,
							String.format(Common.ERR_UNEXPECTED_START_TAG, "KanjiVGReader.readGroupTag", tagName,
									xpp.getLineNumber()));
				}
				break;
			case XmlPullParser.END_TAG:
				tagName = xpp.getName();
				if ("g".equals(tagName)) {
					loop = false;
				} else {
					Log.e(TAG,
							String.format(Common.ERR_UNEXPECTED_END_TAG, "KanjiVGReader.readGroupTag", tagName,
									xpp.getLineNumber()));
				}
				break;
			default:
				Log.e(TAG,
						String.format(Common.ERR_UNKNOWN_EVENTTYPE, "KanjiVGReader.readGroupTag", eventType,
								xpp.getLineNumber()));
				break;
			}
		} while (loop);

		Kanji.Group group = new Kanji.Group(id);
		group.element = (element == null) ? null : element.codePointAt(0);
		group.setPosition(position);
		group.setRadical(radical);
		group.original = (original == null) ? null : original.codePointAt(0);
		group.phon = phon;
		group.part = part;
		group.variant = "true".equals(variant);
		group.partial = Boolean.parseBoolean(partial);
		group.radicalForm = Boolean.parseBoolean(radicalForm);
		if (null != number)
			group.number = Integer.parseInt(number);
		group.tradForm = Boolean.parseBoolean(tradForm);
		group.elements = elements;

		return group;
	}

	private Kanji.Path readPathTag() throws XmlPullParserException, IOException {
		XmlPullParser xpp = this.xmlPullParser;
		int count = xpp.getAttributeCount();

		String id = null;
		String type = null;
		String draw = null;
		for (int i = 0; i < count; i++) {
			String name = xpp.getAttributeName(i);
			String value = xpp.getAttributeValue(i);
			if ("id".equals(name)) {
				id = value;
			} else if ("d".equals(name)) {
				draw = value;
			} else if ("kvg:type".equals(name)) {
				type = value;
			} else {
				Log.e(TAG,
						String.format(Common.ERR_UNKNOWN_ATTRIBUTE, "KanjiVGReader.readPathTag", name,
								xpp.getLineNumber()));
			}
		}

		Kanji.Path path = null;
		int eventType = xpp.nextTag();
		if (XmlPullParser.END_TAG == eventType && "path".equals(xpp.getName())) {
			path = new Kanji.Path(id, type, draw);
		} else {
			Log.e(TAG,
					String.format(Common.ERR_UNEXPECTED_END_TAG, "KanjiVGReader.readPathTag", xpp.getName(),
							xpp.getLineNumber()));
		}

		return path;
	}

}
