/*
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.kanjidb;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

import org.tmatesoft.sqljet.core.SqlJetException;
import org.vono.narau.Log;
import org.xmlpull.v1.XmlPullParserException;

/**
 * Create the kanji database.
 * 
 */
public class Main {
	private static final String TAG = Main.class.getSimpleName();
	private static final int NUM_KANJIS = 10000;

	private static void usage() {
		System.out.println("usage:");
		System.out.println("{java cmd line} <db_creation_dir> <kanjivg_git_dir> <kanjidic2.xml>");
		System.out.println();
		System.out.println("Example:");
		System.out.println("{java cmd line} . ~/src/kanjivg/ ~/Download/kanjidic2.xml");
		System.out.println();
		System.out.println("Kanjidict file direct download link: "
				+ "http://www.csse.monash.edu.au/~jwb/kanjidic2/kanjidic2.xml.gz");
		System.out.println("Kanjivg: $ git clone https://github.com/KanjiVG/kanjivg.git");
	}

	public static void main(String[] args) {
		String dbDirectoryName = null;
		String svgDirectoryName = null;
		String kDictFilename = null;

		long startTime = System.currentTimeMillis();

		if (args.length < 3) {
			Main.usage();
			return;
		}

		dbDirectoryName = args[0];
		svgDirectoryName = args[1];
		kDictFilename = args[2];

		// Check parameters validity
		File dbDirectory = new File(dbDirectoryName);
		if (!(dbDirectory.isDirectory() && dbDirectory.canWrite())) {
			Log.e(TAG, "db_creation_dir parameter is invalid");
			return;
		} else {
			Log.i(TAG, "Database creation directory: " + dbDirectory.getAbsolutePath());
		}

		File svgDirectory = new File(svgDirectoryName + "/kanji");
		if (!(svgDirectory.isDirectory() && svgDirectory.canRead())) {
			Log.e(TAG, "kanjivg_git_dir parameter is invalid: " + svgDirectory.getAbsolutePath());
			return;
		} else {
			Log.i(TAG, "Found svg directory " + svgDirectory.getAbsolutePath());
		}

		File kDictFile = new File(kDictFilename);
		if (!(kDictFile.isFile() && kDictFile.canRead())) {
			Log.e(TAG, "kanjidic2.xml parameter is invalid");
			Log.e(TAG, "Can't read " + kDictFilename);
			return;
		} else {
			Log.i(TAG, "Found kanji dictionary " + kDictFile.getAbsolutePath());
		}

		KanjiVGReader readerVG = null;
		KanjiDictReader readerDict = null;
		HashMap<Integer, Kanji> mapKanjis = new HashMap<Integer, Kanji>(NUM_KANJIS);

		try {
			File[] svgFiles = svgDirectory.listFiles();
			final int numSVGFiles = svgFiles.length;
			Log.i(TAG, "reading " + numSVGFiles + " svg files");
			readerVG = new KanjiVGReader(mapKanjis);
			boolean error = false;
			for (int i = 0; i < numSVGFiles; i++) {
				final File svgFile = svgFiles[i];

				if (!svgFile.getName().contains("-")) {
					if (!readerVG.read(svgFile)) {
						i = numSVGFiles;
						error = true;
					}
				}
			}

			if (!error) {
				readerDict = new KanjiDictReader(kDictFile, mapKanjis);
				readerDict.read();

				ArrayList<Kanji> listKanji = new ArrayList<Kanji>(mapKanjis.values());
				Collections.sort(listKanji);

				KanjiDBWriter writer = new KanjiDBWriter(listKanji);
				writer.createDB(dbDirectory);
			}
		} catch (FileNotFoundException fnfe) {
			Log.e(TAG, fnfe.getMessage(), fnfe);
		} catch (XmlPullParserException xppe) {
			Log.e(TAG, xppe.getMessage(), xppe);
		} catch (IOException ioe) {
			Log.e(TAG, ioe.getMessage(), ioe);
		} catch (SqlJetException sje) {
			Log.e(TAG, sje.getMessage(), sje);
		}

		long endTime = System.currentTimeMillis();
		long sec = (endTime - startTime) / 1000;
		long msec = (endTime - startTime) % 1000;
		Log.i(TAG, "create db time: " + sec + "." + msec + "s");

	}

}
